/***************************************************************************
                          alignment.cpp  -  description
                             -------------------
    begin                : Thu Feb 20 2003
    copyright            : (C) 2003 by Vinh Le Sy and Arndt von Haeseler
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <iostream>
#include <fstream>
#include <string.h>
#include "vec.h"
#include "ali.h"
#include "ptn.h"
#include "optpairseq.h"
#include "outstream.h"
#include "model.h"

extern int nto[NUM_CHAR];
extern int mapNum[NUM_CHAR];

using namespace std;

int Ali::nSeq_;
int Ali::nSite_;
int Ali::longestSeqNameLen_ = -1;

Vec<Seq> Ali::items_;
Mat<double> Ali::genDisMat_;
//check whether or not the genDismat of this class was cmped
int Ali::isGenDisMatCmped_;
char Ali::fileName_[FILENAME_LEN];
int Ali::ptnArr_[MAX_NUM_SITE];



//construction function
Ali::Ali () {
}

//clean all content of this class
void Ali::clean () {
  nSeq_ = 0;
  nSite_ = 0;
  isGenDisMatCmped_ = 0;
}

//this function is used for constructing this class
void Ali::doConstructor () {
  clean ();
}

//--------------------------------------------------------------------
//everything is inited here
void Ali::init () {
  cmpGenDis (fileName_);
}

//=============================================
//set ptnNo to site
void Ali::setPtn (int siteNo, int ptnNo) {
  ptnArr_[siteNo] = ptnNo;
}

//=============================================
int Ali::getPtn (int siteNo) {
  return ptnArr_[siteNo];
}


//get the order of the sequence whose name is input name
int Ali::getSeqNo (Vec<char> &name) {
  for (int count_ = 0; count_ < nSeq_; count_ ++)
    if (items_[count_].getName () == name) 
      return count_;

  return -1;
}

//--------------------------------------------------------------------
int Ali::getLongestSeqNameLen () {
  if (longestSeqNameLen_ > 0)
    return longestSeqNameLen_;
  
  longestSeqNameLen_ = 3;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
      int seqNameLen_ = items_[seqNo_].getName ().getSize ();
      if (seqNameLen_ > longestSeqNameLen_)
          longestSeqNameLen_ = seqNameLen_;
  }

  return longestSeqNameLen_;
}

//--------------------------------------------------------------------
DATA_TYPE Ali::detectDataType () {
  int nState_ = 0;
  int nDNA_ = 0;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) 
     for (int siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++)
        if (items_[seqNo_].items_[siteNo_] != '?') {
           nState_ ++;
           if (items_[seqNo_].items_[siteNo_] == 'A' || items_[seqNo_].items_[siteNo_] == 'C'
              || items_[seqNo_].items_[siteNo_] == 'G' || items_[seqNo_].items_[siteNo_] == 'T')
              nDNA_ ++;
        }

  if (nDNA_ > 0.75 * nState_)
     return NUCLEOTIDE;
  else 
    return AMINO_ACID;
}

//--------------------------------------------------------------------
PAM_TYPE Ali::detectDataSource () {
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) 
     for (int siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++)
        if (items_[seqNo_].items_[siteNo_] == '?')
           return REAL;
  return SIMULATION;
}

//--------------------------------------------------------------------
//get the name of the sequence at certain orther 
Vec<char> &Ali::getName (int ndNo) {
  static Vec<char> name_;
  name_.clean ();
  if (ndNo < nSeq_) {//it is a external node
    name_ = items_[ndNo].getName (); 
    return name_;
  }
  else { //it is an internal node   
    name_.convert (ndNo);
    return name_;
  }    
}

//--------------------------------------------------------------------
//get the name of the sequence at certain orther
void Ali::getName (int ndNo, Vec<char> &name) {
  if (ndNo < nSeq_) //it is a external node
    name = items_[ndNo].getName ();
  else
    name.convert (ndNo);
}

//--------------------------------------------------------------------
//return the content at the given site, only active sequences are taking into account
void Ali::getPtn (const int siteNo, Ptn &ptn) {
  ptn.setSize (nSeq_);
  int len_ = 0;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_++) {
    ptn [len_] = items_[seqNo_][siteNo];
    len_ ++;
  };
}

//--------------------------------  ------------------------------------
//return the number sequences of this alignment
int Ali::getNSeq () {
  return nSeq_;
}

//--------------------------------------------------------------------
//return the number of nucleotides in a sequence
int Ali::getNSite () {
  return nSite_;
}

//--------------------------------------------------------------------
//return the sequence index
Seq &Ali::getSeq (const int index) {
  return items_[index];
}

//--------------------------------------------------------------------
//read input from the file fileName
int Ali::readInput (const char *fileName) {
  ifstream in;
  in.open (fileName);
  if (!in) 
    return 0;

  int tmpNSeq_ = 0; 
  int tmpNSite_ = 0;
  in >> tmpNSeq_ >> tmpNSite_;
  if (tmpNSeq_ == 0 || tmpNSite_ == 0)
    return 0;
  else {
    setLimit (tmpNSeq_, tmpNSite_);
  }
  
  in.ignore (INFINITIVE, '\n');
  for (int count_ = 0; count_ < nSeq_; count_ ++) {
    items_[count_].readInput (in, nSite_, 1);
    items_[count_].setId (count_);
  }

  while (items_[0].getSize () < tmpNSite_) {
    in.ignore (INFINITIVE, '\n');
       int nNu_ = 0;
    for (int count_ = 0; count_ < nSeq_; count_ ++) 
      nNu_ = items_[count_].readInput (in, nSite_, 0);
        if (nNu_ == 0)
           return 0;
  } 
  memcpy (fileName_, fileName, strlen (fileName));
  return 1;
}


//--------------------------------------------------------------------
//sort sequences by their name, it serves for deburging
void Ali::sortSeq () {
  for (int seqNo1_ = 0; seqNo1_ < nSeq_; seqNo1_ ++) 
    for (int seqNo2_ = seqNo1_ + 1; seqNo2_ < nSeq_; seqNo2_ ++) 
      if (items_[seqNo1_].getName ().cmp (items_[seqNo2_].getName () ) == 1){
         Seq tmpSeq_;
         tmpSeq_ = items_[seqNo1_];
         items_[seqNo1_] = items_[seqNo2_];
         items_[seqNo2_] = tmpSeq_;
      }
}

//--------------------------------------------------------------------
/*return the genetic distance between sequence1 and sequence2, 
the distance is cmped by maximum likelihood
*/
double Ali::getGenDis (const int seqNo1, const int seqNo2) {
  if (isGenDisMatCmped_ == 0)
    cmpGenDis (fileName_);
  return genDisMat_ [seqNo1] [seqNo2];
}

int Ali::checkSameName () {
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    for (int seqNo2_ = seqNo_ + 1; seqNo2_ < nSeq_; seqNo2_ ++)
      if (items_[seqNo_].getName () == items_[seqNo2_].getName ()) {
//          std::cout <<"Fuck!!!!" << endl;
  //        std::cout << seqNo_ << "/ (" << items_[seqNo_].getName () <<")" << endl;
    //      std::cout << seqNo2_ <<"/ (" <<items_[seqNo2_].getName () <<")" << endl;
          std::cout << "The name " << items_[seqNo_].getName () << " appears more than one time in the alignment file ";
          exit (EXIT_FAILURE);
      }
//  std::cout <<"no identical name" << endl;
  return 0;
}



//--------------------------------------------------------------------
//return the address of genetic distance matrix
Mat<double> *Ali::getGenDis () {
  if (isGenDisMatCmped_ == 0)
    cmpGenDis (fileName_);
  return &genDisMat_;
}

//--------------------------------------------------------------------
//create the distance matrix for any two sequences basing on maximum likelihood method
void Ali::cmpGenDis (char *aliFileName) {
//  std::cout <<"computing genetic distance ..." << endl;
  for (int seqNo1_ = 0; seqNo1_ < nSeq_; seqNo1_ ++) {
    genDisMat_[seqNo1_][seqNo1_] = 0.0;
    for (int seqNo2_ = seqNo1_ + 1; seqNo2_ < nSeq_; seqNo2_ ++) {
      genDisMat_[seqNo1_][seqNo2_] = OptPairSeq::cmp (Ali::getSeq (seqNo1_), Ali::getSeq (seqNo2_) );
      genDisMat_[seqNo2_][seqNo1_] = genDisMat_[seqNo1_][seqNo2_];
    }                                                                 
  }
  ofstream out;

  char distName_[FILENAME_LEN];
  strcpy (distName_, aliFileName);
  char distTail_[FILENAME_LEN] = ".iqpnni.dist";
  strcat (distName_, distTail_);

  out.open (distName_);
  OutStream::writeGenDis (genDisMat_, out);
  out.close ();

  isGenDisMatCmped_ = 1;
}


//--------------------------------------------------------------------
//compute the state frequency of this alignment
void Ali::estimateStateFrq (DVec20 &stateFrqArr) {    
  int stateNo_;
  int nState_ = Model::getNState ();


  for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
    stateFrqArr [ stateNo_ ] = 1.0 / nState_;

  int NUM_TIME = 8;
  //app = appeareance
  for (int time_ = 0; time_ < NUM_TIME; time_ ++) {
    DVec20 timeAppArr_;
    for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
      timeAppArr_[stateNo_] = 0.0;

    for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
      Seq *seq_;
      seq_ = &items_[seqNo_];
      for (int siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++) {
        int stateNo_ = nto [ static_cast<int> ((*seq_)[siteNo_]) ];

     DVec20 siteAppArr_; //App = appearance
     Ali::getApp (stateNo_, siteAppArr_);

     DVec20 newSiteAppArr_;
     double totalSiteApp_ = 0.0;
     for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {
          newSiteAppArr_[stateNo_] = stateFrqArr[stateNo_] * siteAppArr_[stateNo_];
          totalSiteApp_ += newSiteAppArr_[stateNo_];
        }

        for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
          timeAppArr_[stateNo_] += newSiteAppArr_[stateNo_] / totalSiteApp_;
      }
    }
     
    double totalTimeApp_ = 0.0;
    int stateNo_;
    for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
      totalTimeApp_ += timeAppArr_[stateNo_];


    for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
      stateFrqArr[stateNo_] = timeAppArr_[stateNo_] / totalTimeApp_;
  
  } //end of for time_
  
//  std::cout << "state frequency ..." << endl;
 // for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
   // std::cout << stateFrqArr[stateNo_] << endl;

}


double Ali::estimateProb (BASE baseX, BASE baseY) {
  double totalFrq_ = 0.0;
  for (int siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++) {
     double nBaseX_ = 0.0;
     double nBaseY_ = 0.0;

     for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
        if (nto[ static_cast<int> (items_[seqNo_].items_[siteNo_]) ] == baseX)
          nBaseX_ += 1.0;
        if (nto[ static_cast<int> (items_[seqNo_].items_[siteNo_]) ] == baseY)
          nBaseY_ += 1.0;
     }
 
     totalFrq_ += (nBaseX_ * nBaseY_);
  }
  
  double prob_ = totalFrq_ / (nSeq_ * (nSeq_ - 1) * nSite_);
  return prob_;
}

//==========================================================================================
void Ali::estimateGenPam (double &tsAG, double &tsCT, double &tvAC, double &tvAT, double &tvCG, double &tvGT) {
   tsAG = estimateProb (BASE_A, BASE_G) * 2.0;
   tsCT = estimateProb (BASE_C, BASE_T)  * 2.0;
   tvAC = estimateProb (BASE_A, BASE_C);
   tvAT = estimateProb (BASE_A, BASE_T);
   tvCG = estimateProb (BASE_C, BASE_G);
   tvGT = estimateProb (BASE_G, BASE_T);
}

//==========================================================================================
double Ali::estimateTsTvRatio () {
   double tsAG_ = estimateProb (BASE_A, BASE_G) * 2.0;
   double tsCT_ = estimateProb (BASE_C, BASE_T)  * 2.0;
   double tvAC_ = estimateProb (BASE_A, BASE_C);
   double tvAT_ = estimateProb (BASE_A, BASE_T);
   double tvCG_ = estimateProb (BASE_C, BASE_G);
   double tvGT_ = estimateProb (BASE_G, BASE_T);
  
  return 2.0 * (tsAG_ + tsCT_) / (tvAC_ + tvAT_ + tvCG_ +tvGT_);
}


//set the memory for this class
void Ali::setLimit (int nSeq, int nSite) {
  //do not need set limit again
  if (nSeq_ == nSeq && nSite_ == nSite)
    return;

  nSeq_ = nSeq;
  nSite_ = nSite;
  
  genDisMat_.setLimit (nSeq_, nSeq_);

  items_.set (nSeq_, nSeq_);
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    items_[seqNo_].set (nSite_, nSite_);
}

//--------------------------------------------------------------------
//release all memmory for this class
void Ali::release () {
  genDisMat_.release ();
  
  int realLimit_ = items_.getLimit ();
  items_.setSize (realLimit_);
  for (int count_ = 0; count_ < realLimit_; count_ ++)
    items_[count_].release ();  
  items_.release ();
}

void Ali::getApp (int stateNo, DVec20 &appArr) {
  int nState_ = Model::getNState ();
  int count_;
  for (count_ = 0; count_ < nState_; count_ ++)
    appArr[count_] = 0;
  
  if (Model::getDataType () == AMINO_ACID) {
    if (stateNo == BS_UNKNOWN)
        for (count_ = 0; count_ < nState_; count_ ++)
           appArr[count_] = 1;
    else
        appArr[stateNo] = 1;

    return ;
  }

  int ta = 0;
  int tc = 0;
  int tg = 0;
  int tt = 0;
  switch (stateNo) {
    case BASE_A:
      ta = 1;
      break;
    case BASE_C:
      tc = 1;
      break;
    case BASE_G:
      tg = 1;
      break;
    case BASE_T:
      tt = 1;
      break;
    case BASE_AC:
      ta = tc = 1;
      break;
    case BASE_AG:
      ta = tg = 1;
      break;
    case BASE_AT:
      ta = tt = 1;
      break;
    case BASE_CG:
      tc = tg = 1;
      break;
    case BASE_CT:
      tc = tt = 1;
      break;
    case BASE_GT:
      tg = tt = 1;
      break;
    case BASE_ACG:
      ta = tc = tg = 1;
      break;
    case BASE_ACT:
      ta = tc = tt = 1;
      break;
    case BASE_AGT:
      ta = tg = tt = 1;
      break;
    case BASE_CGT:
      tc = tg = tt = 1;
      break;
    default:
      ta = tc = tg = tt = 1;
  } //end of switch

  appArr[BASE_A] = ta;
  appArr[BASE_C] = tc;
  appArr[BASE_G] = tg;
  appArr[BASE_T] = tt;

}

//--------------------------------------------------------------------
//destruction function
Ali::~Ali () {
  release ();
//  std::cout << "this is the destructor of ali class " << endl;
}
