/***************************************************************************
                          alignment.h  -  description
                             -------------------
    begin                : Thu Feb 20 2003
    copyright            : (C) 2003 by Vinh Le Sy and Arndt von Haeseler
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ALI_H
#define ALI_H

#include <iostream>
#include <fstream>

#include "vec.h"
#include "seq.h"
#include "ptn.h"
#include "mat.h"
#include "dvec20.h"
#include "dmat20.h"

using namespace std;

/*
The Alignment class contains all information and mothods for DNA alignment and Protein alignment,
Note: The Alignment class does not inherit from Sequence class
...Being specified more
*/
class Ali
{
public :
static char fileName_[FILENAME_LEN];

//construction function to init data
Ali ();
//use this function for constructing this class
static void doConstructor ();

//clean all content of this class
static void clean ();

//everything is inited here
static void init ();
static PAM_TYPE detectDataSource ();
static DATA_TYPE detectDataType ();

static int getLongestSeqNameLen ();

//return the content at the given site, only active sequences are taking into account
static void getPtn (const int siteNo, Ptn &ptn);

//return the number sequences of this alignment
static int getNSeq ();

//return the number of nucleotides in a sequence
static int getNSite ();

//return the sequence indexth
static Seq &getSeq (const int index);

//sort sequences by their name, it serves for deburging
static void sortSeq ();

//read input from the file fileName
static int readInput (const char *fileName);


static double getGenDis (const int seqNo1, const int seqNo2);

//return the address of genetic distance matrix
static Mat<double> *getGenDis ();

//create the distance matrix for any two sequences basing on maximum likelihood method
static void cmpGenDis (char *aliFileName);


static void setPtn (int siteNo, int ptnNo);
static int getPtn (int siteNo);

//--------------------------------------------------------------------
//write all information of this alignment 
template <class Tout>
static Tout &write (Tout &out) {
  out << nSeq_ << " " << nSite_ << endl;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    items_[seqNo_].write (out);
  out << "=========end of this alignment====================================" << endl;
  return out;

}

//--------------------------------------------------------------------
//write all information of this alignment 
static void write (const char *fileName) {
  std::ofstream out_;
  out_.open (fileName);
  out_ << nSeq_ << " " << nSite_ << endl;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
    out_ <<'T' << seqNo_ <<"   ";
    items_[seqNo_].writeItem (out_) << endl;
  }
  out_.close ();
}

//get the name of the sequence at certain ndNo
static Vec<char> &getName (int ndNo);
//get the name of the sequence at certain ndNo
static void getName (int ndNo, Vec<char> &name);

//get the order of the sequence whose name is input name
static int getSeqNo (Vec<char> &name);

//set the memory for this class
static void setLimit (int nSeq, int nSite);

//get the base frequency of this alignment
static void estimateStateFrq (DVec20 &stateFrqArr);

static double estimateProb (BASE baseX, BASE baseY);
static void estimateGenPam (double &tsAG, double &tsCT, double &tvAC, double &tvAT, double &tvCG, double &tvGT);
static double estimateTsTvRatio ();

static void getApp (int stateNo, DVec20 &appArr);

static int checkSameName ();

//release all memmory for this class
void release ();


//destruction function to release data
~Ali ();

/**************************************************************************
**************************************************************************
**************************************************************************
**************************************************************************/
private :
//the number of sequence of this alignment, we do not use it, instead of using nActSeq
static int nSeq_;
static int longestSeqNameLen_;


static int ptnArr_[MAX_NUM_SITE];

//the number of sites of this alignment
static int nSite_;

//the alignment information
static Vec<Seq> items_;

//disMat_[i][j] is the distance between sequence i and sequence j
static Mat<double> genDisMat_;

//check whether or not the genDismat of this class was cmped
static int isGenDisMatCmped_;
}; //end class ali
#endif
