#include "aminom.h"
#include "constant.h"
#include "dayhoff.h"
#include "jttm.h"
#include "vt.h"
#include "mtrev24.h"
#include "blosum62.h"
#include "wag.h"
#include "ali.h"
#include "model.h"

AminoM::AminoM () {
}

/* ***************************************************************************/
/* ***************************************************************************/
/*cmp the array of probability of changing from
one amino acid into one amino acid after a period of brLen / subRate_
*/
double AminoM::cmpProbChange (const int stateNo1, const int stateNo2, const double brLen) {
/* compute P(t) */
  double time_ = brLen * 100.0;
  double probChange_ = 0.0;
  for (int stateNo_ = 0; stateNo_ < NUM_AMINO; stateNo_ ++) 
    probChange_ += exp (time_ * eigValArr_[stateNo_]) * eigVecMat_[stateNo1][stateNo_] * invEigVecMat_[stateNo_][stateNo2];
   
  return probChange_;
 //  writeMat (probMat);
}

/* ***************************************************************************/
/* ***************************************************************************/
/*cmp the array of probability of changing from
one amino acid into one amino acid after a period of brLen / subRate_
*/
void AminoM::cmpProbChange (const int stateNo1, const double brLen, DVec20 &probVec) {
/* compute P(t) */
   for (int stateNo2_ = 0; stateNo2_ < NUM_AMINO; stateNo2_ ++)
      probVec[stateNo2_] = cmpProbChange (stateNo1, stateNo2_, brLen);  

 //  writeMat (probMat);
}

/*
cmp the array of probability of changing from
one amino acid into one amino acid after a period of brLen / subRate_
*/
void AminoM::cmpProbChange (const double brLen, DMat20 &probMat) {
/* compute P(t) */

//  double time_ = brLen / COF_BR;
  double time_ = brLen * 100.0;
  int count0_, count1_;
  DMat20 lMat_;
  for (count0_ = 0; count0_ < NUM_AMINO; count0_ ++)
    for (count1_ = 0; count1_ < NUM_AMINO; count1_ ++)
      lMat_[count0_][count1_] = eigVecMat_[count0_][count1_] * exp (time_ * eigValArr_[count1_]);

  for (count0_ = 0; count0_ < NUM_AMINO; count0_ ++)
    for (count1_ = 0; count1_ < NUM_AMINO; count1_ ++) {
      probMat[count0_][count1_] = 0.0;
      for (int count2_ = 0; count2_ < NUM_AMINO; count2_ ++) 
        probMat[count0_][count1_] += lMat_[count0_][count2_] * invEigVecMat_[count2_][count1_];
    }
//  writeMat (probMat);
}

void AminoM::init () {
  switch (modelType_) {
    case JTT:
      JTTM jttM_;
      jttM_.getData (unitRateMat_, stateFrqArr_);
      break;          
    
    case Dayhoff:
      DayhoffM dayhoffM_;
      dayhoffM_.getData (unitRateMat_, stateFrqArr_);
      break;

    case VT:
      VTM vtM_;
      vtM_.getData (unitRateMat_, stateFrqArr_);
      break;

    case MtREV24:
      MtREV24M mtREV24M_;
      mtREV24M_.getData (unitRateMat_, stateFrqArr_);
      break;

    case Blosum62:
      Blosum62M blosum62M_;
      blosum62M_.getData (unitRateMat_, stateFrqArr_);
      break;

    default:
      WAGM wagM_;
      wagM_.getData (unitRateMat_, stateFrqArr_);
  }

  for (int count0_ = 0; count0_ < NUM_AMINO; count0_ ++) {
     for (int count1_ = count0_ + 1; count1_ < NUM_AMINO; count1_ ++)
        unitRateMat_[count1_][count0_] = unitRateMat_[count0_][count1_];
     unitRateMat_[count0_][count0_] = 0.0;
  }
  if (Model::getBaseFrqType () == ESTIMATE) {
     std::cout << "Estimating the amino acid frequecies from the data set..." << endl;
     Ali::estimateStateFrq (stateFrqArr_);
  }
  convfreq (stateFrqArr_);
  tranprobmat ();
}

/* make all frequencies a little different */
void AminoM::convfreq(DVec20 &freqemp)
{
    int i, j, maxi=0;
    double freq, maxfreq, sum;

    sum = 0.0;
    maxfreq = 0.0;
    for (i = 0; i < NUM_AMINO; i++) {
        freq = freqemp[i];
        if (freq < MIN_FRQ) freqemp[i] = MIN_FRQ;
        if (freq > maxfreq) {
            maxfreq = freq;
            maxi = i;
        }

        sum += freqemp[i];
    }
    freqemp[maxi] += 1.0 - sum;
    
    for (i = 0; i < NUM_AMINO - 1; i++) {
        for (j = i + 1; j < NUM_AMINO; j++) {
            if (freqemp[i] == freqemp[j]) {
                freqemp[i] += MIN_FRQ_DIF/2.0;
                freqemp[j] -= MIN_FRQ_DIF/2.0;
            }
        }
    }
} /* convfreq */


AminoM::~AminoM () {
}



/******************************************************************************/
/* computation of Pij(t)                                                      */
/******************************************************************************/


/***************************** internal functions *****************************/


void AminoM::elmhes(DMat20 a, Vec<int> &ordr, int n)
{
    int m, j, i;
    double y, x;


    for (i = 0; i < n; i++)
        ordr[i] = 0;
    for (m = 2; m < n; m++) {
        x = 0.0;
        i = m;
        for (j = m; j <= n; j++) {
            if (fabs(a[j - 1][m - 2]) > fabs(x)) {

                x = a[j - 1][m - 2];
                i = j;
            }
        }
        ordr[m - 1] = i;      /* vector */

        if (i != m) {
            for (j = m - 2; j < n; j++) {
                y = a[i - 1][j];
                a[i - 1][j] = a[m - 1][j];
                a[m - 1][j] = y;
            }
            for (j = 0; j < n; j++) {
                y = a[j][i - 1];
                a[j][i - 1] = a[j][m - 1];
                a[j][m - 1] = y;
            }
        }
        if (x != 0.0) {
            for (i = m; i < n; i++) {
                y = a[i][m - 2];
                if (y != 0.0) {
                    y /= x;
                    a[i][m - 2] = y;
                    for (j = m - 1; j < n; j++)
                        a[i][j] -= y * a[m - 1][j];
                    for (j = 0; j < n; j++)
                        a[j][m - 1] += y * a[j][i];
                }
            }
        }
    }
} /* elmhes */


void AminoM::eltran(DMat20 a, DMat20 zz, Vec<int> &ordr, int n)
{
    int i, j, m;


    for (i = 0; i < n; i++) {
        for (j = i + 1; j < n; j++) {
            zz[i][j] = 0.0;
            zz[j][i] = 0.0;
        }
        zz[i][i] = 1.0;
    }
    if (n <= 2)
        return;
    for (m = n - 1; m >= 2; m--) {
        for (i = m; i < n; i++)
            zz[i][m - 1] = a[i][m - 2];
        i = ordr[m - 1];
        if (i != m) {
            for (j = m - 1; j < n; j++) {
                zz[m - 1][j] = zz[i - 1][j];
                zz[i - 1][j] = 0.0;
            }
            zz[i - 1][m - 1] = 1.0;
        }
    }
} /* eltran */


void AminoM::mcdiv(double ar, double ai, double br, double bi,
    double *cr, double *ci)
{
    double s, ars, ais, brs, bis;


    s = fabs(br) + fabs(bi);
    ars = ar / s;
    ais = ai / s;
    brs = br / s;
    bis = bi / s;
    s = brs * brs + bis * bis;
    *cr = (ars * brs + ais * bis) / s;
    *ci = (ais * brs - ars * bis) / s;
} /* mcdiv */


void AminoM::hqr2(int n, int low, int hgh, DMat20 h,
    DMat20 zz, DVec20 &wr, DVec20 &wi)
{
    int i, j, k, l=0, m, en, na, itn, its;
    double p=0, q=0, r=0, s=0, t, w, x=0, y, ra, sa, vi, vr, z=0, norm, tst1, tst2;
    int notlas; /* boolean */


    norm = 0.0;
    k = 1;
    /* store isolated roots and compute matrix norm */
    for (i = 0; i < n; i++) {
        for (j = k - 1; j < n; j++)
            norm += fabs(h[i][j]);



        k = i + 1;
        if (i + 1 < low || i + 1 > hgh) {
            wr[i] = h[i][i];
            wi[i] = 0.0;
        }
    }
    en = hgh;
    t = 0.0;
    itn = n * 30;
    while (en >= low) {    /* search for next eigenvalues */
        its = 0;
        na = en - 1;
        while (en >= 1) {
            /* look for single small sub-diagonal element */
            for (l = en; l > low; l--) {
                s = fabs(h[l - 2][l - 2]) + fabs(h[l - 1][l - 1]);

                if (s == 0.0)
                    s = norm;
                tst1 = s;
                tst2 = tst1 + fabs(h[l - 1][l - 2]);
                if (tst2 == tst1)
                    goto L100;
            }
            l = low;
    L100:
            x = h[en - 1][en - 1];    /* form shift */
            if (l == en || l == na)
                break;
            if (itn == 0) {
                /* all eigenvalues have not converged */
                std::cout << "\n\n\nHALT: PLEASE REPORT ERROR B TO DEVELOPERS\n\n\n";
                exit(1);
            }
            y = h[na - 1][na - 1];
            w = h[en - 1][na - 1] * h[na - 1][en - 1];
            /* form exceptional shift */
            if (its == 10 || its == 20) {
                t += x;
                for (i = low - 1; i < en; i++)
                    h[i][i] -= x;
                s = fabs(h[en - 1][na - 1]) + fabs(h[na - 1][en - 3]);
                x = 0.75 * s;
                y = x;
                w = -0.4375 * s * s;
            }
            its++;
            itn--;
            /* look for two consecutive small sub-diagonal elements */
            for (m = en - 2; m >= l; m--) {
                z = h[m - 1][m - 1];
                r = x - z;
                s = y - z;
                p = (r * s - w) / h[m][m - 1] + h[m - 1][m];
                q = h[m][m] - z - r - s;
                r = h[m + 1][m];
                s = fabs(p) + fabs(q) + fabs(r);
                p /= s;
                q /= s;
                r /= s;
                if (m == l)
                    break;
                tst1 = fabs(p) *
                        (fabs(h[m - 2][m - 2]) + fabs(z) + fabs(h[m][m]));
                tst2 = tst1 + fabs(h[m - 1][m - 2]) * (fabs(q) + fabs(r));
                if (tst2 == tst1)
                    break;
            }
            for (i = m + 2; i <= en; i++) {
                h[i - 1][i - 3] = 0.0;
                if (i != m + 2)
                    h[i - 1][i - 4] = 0.0;
            }
            for (k = m; k <= na; k++) {
                notlas = (k != na);

                if (k != m) {
                    p = h[k - 1][k - 2];
                    q = h[k][k - 2];
                    r = 0.0;
                    if (notlas)
                        r = h[k + 1][k - 2];
                    x = fabs(p) + fabs(q) + fabs(r);
                    if (x != 0.0) {
                        p /= x;
                        q /= x;
                        r /= x;
                    }
                }
                if (x != 0.0) {
                    if (p < 0.0) /* sign */
                        s = - sqrt(p * p + q * q + r * r);
                    else
                        s = sqrt(p * p + q * q + r * r);
                    if (k != m)
                        h[k - 1][k - 2] = -s * x;
                    else {
                        if (l != m)
                            h[k - 1][k - 2] = -h[k - 1][k - 2];
                    }
                    p += s;
                    x = p / s;
                    y = q / s;
                    z = r / s;
                    q /= p;
                    r /= p;
                    if (!notlas) {
                        for (j = k - 1; j < n; j++) {    /* row modification */
                            p = h[k - 1][j] + q * h[k][j];
                            h[k - 1][j] -= p * x;
                            h[k][j] -= p * y;
                        }
                        j = (en < (k + 3)) ? en : (k + 3); /* min */
                        for (i = 0; i < j; i++) {    /* column modification */
                            p = x * h[i][k - 1] + y * h[i][k];
                            h[i][k - 1] -= p;
                            h[i][k] -= p * q;
                        }
                        /* accumulate transformations */
                        for (i = low - 1; i < hgh; i++) {
                            p = x * zz[i][k - 1] + y * zz[i][k];
                            zz[i][k - 1] -= p;
                            zz[i][k] -= p * q;
                        }
                    } else {
                        for (j = k - 1; j < n; j++) {    /* row modification */
                            p = h[k - 1][j] + q * h[k][j] + r * h[k + 1][j];
                            h[k - 1][j] -= p * x;
                            h[k][j] -= p * y;
                            h[k + 1][j] -= p * z;
                        }
                        j = (en < (k + 3)) ? en : (k + 3); /* min */
                        for (i = 0; i < j; i++) {    /* column modification */
                            p = x * h[i][k - 1] + y * h[i][k] + z * h[i][k + 1];
                            h[i][k - 1] -= p;
                            h[i][k] -= p * q;
                            h[i][k + 1] -= p * r;
                        }
                        /* accumulate transformations */
                        for (i = low - 1; i < hgh; i++) {
                            p = x * zz[i][k - 1] + y * zz[i][k] +
                                z * zz[i][k + 1];
                            zz[i][k - 1] -= p;
                            zz[i][k] -= p * q;
                            zz[i][k + 1] -= p * r;
                        }
                    }
                }
            }           /* for k */
        }               /* while infinite loop */
        if (l == en) {           /* one root found */
            h[en - 1][en - 1] = x + t;
            wr[en - 1] = h[en - 1][en - 1];
            wi[en - 1] = 0.0;
            en = na;
            continue;
        }
        y = h[na - 1][na - 1];

        w = h[en - 1][na - 1] * h[na - 1][en - 1];
        p = (y - x) / 2.0;
        q = p * p + w;
        z = sqrt(fabs(q));
        h[en - 1][en - 1] = x + t;
        x = h[en - 1][en - 1];
        h[na - 1][na - 1] = y + t;
        if (q >= 0.0) {           /* real pair */
            if (p < 0.0) /* sign */
                z = p - fabs(z);
            else
                z = p + fabs(z);
            wr[na - 1] = x + z;
            wr[en - 1] = wr[na - 1];
            if (z != 0.0)
                wr[en - 1] = x - w / z;
            wi[na - 1] = 0.0;
            wi[en - 1] = 0.0;
            x = h[en - 1][na - 1];
            s = fabs(x) + fabs(z);
            p = x / s;
            q = z / s;
            r = sqrt(p * p + q * q);
            p /= r;
            q /= r;
            for (j = na - 1; j < n; j++) {    /* row modification */
                z = h[na - 1][j];
                h[na - 1][j] = q * z + p * h[en - 1][j];
                h[en - 1][j] = q * h[en - 1][j] - p * z;
            }
            for (i = 0; i < en; i++) {    /* column modification */
                z = h[i][na - 1];
                h[i][na - 1] = q * z + p * h[i][en - 1];
                h[i][en - 1] = q * h[i][en - 1] - p * z;
            }
            /* accumulate transformations */
            for (i = low - 1; i < hgh; i++) {
                z = zz[i][na - 1];
                zz[i][na - 1] = q * z + p * zz[i][en - 1];
                zz[i][en - 1] = q * zz[i][en - 1] - p * z;
            }
        } else {           /* complex pair */
            wr[na - 1] = x + p;
            wr[en - 1] = x + p;
            wi[na - 1] = z;
            wi[en - 1] = -z;
        }
        en -= 2;
    }                   /* while en >= low */
    /* backsubstitute to find vectors of upper triangular form */

    if (norm != 0.0) {
        for (en = n; en >= 1; en--) {
            p = wr[en - 1];
            q = wi[en - 1];
            na = en - 1;
            if (q == 0.0) {/* real vector */
                m = en;
                h[en - 1][en - 1] = 1.0;
                if (na != 0) {
                    for (i = en - 2; i >= 0; i--) {
                        w = h[i][i] - p;
                        r = 0.0;
                        for (j = m - 1; j < en; j++)
                            r += h[i][j] * h[j][en - 1];
                        if (wi[i] < 0.0) {
                            z = w;
                            s = r;
                        } else {
                            m = i + 1;
                            if (wi[i] == 0.0) {
                                t = w;
                                if (t == 0.0) {
                                    tst1 = norm;
                                    t = tst1;
                                    do {
                                        t = 0.01 * t;
                                        tst2 = norm + t;
                                    } while (tst2 > tst1);
                                }
                                h[i][en - 1] = -(r / t);
                            } else {    /* solve real equations */
                                x = h[i][i + 1];
                                y = h[i + 1][i];
                                q = (wr[i] - p) * (wr[i] - p) + wi[i] * wi[i];
                                t = (x * s - z * r) / q;
                                h[i][en - 1] = t;
                                if (fabs(x) > fabs(z))
                                    h[i + 1][en - 1] = (-r - w * t) / x;
                                else
                                    h[i + 1][en - 1] = (-s - y * t) / z;
                            }
                            /* overflow control */
                            t = fabs(h[i][en - 1]);
                            if (t != 0.0) {
                                tst1 = t;
                                tst2 = tst1 + 1.0 / tst1;
                                if (tst2 <= tst1) {
                                    for (j = i; j < en; j++)
                                        h[j][en - 1] /= t;
                                }
                            }
                        }
                    }
                }
            } else if (q > 0.0) {
                m = na;
                if (fabs(h[en - 1][na - 1]) > fabs(h[na - 1][en - 1])) {
                    h[na - 1][na - 1] = q / h[en - 1][na - 1];
                    h[na - 1][en - 1] = (p - h[en - 1][en - 1]) /
                                                        h[en - 1][na - 1];
                } else
                    mcdiv(0.0, -h[na - 1][en - 1], h[na - 1][na - 1] - p, q,

                        &h[na - 1][na - 1], &h[na - 1][en - 1]);
                h[en - 1][na - 1] = 0.0;
                h[en - 1][en - 1] = 1.0;
                if (en != 2) {
                    for (i = en - 3; i >= 0; i--) {
                        w = h[i][i] - p;
                        ra = 0.0;
                        sa = 0.0;
                        for (j = m - 1; j < en; j++) {
                            ra += h[i][j] * h[j][na - 1];
                            sa += h[i][j] * h[j][en - 1];
                        }
                        if (wi[i] < 0.0) {
                            z = w;
                            r = ra;
                            s = sa;
                        } else {
                            m = i + 1;
                            if (wi[i] == 0.0)
                                mcdiv(-ra, -sa, w, q, &h[i][na - 1],
                                    &h[i][en - 1]);
                            else {    /* solve complex equations */
                                x = h[i][i + 1];
                                y = h[i + 1][i];
                                vr = (wr[i] - p) * (wr[i] - p);
                                vr = vr + wi[i] * wi[i] - q * q;    
                                vi = (wr[i] - p) * 2.0 * q;
                                if (vr == 0.0 && vi == 0.0) {
                                    tst1 = norm * (fabs(w) + fabs(q) + fabs(x) +
                                                   fabs(y) + fabs(z));
                                    vr = tst1;
                                    do {
                                        vr = 0.01 * vr;
                                        tst2 = tst1 + vr;
                                    } while (tst2 > tst1);

                                }
                                mcdiv(x * r - z * ra + q * sa,
                                     x * s - z * sa - q * ra, vr, vi,
                                     &h[i][na - 1], &h[i][en - 1]);
                                if (fabs(x) > fabs(z) + fabs(q)) {
                                    h[i + 1]
                                        [na - 1] = (q * h[i][en - 1] -
                                                    w * h[i][na - 1] - ra) / x;
                                    h[i + 1][en - 1] = (-sa - w * h[i][en - 1] -
                                                        q * h[i][na - 1]) / x;
                                } else
                                    mcdiv(-r - y * h[i][na - 1],
                                         -s - y * h[i][en - 1], z, q,
                                         &h[i + 1][na - 1], &h[i + 1][en - 1]);
                            }
                            /* overflow control */
                            t = (fabs(h[i][na - 1]) > fabs(h[i][en - 1])) ?
                                 fabs(h[i][na - 1]) : fabs(h[i][en - 1]);
                            if (t != 0.0) {
                                tst1 = t;
                                tst2 = tst1 + 1.0 / tst1;
                                if (tst2 <= tst1) {
                                    for (j = i; j < en; j++) {
                                        h[j][na - 1] /= t;
                                        h[j][en - 1] /= t;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        /* end back substitution. vectors of isolated roots */
        for (i = 0; i < n; i++) {
            if (i + 1 < low || i + 1 > hgh) {
                for (j = i; j < n; j++)
                    zz[i][j] = h[i][j];
            }
        }
        /* multiply by transformation matrix to give vectors of
         * original full matrix. */
        for (j = n - 1; j >= low - 1; j--) {
            m = ((j + 1) < hgh) ? (j + 1) : hgh; /* min */
            for (i = low - 1; i < hgh; i++) {
                z = 0.0;
                for (k = low - 1; k < m; k++)
                    z += zz[i][k] * h[k][j];
                zz[i][j] = z;
            }
        }
    }
    return;
} /* hqr2 */


/* make rate matrix with 0.01 expected substitutions per unit time */
void AminoM::onepamratematrix(DMat20 a)
{
    int i, j;
    double delta, temp, sum;
    DVec20 m;

    for (i = 0; i < NUM_AMINO; i++)
    {
        for (j = 0; j < NUM_AMINO; j++)
        {
            a[i][j] = stateFrqArr_[j]*a[i][j];
        }
    }

    for (i = 0, sum = 0.0; i < NUM_AMINO; i++)
    {
        for (j = 0, temp = 0.0; j < NUM_AMINO; j++)
            temp += a[i][j];
        m[i] = temp; /* row sum */
        sum += temp*stateFrqArr_[i]; /* exp. rate */
    }

    delta = 0.01 / sum; /* 0.01 subst. per unit time */
    for (i = 0; i < NUM_AMINO; i++) {
        for (j = 0; j < NUM_AMINO; j++) {
            if (i != j)
                a[i][j] = delta * a[i][j];
            else
                a[i][j] = delta * (-m[i]);
        }
    }
} /* onepamratematrix */



void AminoM::rtfdata (DMat20 &a, DVec20 &forg) {
  for (int count0_ = 0; count0_ < NUM_AMINO; count0_ ++) {
    forg[count0_] = stateFrqArr_[count0_];
    for (int count1_ = 0; count1_ < NUM_AMINO; count1_ ++)
      a[count0_][count1_] = unitRateMat_[count0_][count1_];
  }
}

void AminoM::writeMat (DMat20 &rate) {
  for (int count0_ = 0; count0_ < NUM_AMINO; count0_ ++) {
    for (int count1_ = 0; count1_ < NUM_AMINO; count1_ ++) {      
      std::cout.width (20);
      std::cout.precision (5);
      std::cout << rate[count0_][count1_];
    }
    std::cout << endl;
  }
  std::cout << endl;
}

void AminoM::writeInt (Vec<int> &vec) {
  std::cout << endl;
  for (int count_ = 0; count_ < NUM_AMINO; count_ ++) {
    std::cout.width (10);
    std::cout << vec[count_];
  }
  std::cout << endl;
}


void AminoM::writeDouble (DVec20 &vec) {
  std::cout << endl;
  for (int count_ = 0; count_ < NUM_AMINO; count_ ++) {
    std::cout.width (10);
    std::cout << vec[count_];
  }
  std::cout << endl;
}

void AminoM::eigensystem(DVec20 &eval, DMat20 &evec)
{
    DVec20 evali, forg;
    DMat20 a, b;
    Vec<int> ordr (NUM_AMINO + 1, NUM_AMINO + 1);
    int i, j, k, error;
    double zero;
    rtfdata(a, forg); /* get relative transition matrix and frequencies */
//    write (a, forg);    
    onepamratematrix(a); /* make 1 PAM rate matrix */
    
//    writeMat (a);    
//    writeDouble (stateFrqArr_);
    
    /* copy a to b */
    for (i = 0; i < NUM_AMINO; i++)
        for (j = 0; j < NUM_AMINO; j++)
            b[i][j] = a[i][j];

    elmhes(a, ordr, NUM_AMINO); /* compute eigenvalues and eigenvectors */
//    writeInt (ordr);

    eltran(a, evec, ordr, NUM_AMINO);

  //  writeMat (evec);    

    hqr2(NUM_AMINO, 1, NUM_AMINO, a, evec, eval, evali);
    
    //writeMat (evec);    
    //writeDouble (eval);
    //writeDouble (evali);
    

    /* check eigenvalue equation */
    error = 0;
    for (j = 0; j < NUM_AMINO; j++) {
        for (i = 0, zero = 0.0; i < NUM_AMINO; i++) {
            for (k = 0; k < NUM_AMINO; k++) zero += b[i][k] * evec[k][j];
            zero -= eval[j] * evec[i][j];
            if (fabs(zero) > 1.0e-5)
                error = 1;    
        }
    }
    if (error)
        std::cout << "\nWARNING: Eigensystem doesn't satisfy eigenvalue equation!\n";


} /* eigensystem */


void AminoM::luinverse(DMat20 inmat, DMat20 imtrx, int size)
{
    double eps = 1.0e-20; /* ! */
    int i, j, k, l, maxi=0, idx, ix, jx;
    double sum, tmp, maxb, aw;
    Vec<int> index (NUM_AMINO, NUM_AMINO);
    double *wk;
    DMat20 omtrx;

    
    
    /* copy inmat to omtrx */
    for (i = 0; i < NUM_AMINO; i++)
        for (j = 0; j < NUM_AMINO; j++)
            omtrx[i][j] = inmat[i][j];
    
    wk = (double *) calloc((size_t)size, sizeof(double));
    aw = 1.0;
    for (i = 0; i < size; i++) {
        maxb = 0.0;
        for (j = 0; j < size; j++) {
            if (fabs(omtrx[i][j]) > maxb)
                maxb = fabs(omtrx[i][j]);
        }
        if (maxb == 0.0) {
            /* Singular matrix */
            std::cout << "\n\n\nHALT: PLEASE REPORT ERROR C TO DEVELOPERS\n\n\n";
            exit(1);
        }
        wk[i] = 1.0 / maxb;
    }
    for (j = 0; j < size; j++) {
        for (i = 0; i < j; i++) {
            sum = omtrx[i][j];
            for (k = 0; k < i; k++)
                sum -= omtrx[i][k] * omtrx[k][j];
            omtrx[i][j] = sum;
        }
        maxb = 0.0;
        for (i = j; i < size; i++) {
            sum = omtrx[i][j];
            for (k = 0; k < j; k++)
                sum -= omtrx[i][k] * omtrx[k][j];
            omtrx[i][j] = sum;
            tmp = wk[i] * fabs(sum);
            if (tmp >= maxb) {
                maxb = tmp;
                maxi = i;
            }
        }
        if (j != maxi) {
            for (k = 0; k < size; k++) {
                tmp = omtrx[maxi][k];
                omtrx[maxi][k] = omtrx[j][k];
                omtrx[j][k] = tmp;
            }
            aw = -aw;
            wk[maxi] = wk[j];
        }
        index[j] = maxi;
        if (omtrx[j][j] == 0.0)
            omtrx[j][j] = eps;
        if (j != size - 1) {
            tmp = 1.0 / omtrx[j][j];
            for (i = j + 1; i < size; i++)
                omtrx[i][j] *= tmp;
        }
    }
    for (jx = 0; jx < size; jx++) {
        for (ix = 0; ix < size; ix++)
            wk[ix] = 0.0;
        wk[jx] = 1.0;
        l = -1;
        for (i = 0; i < size; i++) {
            idx = index[i];
            sum = wk[idx];
            wk[idx] = wk[i];
            if (l != -1) {
                for (j = l; j < i; j++)
                    sum -= omtrx[i][j] * wk[j];
            } else if (sum != 0.0)
                l = i;
            wk[i] = sum;
        }
        for (i = size - 1; i >= 0; i--) {
            sum = wk[i];
            for (j = i + 1; j < size; j++)
                sum -= omtrx[i][j] * wk[j];
            wk[i] = sum / omtrx[i][i];
        }
        for (ix = 0; ix < size; ix++)
            imtrx[ix][jx] = wk[ix];
    }

    free((char *)wk);
    wk = NULL;
} /* luinverse */


void AminoM::checkevector(DMat20 evec, DMat20 ivec, int nn)
{
    int i, j, ia, ib, ic, error;
    DMat20 matx;
    double sum;


    /* multiply matrix of eigenvectors and its inverse */
    for (ia = 0; ia < nn; ia++) {
        for (ic = 0; ic < nn; ic++) {
            sum = 0.0;
            for (ib = 0; ib < nn; ib++) sum += evec[ia][ib] * ivec[ib][ic];
            matx[ia][ic] = sum;
        }
    }
    /* check whether the unitary matrix is obtained */
    error = 0;
    for (i = 0; i < nn; i++) {
        for (j = 0; j < nn; j++) {
            if (i == j) {
                if (fabs(matx[i][j] - 1.0) > 1.0e-5)
                    error = 1;
            } else {
                if (fabs(matx[i][j]) > 1.0e-5)
                    error = 1;
            }
        }
    }
    if (error) {
        std::cout << "\nWARNING: Inversion of eigenvector matrix not perfect!\n";
    }
} /* checkevector */


/***************************** exported functions *****************************/


/* compute 1 PAM rate matrix, its eigensystem, and the inverse matrix thereof */
void AminoM::tranprobmat()
{
//   writeDouble (stateFrqArr_);
   eigensystem(eigValArr_, eigVecMat_); /* eigensystem of 1 PAM rate matrix */
   luinverse(eigVecMat_, invEigVecMat_, NUM_AMINO); /* inverse eigenvectors are in Ievc */
   checkevector(eigVecMat_, invEigVecMat_, NUM_AMINO); /* check whether inversion was OK */
} /* tranprobmat */
