/***************************************************************************
                          model.cpp  -  description
                             -------------------
    begin                : Wed Mar 5 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <iostream>

#include "basem.h"
#include "mat.h"
#include "constant.h"
#include "ali.h"
#include "opturtree.h"
#include "brent.h"
#include "vec.h"

using namespace std;

DVec20 BaseM::stateFrqArr_;

DMat20 BaseM::unitRateMat_;
DMat20 BaseM::eigVecMat_;
DMat20 BaseM::invEigVecMat_;
DVec20 BaseM::eigValArr_;

double BaseM::tsTvRatio_;
double BaseM::pyPuRatio_;

double BaseM::tsAG_;
double BaseM::tsCT_;

double BaseM::tvAC_;
double BaseM::tvAT_;
double BaseM::tvCG_;
double BaseM::tvGT_;

MODEL_PAM_TYPE BaseM::isActPam_;

int    BaseM::nState_ = 0;

MODEL BaseM::modelType_;

//the substitution rate of change per unit time
double BaseM::subRate_;


int BaseM::isDNA () {
  return (modelType_ ==HKY85 || modelType_ == GTR);
}

void BaseM::setModel (MODEL modelType) {
  modelType_ = modelType;
  if (isDNA () == 1)
    nState_ = NUM_BASE;
  else
    nState_ = NUM_AMINO;
}

//--------------------------------------------------------------------
//the constructor function
BaseM::BaseM () {
  tsTvRatio_ = -1.0;

  tsAG_ = 1.0 / 6.0;
  tsCT_ = 1.0 / 6.0;
  tvAC_ = 1.0 / 6.0;
  tvAT_ = 1.0 / 6.0;
  tvCG_ = 1.0 / 6.0;
  tvGT_ = 1.0 / 6.0;
}

//--------------------------------------------------------------------
//compute the substitution rate per unit brLen for this model
double BaseM::cmpSubRate () {
  return -1.0;
}

//--------------------------------------------------------------------
//set the tsTvRatio for this model
void BaseM::setTsTvRatio (const double tsTvRatio) {
  tsTvRatio_ = tsTvRatio;
} 

//--------------------------------------------------------------------
//set the tsTvRatio for this model
void BaseM::setPyPuRatio (const double pyPuRatio) {
  pyPuRatio_ = pyPuRatio;
} 

void BaseM::setGenPam (const double tsAG, const double tsCT, 
                                      const double tvAC, const double tvAT, const double tvCG, const double tvGT) {
   tsAG_ = tsAG;
   tsCT_ = tsCT;

   tvAC_ = tvAC;
   tvAT_ = tvAT;
   tvCG_ = tvCG;
   tvGT_ = tvGT;

   init ();
}

void BaseM::setStateFrq (DVec20 &stateFrqArr) {
  for (int count_ = 0; count_ < nState_; count_ ++)
    stateFrqArr_[count_] = stateFrqArr[count_];

  if (stateFrqArr_[0] < -ZERO) 
    Ali::estimateStateFrq (stateFrqArr_);
}

//--------------------------------------------------------------------
//compute the state frequencies for this alignment
void BaseM::cmpStateFrq () {
}

//--------------------------------------------------------------------
//compute the substitution rate
double cmpSubRate () {
  return -1.0;
}

//--------------------------------------------------------------------
//all things are inited here
void BaseM::init () {
}

//--------------------------------------------------------------------
//compute the state frequencies for this alignment
void BaseM::getStateFrq (DVec20 &stateFrqArr) {  
  for (int count_ = 0; count_ < nState_; count_ ++)
    stateFrqArr [count_] = stateFrqArr_[count_];
}

//--------------------------------------------------------------------
//get the tsTvratio
double BaseM::getTsTvRatio () {
  return tsTvRatio_;
}

//--------------------------------------------------------------------
//get the pyPuratio
double BaseM::getPyPuRatio () {
  return pyPuRatio_;
}

//================================================================
void BaseM::getGenPam (double &tsAG, double & tsCT,
                                     double & tvAC, double & tvAT, double & tvCG, double & tvGT) {
  
   tsAG  = tsAG_;
   tsCT  = tsCT_;
   tvAC  = tvAC_;
   tvAT  = tvAT_;
   tvCG  = tvCG_;
   tvGT  = tvGT_;
}


//--------------------------------------------------------------------
/*optimize all parameter evolving to sequence evolution
1. tsTvRatio
2. TsRatio
3. find state frequencies
*/
//optimize all parameters of this model

//--------------------------------------------------------------------
//compute the probability mat
void BaseM::cmpProbChange (const double time, DMat20 &probMat) {
  std::cout << time << ": " <<  probMat[0][0] << endl;
}

//--------------------------------------------------------------------
//compute the probability vec
void BaseM::cmpProbChange (const int stateNo, const double time, DVec20 &probVec) {
  std::cout << stateNo << ":" << time << ": " <<  probVec[0] << endl;
}

//--------------------------------------------------------------------
//compute the probability of changing from stateNo1 into stateNo2 after a period time of time
double BaseM::cmpProbChange (const int stateNo1, const int stateNo2, const double time) {
  std::cout << "(" << stateNo1 << "," << stateNo2 << "): " << time << endl;
  return -1.0;
}

void BaseM::optPam (const int maxNIter, const PAM_TYPE tsTvRatioType, const PAM_TYPE pyPuRatioType) {
  std::cout << maxNIter << ": " << tsTvRatioType << ": " << pyPuRatioType << endl;
}

//--------------------------------------------------------------------
//optimize the paramters using Brent method
void BaseM::optPam (const int maxNIter) {
  std::cout << maxNIter << endl;
}

//--------------------------------------------------------------------
//optimize the paramters using Brent method
void BaseM::optPam () {
}

//--------------------------------------------------------------------
//release all memmory for this class
void BaseM::release () {
}

//--------------------------------------------------------------------
//the destructor function of this class
BaseM::~BaseM () {
  release ();
//  std::cout << "this is the destructor of state model class " << endl;
}
