/***************************************************************************
                          statem.h  -  description
                             -------------------
    begin                : Wed Mar 5 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#ifndef BASEM_H
#define BASEM_H

#include "constant.h"
#include "mat.h"
#include "vec.h"
#include "dvec20.h"
#include "dmat20.h"

//this is the state model for all current models, all current models inherit from stateM
class BaseM
{
public: 
//constructor function
BaseM();

void setModel (MODEL modelType);

int isDNA ();

//set the transition transvertion ratio
void setTsTvRatio (const double tsTvRatio);

//set the pyridimine/purine ratio
void setPyPuRatio (const double pyPuRatio);

//set the transition transvertion ratio
void setStateFrq (DVec20 &stateFrqArr);

void setGenPam (const double tsAG, const double tsCT, 
                          const double tvAC, const double tvAT, const double tvCG, const double tvGT);

void getGenPam (double &tsAG,  double &tsCT, 
                          double &tvAC,  double &tvAT, double &tvCG, double &tvGT);

//get the state frequencies of this alignment
void getStateFrq (DVec20 &stateFrqArr);

//get the tsTvratio
virtual double getTsTvRatio ();
//get the pyPuratio
virtual double getPyPuRatio ();

//compute the probability mat
virtual void cmpProbChange (const double time, DMat20 &probMat);


//compute the probability vec
virtual void cmpProbChange (const int stateNo, const double time, DVec20 &probVec);


//compute the probability of changing from stateNo1 into stateNo2 after a period time of time
virtual double cmpProbChange (const int stateNo1, const int stateNo2, const double time);


//all things are inited here
virtual void init ();


virtual void optPam (const int maxNIter, const PAM_TYPE tsTvRatioType, const PAM_TYPE pyPuRatioType);
virtual void optPam (const int maxNIter);
virtual void optPam ();

//release all memmory for this class
void release ();

//the destructor function
virtual ~BaseM();


/************************************************************************************
************************************************************************************
************************************************************************************
************************************************************************************
************************************************************************************/
protected :

static MODEL modelType_;

static int nState_;
static MODEL_PAM_TYPE isActPam_;

//the substitution rate of change per unit time
static double subRate_;

//the transition transvertion ratio
static double tsTvRatio_;
//the pyridimine/purine ratio
static double pyPuRatio_;

static double tsAG_;
static double tsCT_;

static double tvAC_;
static double tvAT_;
static double tvCG_;
static double tvGT_;

//the state frequency list
static DVec20 stateFrqArr_;
static double tsTvRatioError_;

//for amino acid model
static DMat20 unitRateMat_; 
static DMat20 eigVecMat_;
static DMat20 invEigVecMat_;
static DVec20 eigValArr_;


//compute the substitution rate per unit time for this model
static double cmpSubRate ();

//cmp the state frequencies for this alignment
virtual void cmpStateFrq ();

}; //end of stateM class
#endif
