#ifndef BRARR_H
#define BRARR_H

#include <iostream>
#include "vec.h"
#include "br.h"

//this class is a collect of all branches of the unrooted tree
template <class T = double>
class BrArr
{
public :

//constructor function
BrArr () {
  curNBr_ = 0;
  maxNBr_ = 0;
}

//constructor function
BrArr (const int maxNBr) {
  curNBr_ = 0;
  setLimit (maxNBr);
}

//--------------------------------------------------------------------
//set limit of this list
void setLimit (const int maxNBr) {
  maxNBr_ = maxNBr;
  items_.set (maxNBr_, maxNBr_);
  brNoLs_.setLimit (maxNBr_);
}

int findBr (int ndNo0, int ndNo1) {
  for (int brNo_ = 0; brNo_ < maxNBr_; brNo_ ++)
    if (items_[brNo_].getId () != -1) {
	  int headNdNo_ = items_[brNo_].getHeadNd ();
	  int tailNdNo_ = items_[brNo_].getTailNd ();
	  if (Utl::isEqualPair (headNdNo_, tailNdNo_, ndNo0, ndNo1) == 1)
	    return brNo_;
	}
  return -1;
}
//--------------------------------------------------------------------
//overload the operator []
Br<T> &operator[] (const int index) {
  return items_[index];
}

//--------------------------------------------------------------------
//overload the operator +=
void operator += (Br<T> &br) {
  int brNo_ = br.getId ();
  items_[brNo_] = br;
  curNBr_ ++;
  brNoLs_ += brNo_;
}

//--------------------------------------------------------------------
//overload the operator -=
void operator -= (int brNo) {
  items_[brNo].clean ();
  brNoLs_.delItem (brNo);
  curNBr_ --;
}

//--------------------------------------------------------------------
//change the data of br
void change (Br<T> &br) {
  int brNo_ = br.getId ();
  items_[brNo_] = br;
}

//--------------------------------------------------------------------
//clean all content of branch array
void clean () {
  for (int count_ = 0; count_ < maxNBr_; count_ ++)
    items_[count_].clean ();

  curNBr_ = 0;
  brNoLs_.clean ();
}

//--------------------------------------------------------------------
//get the the curretn number of the bres list
int getCurNBr () {
  return curNBr_;
}

//--------------------------------------------------------------------
//get the maximum number branches of this class
int getMaxNBr () {
  return maxNBr_;
}

//--------------------------------------------------------------------
//return the first free branch
int getNewBr () {
  for (int brNo_ = 0; brNo_ < curNBr_ + 1; brNo_ ++)
    if (items_[brNo_].getId () == -1)
      return brNo_;

  return -1;
}

//--------------------------------------------------------------------
//return items
typedef Br<T> TBr;
Vec<TBr> &getItem () {
  return items_;
}

//--------------------------------------------------------------------
//return the brNoLs_
Vec<int> &getBrNo () {
  return brNoLs_;
}

//--------------------------------------------------------------------
//this = brArr
void operator = (BrArr<T> &brArr) {
  this->items_ = brArr.getItem ();
  this->curNBr_ = brArr.getCurNBr ();
  this->maxNBr_ = brArr.getMaxNBr ();
  this->brNoLs_ = brArr.getBrNo ();
}

//--------------------------------------------------------------------
//release all memory of this class
void release () {
  int realLimit_ = items_.getLimit ();
  items_.setSize (realLimit_);
  for (int count_ = 0; count_ < realLimit_; count_ ++)
    items_[count_].release ();    
  items_.release ();

  brNoLs_.release ();

}

Vec<int> *getBrLs () {
  return &brNoLs_;
}
//--------------------------------------------------------------------
//destructor function
virtual ~BrArr () {
  release ();
//  std::cout << "this is destructor function of BrArr class " << endl;
}

/********************************************************************************
********************************************************************************
********************************************************************************
********************************************************************************/

private :
//the number of current bres
int curNBr_;

//the maximum number branches of this branch array
int maxNBr_;

//the array of all bres
Vec<TBr> items_;

//the list off all current branches no
Vec<int> brNoLs_;

}; //end of brArr class

#endif
