#include <iostream>
#include <stdlib.h>
#include "constant.h"
#include "brent.h"
#include "math.h"

using namespace std;
int Brent::isOptedPam_ = 0;

//constructor function
Brent::Brent() {
}

int Brent::isOptedPam () {
  return isOptedPam_;
}

void Brent::turnOnOptedPam () {
  isOptedPam_ = 1;
}

void Brent::turnOffOptedPam () {
  isOptedPam_ = 0;
}

/******************************************************************************/
/* minimization of a function by Brents method (Numerical Recipes)            */
/******************************************************************************/


//-------------------------------------------------------------------------------------------------------
#define ITMAX 100
#define CGOLD 0.3819660
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define ZEPS 1.0e-10
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

/* Brents method in one dimension */
double Brent::opt (double ax, double bx, double cx, double (*f)(double), double tol,
             double *foptx, double *f2optx, double fax, double fbx, double fcx)
{
  int iter;
  double a,b,d=0,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double xw,wv,vx;
  double e=0.0;

  a=(ax < cx ? ax : cx);
  b=(ax > cx ? ax : cx);
  x=bx;
  fx=fbx;
  if (fax < fcx) {
    w=ax;
    fw=fax;
    v=cx;
    fv=fcx;
  } else {
    w=cx;
    fw=fcx;
    v=ax;
    fv=fax;	
  }
  
  for (iter=1;iter<=ITMAX;iter++) {
    xm=0.5*(a+b);
    tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
    if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
      *foptx = fx;
      xw = x-w;
      wv = w-v;
      vx = v-x;
      *f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
                (v*v*xw + x*x*wv + w*w*vx);
      return x;
    }

    if (fabs(e) > tol1) {
      r=(x-w)*(fx-fv);
      q=(x-v)*(fx-fw);
      p=(x-v)*q-(x-w)*r;
      q=2.0*(q-r);
      if (q > 0.0) 
        p = -p;
      q=fabs(q);
      etemp=e;
      e=d;
      if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
        d=CGOLD*(e=(x >= xm ? a-x : b-x));
      else {
        d=p/q;
        u=x+d;
        if (u-a < tol2 || b-u < tol2)
          d=SIGN(tol1,xm-x);
      }
    } else {
      d=CGOLD*(e=(x >= xm ? a-x : b-x));
    }

    u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
    fu=(*f)(u);
    if (fu <= fx) {
      if (u >= x) 
        a=x; 
      else 
        b=x;
      
      SHFT(v,w,x,u)
      SHFT(fv,fw,fx,fu)
    } else {
      if (u < x) 
        a=u; 
      else 
        b=u;
      if (fu <= fw || w == x) {
        v=w;
        w=u;
        fv=fw;
        fw=fu;
      } else 
        if (fu <= fv || v == x || v == w) {
          v=u;
          fv=fu;
        }
      }
    }

    *foptx = fx;
    xw = x-w;
    wv = w-v;
    vx = v-x;
    *f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/(v*v*xw + x*x*wv + w*w*vx);
    
    return x;
}

#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SHFT
#undef SIGN
#undef GOLD
#undef GLIMIT
#undef TINY


//-------------------------------------------------------------------------------------------------------
/* one-dimensional minimization - as input a lower and an upper limit and a trial
value for the minimum is needed: xmin < xguess < xmax
the function and a fractional tolerance has to be specified
onedimenmin returns the optimal x value and the value of the function
and its second derivative at this point
*/
double Brent::optOneDim(double xmin, double xguess, double xmax, double (*f)(double),
                 double tol, double *fx, double *f2x)
{
  double eps, optx, ax, bx, cx, fa, fb, fc;

  /* first attempt to bracketize minimum */
  if (isOptedPam_ == 1) {
    bx = xguess;
    fb = (*f)(bx);

    eps = xguess * 0.5;
    ax = xguess - eps;
    if (ax < xmin) 
      ax = xmin;
    fa = (*f)(ax);
    if (fa < fb) {
      ax = xmin;
      fa = (*f)(ax);
    }
 
    cx = xguess + eps;
    if (cx > xmax) 
      cx = xmax;
    fc = (*f)(cx);
    if (fc < fb) {
      cx = xmax;
      fc = (*f)(cx);
    }
     double optx_ = opt(ax, bx, cx, f, tol, fx, f2x, fa, fb, fc);	
     return optx_;
  } //end of isOptedPam_

   eps = tol;

   ax = xguess - eps;
   if (ax < xmin) 
      ax = xmin;
    bx = xguess;
    cx = xguess + eps;
    if (cx > xmax) 
      cx = xmax;
    /* check if this works */
    fa = (*f)(ax);
    fb = (*f)(bx);
    fc = (*f)(cx);
  if (fa >= fb && fc >= fb)
    return bx;

  eps = xguess * 0.5;

  if (fa <= fb) {
    ax = xguess - eps;
    if (ax < xmin) 
      ax = xmin;
    fa = (*f) (ax);
  }

  if (fc <= fb) {
    cx = xguess + eps;
    if (cx > xmax) 
      cx = xmax;
    fc = (*f) (cx);
  }

  
  /* if it works use these borders else be conservative */
  if ((fa < fb) || (fc < fb)) {
    if (ax != xmin)  {
      ax = xmin;
      fa = (*f)(ax);
    }

 if (ax == xmin) {
   double fa1 = (*f)(ax + 0.5 * tol);
   if ( fa1 >  fa) {
     double fa2 = (*f)(ax + tol);
     if (fa2 > fa1) 
       return xmin;
   }
 }

 if (cx != xmax) 
   fc = (*f)(xmax);
   optx = opt(xmin, xguess, xmax, f, tol, fx, f2x, fa, fb, fc);
 } else {
   optx = opt(ax, bx, cx, f, tol, fx, f2x, fa, fb, fc);	
 }
  
  return optx; /* return optimal x */
}


//--------------------------------------------------------------------
//release the memory of this class
void Brent::release () {
}

//-------------------------------------------------------------------------------------------------------
//the destructor function
Brent::~Brent() {
  release ();
  std::cout << "the constructor of Brent class " << endl;
}
