#ifndef BRENT_H
#define BRENT_H

//this class includes some methods for optimizing function
class Brent  
{
public:


//the constructor function
Brent ();

/* one-dimensional minimization - as input a lower and an upper limit and a trial
value for the minimum is needed: xmin < xguess < xmax
the function and a fractional tolerance has to be specified
onedimenmin returns the optimal x value and the value of the function
and its second derivative at this point
*/

static double optOneDim (double xmin, double xguess, double xmax, double (*f) (double), double tol, double *fx, double *f2x);

/* two-dimensional minimization - as input lower limits and upper limits and trial
values for the minimums are needed: min1 < *x1 < max1; min2 < *x2 < max2
the function and a fractional tolerance has to be specified
optimizeTwodimenmin returns the optimal x1, x2 values and the value of their second derivatives at these points
*/
static  void  optTwoDim (
       double tol1, 
       int active1, double min1, double *x1, double max1, double (*func1)(double), double *err1, 

       double tol2,
       int active2, double min2, double *x2, double max2, double (*func2)(double), double *err2
       
       );

static void turnOnOptedPam ();
static void turnOffOptedPam ();
static int isOptedPam ();

//release all memmory for this class
void release ();

//the destructor function
virtual ~Brent ();

/***************************************************************************************
***************************************************************************************

***************************************************************************************
***************************************************************************************/
private :
/* Brents method in one dimension */
static double opt (double ax, double bx, double cx, double (*f)(double), double tol, double *foptx, double *f2optx, double fax, double fbx, double fcx);
static int isOptedPam_;

}; //end of class

#endif
