/***************************************************************************
                          constants.h  -  description
                             -------------------
    begin                : Thu Feb 20 2003
    copyright            : (C) 2003 by Vinh Le Sy and Arndt von Haeseler
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CONSTANT_H
#define CONSTANT_H
#include <math.h>
/*
The Constants.h contains all constant values for whole project.
Every constant values MUST BE declared here.

...Being specified more
*/

/*IN ORDER TO REDUCE THE SOURCE CODES, SOME COMMOM ABBREVIATIONS ARE USED
alignment                      : ali
branch                         : br
list                           : ls
JukeCantor69Model              : JK69M
kimura80Model                  : K2P
equalInputModel                : F81M
hasegawa85Model                : HKY85M
felsenstein84Model             : F84M
tamuraNei93Model               : TM93M
generalReversibleModel         : GRM
internal                       : in
external                       : ex
array                          : arr
node                           : nd
likelihood                     : li;
optimize, optimal              : opt
unroot                         : ur
quartetTree                    : qtree
puzlle                         : puz
sequence                       : seq
utility                        : utl
transition                     : ts
transversion                   : tv
interation                     : it
pattern                        : ptn
dimension                      : dim
EPSILON                        : EPS
value                          : val
Substitution                   : sub
vector                         : vec
number                         : num
length                         : len
increase, increament           : inc
decrease, decreament           : dec
compute                        : cmp
parametter                     : pam
frequence                      : frq
active                         : act
small                          : s
middle                         : m
great                          : g
left                           : l
right                          : r
binary                         : bi
distance                       : dis
remain                         : rem
estimate                       : est
interation                     : it
negative                       : neg
child                          : chi
parrent                        : par
structure                      : str
current                        : cur
approximate                    : app
substitution                   : sub
instantaneous                  : ins
coefficient                    : cof
purine                         : purine (A, G)
pyrimidine                     : pyri (C, T)
average                        : ave
quartet                        : qua (q)
position                       : pos
descendant                     : des
pointer                        : poi
similar                        : sim
height                         : hei
small                          : sma
great                          : gre
company                        : com

*/

enum BASE  {BASE_A, BASE_C, BASE_T, BASE_G, BASE_AC, BASE_AG, BASE_AT, BASE_CG, BASE_CT, BASE_GT, BASE_CGT, BASE_AGT, BASE_ACT, BASE_ACG};

enum STATE {STATE_A, STATE_R, STATE_N, STATE_D, STATE_C, STATE_Q, STATE_E, STATE_G, 
            STATE_H, STATE_I, STATE_L, STATE_K, STATE_M, STATE_F, STATE_P, STATE_S, STATE_T, 
            STATE_W, STATE_Y, STATE_V, BS_UNKNOWN};

enum MODEL {HKY85, TN93, GTR, WAG, JTT, VT, MtREV24, Blosum62, Dayhoff};

enum EX_IN {EX, IN};
enum OPT_STATUS {APPRO, COMPLETE};
enum NEWICK_TREE_STATUS {WRONG_BRACKET_ORDER, NEWICK_TREE_FORM, MISSING_OPEN_BRACKET, MISSING_CLOSE_BRACKET, MISSING_FIRST_OPEN_BRACKET, MISSING_LAST_CLOSE_BRACKET, MISSING_COLON};
enum DRAW_TREE_TYPE {NON_BR_LEN, BR_LEN};
enum BR_DIR {StG, GtS};

enum DATA_TYPE {NUCLEOTIDE, AMINO_ACID};
enum RATE_TYPE {UNIFORM, SITE_SPECIFIC, GAMMA};

enum PAM_TYPE {ESTIMATE, USER_DEFINED, EQUAL, REAL, SIMULATION};
enum MODEL_PAM_TYPE {TS_TV, TS_AG, TS_CT, TV_AC, TV_AT, TV_CG, TV_GT, TS_TV_RATIO, PY_PU_RATIO};
enum STOPPING_TYPE {YES, YES_MIN_ITER, YES_MAX_ITER, NO};


const int SMALL_DATA_SET = 101;
const int  MAX_ITERATION = 10000;
const int  START_PREDICTION = 4;

const char BAD_MEMORY_ACCESS[] = "Accessed out of array. Please check the input data or contact the authors!!!";
const char PRO_VERSION[] = "Version 2.6";

const char MEMORY_LOCATION[] = "Out of memory!!!";


const char BAD_ASSIGNMENT[] = "Can not perform matrix A = matrix B while their rows and columns are different";

const char BAD_CMP_LI[] = "The value of likelihood is SMALLER THAN 0.0 !!!";

const char NEWICK_TREE_ALIGNMENT_ERROR[] = "the newick tree form and the alignment is not adaptive";

const char OUT_GRP_ND_ERROR[] = "The out group node is improper...";

const int NUM_CAN_STATE = 21;
const int MAX_SEQ_NAME_LEN = 1000;

const int NUM_BASE = 4;
const int NUM_AMINO = 20;

//the number of ANSSI character set
const int NUM_CHAR = 256;

const int FILENAME_LEN = 10000;


//the last control charater
const int LAST_CONTROL_CHAR = 32;

//the maximum neighbor nodes of an internal nodes
const int MAX_NUM_NEI_ND = 3;

//the maximum number childrent of a internal node
const int MAX_NUM_CHI = 3;


/*if baseNo1 + baseNo2 = TRANSITION_SIGN, it means that baseNo1 and baseNo2 are
either purine or phyrimidine
*/

const int TS_SIGN = 3;

//the INCREMENT is for vector template
const int INC = 4;

//the maximum number of sequences
const int MAX_NUM_SEQ = 2005;
const int MAX_NUM_CLUSTER = MAX_NUM_SEQ * 100;

const int MAX_NUM_REP       = 31;

//the max number of positive pattern
const int MAX_NUM_PTN  = 100100;
const int MAX_NUM_SITE = 100000;

//the max number of rate for heterogeneity
const int MAX_NUM_RATE = 32;

//the mean of substitution rate (muy)
const double MEAN_SUB_RATE = 1.0;

const int INFINITIVE = 1000000000;

const int MIN_EXP = 200;

const long double MIN_EXP_VAL = exp (-MIN_EXP);

//the is the maximum length of a line in the alignment file
const int MAX_LINE_LEN = 100000;

//the maximum newich tree form length

const int MAX_NEWICK_TREE_LEN = 100000;

const double ZERO = 0.000000001;

//the limits of branch length
const double MAX_BR_LEN = 9.00; /* upper limit on branch length */
const double MIN_BR_LEN = 0.000001; /* lower limit on branch length */


/* error in branch length, using for optimizing the branch length be Brent method
for the pair sequences optimizing*/
const double EPS_OPT_PAIR_SEQ_BR	= 0.000001;

/* error in branch length, using for optimizing the branch length be Brent method
for the quartet optimizing*/
const double EPS_OPT_UR_TREE_BR	    = 0.000001;

const double EPS_GAMMA_SHAPE_ERROR =  0.001; /*using for optimizing parameters */
const double EPS_MODEL_PAM_ERROR      =  0.0001; /*using for optimizing parameters */


/* maximum number of iterations of smoothing using
for maximizing all brances of the unrooted tree*/
const int MAX_IT_UR_BR = 100;

//the optimal urooted tree error, using for (*opt) function
const double BR_DIF_ERROR = 0.000001;

//using for optimizing parameters
const int MAX_IT_UR_BR_PAM_OPT = 3;
const int MAX_IT_ALL_PAM_OPT = 3;

const double MIN_FRQ                = 0.0001;
const double MIN_FRQ_DIF            = 0.00002;

const double MIN_STATE_FRQ = 0.0001;  /* lower limit on base frequencies = 0.01% */

const double MIN_RATE = 0.00001; //all variable rates should be greater than MIN_RATE

//the limits of tsTvRatio, using for optimizing tsTvRatio
const double MAX_TS_TV_RATIO = 32.0;
const double MIN_TS_TV_RATIO = 0.2;

const double MAX_PY_PU_RATIO = 32.0;
const double MIN_PY_PU_RATIO = 0.2;

const double MIN_GEN_PAM = 0.0001;
const double MAX_GEN_PAM = 0.99;

//using for optimizing shape parameters of discrete gamma distribution
const double MIN_GAMMA_SHAPE = 0.1;
const double MAX_GAMMA_SHAPE = 100.0;
const double GAMMA_SHAPE_ERROR = 0.1;
const double EPS_LOGLI_SITE_SPECIFIC_RATE_OPT = 0.01;
const double MAX_IT_TREE_SITE_SPECIFIC_RATE_OPT = 10;
const int MAX_IT_SITE_SPECIFIC_RATE_OPT = 100;

const double MIN_PTN_RATE = 0.000001;
const double MAX_PTN_RATE = 100.0;
const double EPS_PTN_ERROR = 0.001;


//using for drawing the tree
const int NUM_ABSTRACT_BR_LEN_CLASS = 3;

//using for drawing the tree
const int MIN_ABSTRACT_BR_LEN = 4;

#endif

