/***************************************************************************
                          hasegawa85model.h  -  description
                             -------------------
    begin                : Thu Feb 20 2003
    copyright            : (C) 2003 by Vinh Le Sy and Arndt von Haeseler
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HKY85M_H
#define HKY85M_H


#include "basem.h"
 /*
Hasegawa et al. (HKY1985) takes into account of both the different equilibrium frequencies of bases
( the frequence of A # the frequence of C # the frequence of G # the frequence of T  )
and the common observation that transitions and transversions occur at different rates.
  transition transvertion ratio = transition rate / transversition rate

When ( the frequence of A = the frequence of C = the frequence of G = the frequence of T  )
the HYK85 model becomes the Kimura 1980 model,
When the transition tranverstion ratio = 1, the HYK85 model becomes the EqualInputModel (F81) model,

THE NUMBER OF FREE VARIABLES IS 4
the Hasegawa 1985 model is A brLen-REVERSIBLE MODEL
...Being specified more
*/
class HKY85M : public BaseM
{
public :

//the constructor function
HKY85M ();

//compute the substitution rate of this model
static double cmpSubRate ();

//compute the substitution rate of this model
static void reCmpSubRate ();

//cmp the base frequencies for this alignment
//virtual void cmpBaseFrq ();

//all things are inited here
virtual void init ();

/*
cmp the probability of changing from nucleotide baseNo1 into nucleotide baseNo2
after a period of brLen  / subRate_
*/
virtual double cmpProbChange (const int baseNo1, const int baseNo2, const double brLen);


//cmp the vector probability of change from baseNo
virtual void cmpProbChange (const int baseNo, const double brLen, DVec20 &probVec);

/*
cmp the array of probability of changing from
one nucleotide into one nucleotide after a period of brLen / subRate_
*/
virtual void cmpProbChange (const double brLen, DMat20 &probMat);


virtual void optPam ();

/*compute the log likelihood when given tsTvRatio,
this function is used to optimize the tsTvRatio of this model
*/
static double cmpNegLogLiTsTvRatio (double tsTvRatio);

//release all memmory for this class
void release ();

//the destructor function
virtual ~HKY85M ();

}; //end of hky85 model class

#endif
