#ifndef INTERFACE_H
#define INTERFACE_H

#include <iostream>
#include <fstream>

#include "model.h"
#include "ali.h"
#include "rate.h"
#include "outstream.h"
#include "ptnls.h"
#include "model.h"

extern int isContinuous;

class Interface {

public :

int cmpnIter () {
  if (nSeq_ < 31)
    return 20;

  if (nSeq_ < 101)
    return nSeq_;

  if (nSeq_ < 151)
    return 100;

  if (nSeq_ < 201)
    return 200;

  if (nSeq_ < 1001)
    return 300;  

  return 500;
}

double cmpProbSel (PAM_TYPE sourceType) {   
  if (sourceType == SIMULATION)    
    return 0.3;

  if (nSeq_ < 51)
    return 0.5;

  if (nSeq_ < 100)
    return 0.3;

  if (nSeq_ < 200)
    return 0.2;

  return 0.1;
}

void getIntNumber (const char *msg, double lowerBound, double upperBound, int &value) {
  char sValue_[20];
  char *stopPos_;
  do {
     std::cout << msg;
     std::cin >> sValue_;
     value = static_cast<int> (strtod (sValue_, &stopPos_));

     std::cout.precision (12);
     if (value < lowerBound || value > upperBound)
       std::cout <<"The entered number must be between " << lowerBound << " and " << upperBound << ". Please, try again!!!" << endl;
  } while (value < lowerBound || value > upperBound);
}

void getDoubleNumber (const char *msg, double lowerBound, double upperBound, double &value) {
  char sValue_[20];
  char *stopPos_;
  do {
     std::cout << msg;
     std::cin >> sValue_;
      value = strtod (sValue_, &stopPos_);

     std::cout.precision (12);
     if (value < lowerBound || value > upperBound)
       std::cout <<"The entered number must be between " << lowerBound << " and " << upperBound << ". Please, try again!!!" << endl;
  } while (value < lowerBound || value > upperBound);
}

//=====================================================================================
//=====================================================================================
int getOutGrpSeq () {
     std::cout << "Enter name of outgroup sequence: ";
     char tmpOutGrpSeqName_[MAX_SEQ_NAME_LEN];
     std::cin >> tmpOutGrpSeqName_;

     Vec<char> userOutGrpSeqName_(MAX_SEQ_NAME_LEN);
     int pos_ = 0;
      while (Utl::isControlChar (tmpOutGrpSeqName_[pos_] ) == 1)
         pos_ ++;
      while (tmpOutGrpSeqName_[pos_] != '\0') {
         userOutGrpSeqName_ += tmpOutGrpSeqName_[pos_];
         pos_ ++;
      }

     int userOutGrpSeqNo_ = Ali::getSeqNo (userOutGrpSeqName_);
     if (userOutGrpSeqNo_ != -1) 
        return userOutGrpSeqNo_;
     else {
        std::cout <<"Warning!!! The entered outgroup sequence \"" << tmpOutGrpSeqName_ << "\" does not exist!!!" << endl;
        return -1;
     } 
}

//=========================================================================================
//=========================================================================================
//=========================================================================================
int getParameter (
                        int        &nIter,       
                        double     &probSel,
                        int        &nRep,
                        STOPPING_TYPE     &stoppingRule,
                        DATA_TYPE  &dataType,
                        MODEL      &model,
                        double      &baseA, double &baseC, double &baseG, double &baseT,
                        PAM_TYPE    &baseFrqType,

                        double      &tsTvRatio,
                        PAM_TYPE    &tsTvRatioType,

                        double      &pyPuRatio,
                        PAM_TYPE    &pyPuRatioType,

                        double      &tsAG, 
                        double      &tsCT, 
                        double      &tvAC, 
                        double      &tvAT, 
                        double      &tvCG, 
                        double      &tvGT, 
                        PAM_TYPE    &genPamType,

                        RATE_TYPE   &rateType,
                        int         &nRate,
                        double      &gammaShape,
                        PAM_TYPE    &gammaShapeType,

                        int         &outGrpSeqNo) {

  std::cout << endl << endl << endl << endl << endl;

  nSeq_ = Ali::getNSeq ();

  char treeLsFileName_[FILENAME_LEN];
  char checkPointFileName_[FILENAME_LEN];
  strcpy (treeLsFileName_, Ali::fileName_);
  strcpy (checkPointFileName_, Ali::fileName_);
  char treeLsTail_[FILENAME_LEN] = ".iqpnni.treels";
  char checkPointTail_[FILENAME_LEN] = ".iqpnni.checkpoint";
  strcat (treeLsFileName_, treeLsTail_);
  strcat (checkPointFileName_, checkPointTail_);


  int isPosContinuous_ = 0;
  if (isExistedFile (treeLsFileName_) == 1  && isExistedFile (checkPointFileName_) == 1) 
     isPosContinuous_= getPamCheckPoint (checkPointFileName_,
                                   nIter, probSel, nRep, stoppingRule,
                                   dataType, model,
                                   baseA, baseC, baseG, baseT,
                                   baseFrqType,   

                                   tsTvRatio,
                                   tsTvRatioType,
 
                                   pyPuRatio,
                                   pyPuRatioType,

                                   tsAG, tsCT, tvAC, tvAT, tvCG, tvGT,
                                   genPamType, 
                                      
                                   rateType, nRate, gammaShape,
                                   gammaShapeType,
                                   outGrpSeqNo); 

  isContinuous = isPosContinuous_; 
  if (isContinuous == 0) {
    nRep = 4;
    nIter = cmpnIter ();
    PAM_TYPE sourceType_;
    sourceType_ = Ali::detectDataSource ();
    dataType = Ali::detectDataType ();

    if (dataType == NUCLEOTIDE)
      model  = HKY85;
    else
     model = WAG;

    probSel = cmpProbSel (sourceType_);

    baseFrqType = ESTIMATE;
    tsTvRatioType = ESTIMATE;
    pyPuRatioType = ESTIMATE;
    genPamType = ESTIMATE;
    gammaShapeType = ESTIMATE;
 
    rateType = UNIFORM;
    nRate = 1;
    gammaShape = 1.0;

    stoppingRule = YES_MIN_ITER;
    outGrpSeqNo = 0;
  } //end of if

  Vec<char> outGrpSeqName_;
  Ali::getName (outGrpSeqNo, outGrpSeqName_);

  do {
     std::cout <<     "GENERAL OPTIONS" << endl;
     if (isPosContinuous_ == 1)
       if (isContinuous == 1)
         std::cout <<      " b        Restart from the last stopped point? Yes" << endl;
       else 
         std::cout <<      " b        Restart from the last stopped point? No (start a new search)" << endl;
     
     std::cout <<         " o                        Display as outgroup? " << outGrpSeqName_ << endl;
     std::cout <<         " n                       Number of iterations? " << nIter << endl;
     switch (stoppingRule) {
        case YES:
          std::cout <<         " s              Stopping rule (if applicable)? " << "Yes" << endl << endl;
          break;
        case YES_MIN_ITER:
          std::cout <<         " s              Stopping rule (if applicable)? " << "Yes, but at least " << nIter << " iterations" << endl << endl;
          break;
        case YES_MAX_ITER:
          std::cout <<         " s              Stopping rule (if applicable)? " << "Yes, but at most " << nIter << " iterations" << endl << endl;
          break;         
        default: 
          std::cout <<         " s                              Stopping rule? " << "No" << endl << endl;
     }
         
     std::cout <<     "IQP OPTIONS" << endl;
     std::cout <<         " p         Probability of deleting a sequence? " << probSel << endl;
     std::cout <<         " k                     Number representatives? " << nRep << endl << endl;
  
     std::cout << "SUBSITUTION PROCESS" << endl;
     if (dataType == AMINO_ACID) {
      std::cout <<       " d                Type of sequence input data? " << "Amino acids" << endl;
      switch (model) {
        case WAG:
           std::cout <<     " m                      Model of substitution? " << "WAG (Whelan-Goldman 2000)" << endl;       
           break;
        case JTT:
          std::cout <<     " m                      Model of substitution? " << "JTT (Jones et at. 1992)" << endl;
          break;
        case VT:                              
           std::cout <<     " m                      Model of substitution? " << "VT (Mueller-Vingron 2000)" << endl;       
           break;
        case MtREV24:
           std::cout <<     " m                      Model of substitution? " << "mtREV24 (Adachi-Hasegawa 1996)" << endl;       
           break;
        case Blosum62:
           std::cout <<     " m                      Model of substitution? " << "BLOSUM62 (Henikoff-Henikoff 1992)" << endl;       
           break;
        default:
           std::cout <<     " m                      Model of substitution? " << "Dayhoff (Dayhoff et al. 1978)" << endl;       
       } //end of switch

       if (baseFrqType == ESTIMATE)
           std::cout <<   " f                    Amino acide frequencies? " << "Estimate from data" << endl;
       else
           std::cout <<   " f                    Amino acide frequencies? " << "Default" << endl;

     } else {
       std::cout <<       " d                Type of sequence input data? " << "Nucleotides" << endl;
       switch (model) {
         case HKY85:
           std::cout <<     " m                      Model of substitution? " << "HKY85 (Hasegawa et al. 1985)" << endl;       
           if (tsTvRatioType == ESTIMATE)
             std::cout <<     " t                 Ts/Tv ratio (0.5 for JC69)? " << "Estimate from data" << endl;
           else 
            std::cout <<     " t                 Ts/Tv ratio (0.5 for JC69)? " << tsTvRatio << endl;
           break;

         case TN93:
           std::cout <<     " m                      Model of substitution? " << "TN93 (Tamura-Nei 1993)" << endl;       
           if (tsTvRatioType == ESTIMATE)
             std::cout <<     " t                 Ts/Tv ratio (0.5 for JC69)? " << "Estimate from data" << endl;
           else 
             std::cout <<     " t                 Ts/Tv ratio (0.5 for JC69)? " << tsTvRatio << endl;

           if (pyPuRatioType == ESTIMATE)
             std::cout <<     " u                    Pyrimidine/Purine ratio? " << "Estimate from data" << endl;
           else 
             std::cout <<     " u                    Pyrimidine/Purine ratio? " << pyPuRatio << endl;
           break;

         default:
           std::cout <<     " m                      Model of substitution? " << "General time reversible" << endl;       
           if (genPamType == ESTIMATE) 
             std::cout <<    " g   General time reversible model parameters? " << "Estimate from data" << endl;
           else
             std::cout <<    " g   General time reversible model parameters? " << "User defined" << endl;
       }
       
       if (baseFrqType == ESTIMATE)
           std::cout <<       " f                           Base frequencies? " << "Estimate from data" << endl;
        else
         if (baseFrqType == EQUAL)
           std::cout <<   " f                           Base frequencies? " << "Equal" << endl;
         else                                 
           std::cout <<   " f                           Base frequencies? " << "User defined" << endl;
    } //end of model

  std::cout <<  endl <<   "RATE HETEROGENEITY" << endl;
  if (rateType == UNIFORM)
    std::cout <<        " w                Model of rate heterogeneity? " << "Uniform rate" << endl;
  else
     if (rateType == SITE_SPECIFIC) 
       std::cout <<     " w                Model of rate heterogeneity? " << "Site-specific substitution rates" << endl;
     else {
        std::cout <<    " w                Model of rate heterogeneity? " << "Gamma distributed rates" << endl;
        if (gammaShapeType == ESTIMATE) 
           std::cout << " a         Gamma distribution parameter alpha? " << "ESTIMATE" << endl;
        else


           std::cout << " a         Gamma distribution parameter alpha? " << gammaShape << endl;   

          std::cout <<  " c            Number of Gamma rate categogies? " << nRate << endl;    
      }  
    
    std::cout << endl << "quit [q]," << " confirm [y]," << "or change [menu] settings: ";
  char option_;

  std::cin >> option_;
  if (option_ == 'q')
    return 0;
  if (option_ == 'y')
    break;

  if (option_ == 'b' && isPosContinuous_ == 1) {
    isContinuous = 1 - isContinuous;
    if (isContinuous == 1) 
       getPamCheckPoint (checkPointFileName_,
                                   nIter, probSel, nRep, stoppingRule,
                                   dataType, model,
                                   baseA, baseC, baseG, baseT,
                                   baseFrqType,   

                                   tsTvRatio,
                                   tsTvRatioType,
 
                                   pyPuRatio,
                                   pyPuRatioType,

                                   tsAG, tsCT, tvAC, tvAT, tvCG, tvGT,
                                   genPamType, 
                                      
                                   rateType, nRate, gammaShape,
                                   gammaShapeType,
                                   outGrpSeqNo); 
  } //end of option 'b'

  if (option_ == 's') {
    switch (stoppingRule) {
       case YES:
          stoppingRule = YES_MIN_ITER;
          break;
       case YES_MIN_ITER:
          stoppingRule = YES_MAX_ITER;
          break;
       case YES_MAX_ITER:
          stoppingRule = NO;
          break;
       default:
         stoppingRule = YES;
    }
  } //end of option 's'
  
  if (option_ == 'o')  {
     int tmpOutGrpSeqNo_ = getOutGrpSeq ();
     if (tmpOutGrpSeqNo_ >= 0 && tmpOutGrpSeqNo_ < nSeq_)
       outGrpSeqNo = tmpOutGrpSeqNo_;
     Ali::getName (outGrpSeqNo, outGrpSeqName_);
  }

  if (option_ == 'n') 
    getIntNumber ("Enter number of iterations: ", 1, INFINITIVE, nIter);

  if (option_ == 'p') {
     getDoubleNumber ("Enter probability of deleting a sequence (0.01, 1.0): ", 0.01, 1.0, probSel);
  }

  if (option_ == 'k') {
    getIntNumber ( "Enter the number of representatives: ", 1, nSeq_, nRep);
  }


  if (option_ == 'f') {
    if (dataType == AMINO_ACID) {
      if (baseFrqType == ESTIMATE)
        baseFrqType = USER_DEFINED;
      else
        baseFrqType = ESTIMATE;   
    } else {
      if (baseFrqType == ESTIMATE) {
         baseFrqType = EQUAL;
         baseA = 0.25;
         baseC = 0.25;
         baseG = 0.25;
         baseT = 0.25;
       } else
         if (baseFrqType == EQUAL) {
           do {
             baseFrqType = USER_DEFINED;
             getDoubleNumber (" Base A (0.0, 1.0): ", 0.0, 1.0, baseA);
             getDoubleNumber (" Base C (0.0, 1.0): ", 0.0, 1.0, baseC);
             getDoubleNumber (" Base G (0.0, 1.0): ", 0.0, 1.0, baseG);

             if (baseT + baseC + baseG > 1.0) 
               std::cout <<"Error!!! the sum of base frequencies is greater than 1.0. Please, enter again..." << endl;
              baseT = 1.0 - baseA - baseC - baseG;
            } while (baseT < -ZERO);
         } else
           baseFrqType = ESTIMATE;
     } //end of else
  }//end of option 'f'


  if (option_ == 't') {
    if (tsTvRatioType == USER_DEFINED) 
      tsTvRatioType = ESTIMATE;
    else {
      tsTvRatioType = USER_DEFINED;
      getDoubleNumber ("Enter ts/tv ratio (sensible number): ", ZERO, INFINITIVE, tsTvRatio);
    }
  }
 
  if (option_ == 'u') {
    if (pyPuRatioType == USER_DEFINED) 
      pyPuRatioType = ESTIMATE;
    else {
      pyPuRatioType = USER_DEFINED;
      getDoubleNumber ("Enter pyridinine/purine ratio (sensible number): ", ZERO, INFINITIVE, pyPuRatio);
    }
  }

  if (option_ == 'g') {
    if (genPamType == ESTIMATE)  {
      genPamType = USER_DEFINED;
      getDoubleNumber ("Enter transversion rate from A to C: ", ZERO, INFINITIVE, tvAC);       
      getDoubleNumber ("Enter transition   rate from A to G: ", ZERO, INFINITIVE, tsAG);       
      getDoubleNumber ("Enter transversion rate from A to T: ", ZERO, INFINITIVE, tvAT);       
      getDoubleNumber ("Enter transversion rate from C to G: ", ZERO, INFINITIVE, tvCG);       
      getDoubleNumber ("Enter transition   rate from C to T: ", ZERO, INFINITIVE, tsCT);       
      getDoubleNumber ("Enter transversion rate from G to T: ", ZERO, INFINITIVE, tvGT);       
    }
    else
     genPamType = ESTIMATE;
  }

  if (option_ == 'd')
    if (dataType == NUCLEOTIDE)  {
      dataType = AMINO_ACID;
      model = WAG;
    } else {
      dataType = NUCLEOTIDE;
      model = HKY85;
    }
    
  if (option_ == 'm') {
    if (dataType == AMINO_ACID) {
      switch (model) {
         case WAG:
            model = JTT;
            break;
         case JTT:
           model = VT; 
           break;
        case VT:
          model = MtREV24;
          break;
       case MtREV24:
          model = Blosum62;
          break;
       case Blosum62:
          model = Dayhoff;
          break;
       default:
          model = WAG;
     } // end of swith
    } else {//DNA 
       if (model == HKY85)
         model = TN93;
       else
         if (model == TN93)
          model = GTR;
         else
          model = HKY85;
    }
  }  //end of option 'm'
    
  if (option_ == 'w') {
    if (rateType == SITE_SPECIFIC) {
      nRate = 1;
      rateType = UNIFORM;
    }  else 
      if (rateType == UNIFORM) {
         rateType = GAMMA;
         nRate = 4;
         gammaShape = 1.0;
      } else {
         nRate = Ali::getNSite ();
         rateType = SITE_SPECIFIC;
      }
  }
    
  if (option_ == 'a') {
     if (gammaShapeType == USER_DEFINED) 
        gammaShapeType = ESTIMATE;      
    else {
       gammaShapeType = USER_DEFINED;
       getDoubleNumber ("Gamma distribution parameter alpha (sensible number): ", ZERO, INFINITIVE, gammaShape);
    } 
  }

  if (option_ == 'c') {
    getIntNumber ("Number of Gamma rate categories ( 1, 32 ): ", 1,  32, nRate);
  }

    for (int count_ = 0; count_ < 15; count_ ++)
    std::cout << endl;
  } while (1); //finish user defined parameters!!!!!!!!!!!
  

  if (baseFrqType == ESTIMATE) {
    baseA = -1.0;
    baseC = -1.0;
    baseG = -1.0;
    baseT = -1.0;
  }  
  
  if (tsTvRatioType == ESTIMATE) 
    tsTvRatio = - 1.0;
  if (pyPuRatioType == ESTIMATE)
    pyPuRatio = -1.0;
  if (genPamType == ESTIMATE) {
     tsAG = -1.0;
     tsCT = -1.0;
     tvAC = -1.0;
     tvAT = -1.0;
     tvCG = -1.0;
     tvGT = -1.0;
  }
  

  if (rateType == GAMMA) {
    if (gammaShapeType == ESTIMATE) 
      gammaShape = -1.0;
  } else {
     gammaShape = -1.0;
     if (rateType == SITE_SPECIFIC)
       nRate = -1;
  }

  return 1;
}


//====================================================================================
//ab:cccc:dd: X
int getIntNum (const char  *aLine) { 
   char intNumStr_[20];
   int numPos_ = 0;
  
   int len_ = strlen (aLine);
   int semiColonPos_ = len_ - 1;
   while (aLine[semiColonPos_] != ':')
     semiColonPos_ --;

   for (int linePos_ = semiColonPos_ + 1; linePos_ < len_; linePos_ ++) 
      if (aLine[linePos_] >= '0' && aLine[linePos_] <= '9') {
         intNumStr_[numPos_] = aLine[linePos_];
         numPos_ ++;
      }        
  intNumStr_[numPos_]= '\0';
  char *endPos_;
  double doubleVal_ = strtod (intNumStr_, &endPos_);

  return static_cast<int> (doubleVal_ + ZERO);
}

//====================================================================================
//ab:cccc:dd: X
double getDoubleNum (const char  *aLine) { 

   char doubleNumStr_[40];
   int numPos_ = 0;
  
   int len_ = strlen (aLine);
   int semiColonPos_ = len_ - 1;
   while (aLine[semiColonPos_] != ':')
     semiColonPos_ --;

   for (int linePos_ = semiColonPos_ + 1; linePos_ < len_; linePos_ ++) 
      if (aLine[linePos_] == '.' || aLine[linePos_] == '-' || 
         (aLine[linePos_] >= '0' && aLine[linePos_] <= '9') ) {
         doubleNumStr_[numPos_] = aLine[linePos_];

         numPos_ ++;
      }        

  doubleNumStr_[numPos_]= '\0';
  char *endPos_;
  double doubleVal_ = strtod (doubleNumStr_, &endPos_);
  return doubleVal_;
}

//====================================================================================
//====================================================================================
//get all parameters from the last stopped point
int getPamCheckPoint ( char          *checkPointFileName, 
                        int        &nIter,       
                        double     &probSel,
                        int        &nRep,
                        STOPPING_TYPE     &stoppingRule,
                        DATA_TYPE  &dataType,
                        MODEL      &model,
                        double      &baseA, double &baseC, double &baseG, double &baseT,
                        PAM_TYPE    &baseFrqType,


                        double      &tsTvRatio,
                        PAM_TYPE    &tsTvRatioType,

                        double      &pyPuRatio,
                        PAM_TYPE    &pyPuRatioType,

                        double      &tsAG, 
                        double      &tsCT, 
                        double      &tvAC, 
                        double      &tvAT, 
                        double      &tvCG, 
                        double      &tvGT, 
                        PAM_TYPE    &genPamType,

                        RATE_TYPE   &rateType,
                        int         &nRate,
                        double      &gammaShape,
                        PAM_TYPE    &gammaShapeType,

                        int         &outGrpSeqNo) {

   ifstream checkPointFile_;
   checkPointFile_.open (checkPointFileName);

   char aLine_[MAX_LINE_LEN];


   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   nIter = getIntNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   probSel = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);

   nRep = getIntNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int stoppingRuleNum_ = getIntNum (aLine_);
   switch (stoppingRuleNum_) {
      case 0: 
         stoppingRule = YES;
         break;
      case 1: 
         stoppingRule = YES_MIN_ITER;
         break;
      case 2:
        stoppingRule = YES_MAX_ITER;
        break;
      default:
        stoppingRule = NO;
   }

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int dataTypeNum_ = getIntNum (aLine_);

   if (dataTypeNum_ == 0)
     dataType = NUCLEOTIDE;
  else
     dataType = AMINO_ACID;

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int modelTypeNum_ = getIntNum (aLine_);
   if (modelTypeNum_ == 0)
        model = HKY85;
   if (modelTypeNum_ == 1)
        model = TN93;
   if (modelTypeNum_ == 2)
        model = GTR;
   if (modelTypeNum_ == 3)
        model = WAG;
   if (modelTypeNum_ == 4)
        model = JTT;
   if (modelTypeNum_ == 5)
        model = VT;
   if (modelTypeNum_ == 6)
        model = MtREV24;
   if (modelTypeNum_ == 7)
        model = Blosum62;
   if (modelTypeNum_ == 8)
        model = Dayhoff;

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   baseA = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   baseC = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   baseG = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   baseT = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int baseFrqTypeNum_ = getIntNum (aLine_);
   if (baseFrqTypeNum_ == 0)
     baseFrqType = ESTIMATE;
   if (baseFrqTypeNum_ == 1)
     baseFrqType = USER_DEFINED;
   if (baseFrqTypeNum_ == 2)
     baseFrqType = EQUAL;

   
//======================================================
   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tsTvRatio = getDoubleNum (aLine_);
   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int tsTvRatioTypeNum_ = getIntNum (aLine_);
   if (tsTvRatioTypeNum_ == 0)
      tsTvRatioType = ESTIMATE;
   else
     tsTvRatioType = USER_DEFINED;


   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   pyPuRatio = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int pyPuRatioTypeNum_ = getIntNum (aLine_);
   if (pyPuRatioTypeNum_ == 0)
      pyPuRatioType = ESTIMATE;
   else
     pyPuRatioType = USER_DEFINED;

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tsAG = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tsCT = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tvAC = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tvAT = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tvCG = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   tvGT = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int genPamTypeNum_ = getIntNum (aLine_);
   if (genPamTypeNum_ == 0)
      genPamType = ESTIMATE;
   else
     genPamType = USER_DEFINED;


   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int rateTypeNum_ = getIntNum (aLine_);
   if (rateTypeNum_ == 0)
      rateType = UNIFORM;
   if (rateTypeNum_ == 1)
      rateType = SITE_SPECIFIC;
   if (rateTypeNum_ == 2)
      rateType = GAMMA;

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   nRate = getIntNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   gammaShape = getDoubleNum (aLine_);

   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int gammaShapeTypeNum_ = getIntNum (aLine_);
   if (gammaShapeTypeNum_ == 0)
      gammaShapeType = ESTIMATE;
   else
     gammaShapeType = USER_DEFINED;


   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   outGrpSeqNo = getIntNum (aLine_);

   
   checkPointFile_.getline (aLine_, MAX_LINE_LEN - 1);
   int isContinuous_ = (1 - getIntNum (aLine_));
   return isContinuous_;
}

//====================================================================================
//====================================================================================
int isExistedFile (const char *fileName) {
  ifstream existedFile_;
  existedFile_.open (fileName);


  int isExistedFile_ = 1;
  if (existedFile_ == 0)
    isExistedFile_ = 0;

  existedFile_.close ();

  return isExistedFile_;
}

//====================================================================================
//====================================================================================
int getUserPam ( char   *aliFileName, 
                        int        &nIter,       
                        double     &probSel,
                        int        &nRep,
                        STOPPING_TYPE     &stoppingRule,
                        DATA_TYPE  &dataType,
                        MODEL      &model,
                        double      &baseA, double &baseC, double &baseG, double &baseT,
                        PAM_TYPE    &baseFrqType,

                        double      &tsTvRatio,
                        PAM_TYPE    &tsTvRatioType,

                        double      &pyPuRatio,
                        PAM_TYPE    &pyPuRatioType,

                        double      &tsAG, 
                        double      &tsCT, 
                        double      &tvAC, 
                        double      &tvAT, 
                        double      &tvCG, 
                        double      &tvGT, 
                        PAM_TYPE    &genPamType,

                        RATE_TYPE   &rateType,
                        int         &nRate,
                        double      &gammaShape,
                        PAM_TYPE    &gammaShapeType,

                        int         &outGrpSeqNo) {

  std::cout << "WELCOME TO IQPNNI " << PRO_VERSION << endl << endl << endl;

   if (isExistedFile (aliFileName) == 0) {
     std::cout << "Please enter a file name for the sequence data: ";
     std::cin >> aliFileName;
     do {
        if (isExistedFile (aliFileName) == 0)     
          std::cout << "file " << aliFileName << " not found, please enter alternative name or type q to quit: ";  
        else
           break;
        std::cin >> aliFileName;
        if (aliFileName[0] == 'q')
          return 0;
     } while (1);
   }

  std::cout <<"Reading data..." << endl;
  int aliStatus_ = Ali::readInput (aliFileName);
  Ali::sortSeq ();

  if (aliStatus_ == 0) {
    std::cout << "The alignment file " << aliFileName << " is not correct!!!";
    return EXIT_SUCCESS;
  }

  Ali::checkSameName ();
  nSeq_ = Ali::getNSeq ();

  if (nSeq_ < 4) {
     std::cout <<"The alignment file MUST consist of at least 4 sequences !!!";
     return 0;
  }




  int status_ = getParameter    (
                                  nIter, probSel, nRep, stoppingRule,
                                   dataType, model,
                                   baseA, baseC, baseG, baseT,
                                   baseFrqType,   

                                   tsTvRatio,
                                   tsTvRatioType,
 
                                   pyPuRatio,
                                   pyPuRatioType,

                                   tsAG, tsCT, tvAC, tvAT, tvCG, tvGT,
                                   genPamType, 
                                      
                                   rateType, nRate, gammaShape,
                                   gammaShapeType,
                                   outGrpSeqNo); 

 
   return status_;
}

//=================================================================
//=================================================================
void writeSpecificRate (const char *rateFileName) {
  ofstream rateFile_;
  rateFile_.open (rateFileName);
  Vec<double> *ptnRateArr_;
  ptnRateArr_ = Rate::getPtnRate ();
   
  int nSite_ = Ali::getNSite ();
//  double totalRate_ = 0.0;
  for (int siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++) {
    int ptnNo_ = Ali::getPtn (siteNo_);
    double siteRate_ = (*ptnRateArr_)[ptnNo_];                                                        
    rateFile_ << siteNo_ + 1 << "\t";
    rateFile_.width (10);
    rateFile_.precision (4);
    rateFile_.setf (ios::left);
//    totalRate_ += siteRate_;

    rateFile_  << siteRate_ << endl;
  }
//  std::cout <<"total rates: " << totalRate_ << endl;
  rateFile_.close ();
}

//=====================================================================================================
//=====================================================================================================
//=====================================================================================================
void writeOut (const char *aliFile,  
               UrTree &tree, 
               int nIter, double probSel, int nRep, 
               STOPPING_TYPE stoppingRule,  
               Vec<char> &conTree, 
               const char *startedDate, double progTime) {

  std::ofstream outFile_;
  char outFileName_[FILENAME_LEN];
  strcpy (outFileName_, aliFile);
  char outTail_[FILENAME_LEN] = ".iqpnni";
  strcat (outFileName_, outTail_);

  char topoFileName_[FILENAME_LEN];
  strcpy (topoFileName_, aliFile);
  char topoTail_[FILENAME_LEN] = ".iqpnni.treefile";                                                       
  strcat (topoFileName_, topoTail_);      

  char rateFileName_[FILENAME_LEN];
  strcpy (rateFileName_, aliFile);
  char rateTail_[FILENAME_LEN] = ".iqpnni.rate";                                                       
  strcat (rateFileName_, rateTail_);      

  char treeLsFileName_[FILENAME_LEN];
  strcpy (treeLsFileName_, aliFile);
  char treeLsTail_[FILENAME_LEN] = ".iqpnni.treels";                                                       
  strcat (treeLsFileName_, treeLsTail_);      

  outFile_.open (outFileName_);  



  outFile_ << "THESE RESULTS ARE FROM IQPNNI " << PRO_VERSION  << "!!!!" << endl;
  outFile_ << "Le Sy Vinh and Arndt von Haeseler, IQPNNI: Moving Fast Through Tree Space" << endl;
  outFile_ << "and Stopping in Time, Mol. Biol. Evol., 21(8):1565-1571.2004" << endl << endl << endl;
  
  outFile_ << "Some parts of the code were taken from TREE-PUZZLE package:" << endl;
  outFile_ << "Heiko A. Schmidt, Korbinian Strimmer, Martin Vingron, and Arndt von Haeseler," << endl;
  outFile_ << "TREE-PUZZLE: Maximum likelihood phylogenetic analysis" << endl;
  outFile_ << "using quartets and parallel computing, Bioinformatics, 18:502-504, 2002" << endl << endl;

  outFile_ << "The source codes to construct the BIONJ tree were taken from BIONJ software:" << endl;
  outFile_ << "Oliver Gascuel, BIONJ: An Improved Version of the NJ Algorithm" << endl;
  outFile_ << "Based on a Simple Model of Sequence Data, Mol. Bio. Evol., 14:685-695, 1997" << endl << endl;

  outFile_ <<"================================================================================="<< endl;
  outFile_ <<"                              INPUT DATA INFORMATION                                                                         " << endl;
  outFile_ <<"================================================================================="<< endl << endl;
  outFile_ <<"General information:" << endl;
  outFile_ << "Alignment file name: " << aliFile << endl;
  int nSeq_ = Ali::getNSeq ();
  int seqLen_ = Ali::getSeq (0).getSize ();
  outFile_ << "Number of sequences: "  << nSeq_ << endl;
  outFile_ << "Number of sites: " << seqLen_ << endl;

  int outGrpNdNo_ = tree.getOutGrpNd ();
  Vec<char> outGrpSeqName_;
  Ali::getName (outGrpNdNo_, outGrpSeqName_);
  outFile_ << "Display as outgroup: "; 
  OutStream::write (outGrpSeqName_, outFile_) << endl;
  outFile_ << "Number of iterations: " << nIter << endl;
  switch (stoppingRule) {
     case 0:
       outFile_ <<"Stopping rule: Yes" << endl << endl;
       break;
     case 1:
       outFile_ <<"Stopping rule: Yes, but at least " << nIter << " iterations" << endl << endl;
       break;
     case 2:
       outFile_ <<"Stopping rule: Yes, but at most " << nIter << " iterations" << endl << endl;
       break;
    default:
       outFile_ <<"Stopping rule: No" << endl << endl;
  } //end of switch
       
  outFile_ <<"IQP information:" << endl;
  outFile_ << "Probablity of deleting one sequence: " << probSel << endl;
  outFile_ << "Number of representatives: " << nRep << endl << endl;

  outFile_ <<"Substitution process: " << endl;
  MODEL model_ = Model::getModelType ();

  if (Model::getDataType () == AMINO_ACID) {
    outFile_ << "Type of sequence input data: Amino acids" << endl;
    outFile_ << "Model of substitution: ";  
    switch (model_) {
      case JTT: 
        outFile_ << "Jones et al. 1992 model " << endl;
        break;
      case Dayhoff:
        outFile_ << "Dayhoff et al. 1978 model " << endl;
        break;
      case VT:
        outFile_ << "Mueller, T. and Vingron 2000 " << endl;
        break;    

      case MtREV24:
        outFile_ << "Adachi, J. and Hasegawa, M. (1996)" << endl;
        break;
      case Blosum62:
        outFile_ << "S. Henikoff and J. G. Henikoff.  1992" << endl;
        break;

      default:
        outFile_ << "S. Whelan and N. Goldman.  2000" << endl;
     } //end of switch

     if (Model::getBaseFrqType () == ESTIMATE) 
       outFile_ << endl << "Amino acid Frequencies (Estimate from the data set):" << endl;
     else
       outFile_ << endl << "Amino acid Frequencies (From the model):" << endl;

     DVec20 stateFrqArr_;
     Model::getStateFrq (stateFrqArr_);
     outFile_ << "Amino acid A: " << stateFrqArr_[STATE_A] << endl; //1
     outFile_ << "Amino acid C: " << stateFrqArr_[STATE_C] << endl; //2
     outFile_ << "Amino acid D: " << stateFrqArr_[STATE_D] << endl; //3
     outFile_ << "Amino acid E: " << stateFrqArr_[STATE_E] << endl; //4
     outFile_ << "Amino acid F: " << stateFrqArr_[STATE_F] << endl; //5

     outFile_ << "Amino acid G: " << stateFrqArr_[STATE_G] << endl; //6
     outFile_ << "Amino acid H: " << stateFrqArr_[STATE_H] << endl; //7
     outFile_ << "Amino acid I: " << stateFrqArr_[STATE_I] << endl; //8
     outFile_ << "Amino acid K: " << stateFrqArr_[STATE_K] << endl; //9
     outFile_ << "Amino acid L: " << stateFrqArr_[STATE_L] << endl; //10

     outFile_ << "Amino acid M: " << stateFrqArr_[STATE_M] << endl; //11
     outFile_ << "Amino acid N: " << stateFrqArr_[STATE_N] << endl; //12
     outFile_ << "Amino acid P: " << stateFrqArr_[STATE_P] << endl; //13
     outFile_ << "Amino acid Q: " << stateFrqArr_[STATE_Q] << endl; //14
     outFile_ << "Amino acid R: " << stateFrqArr_[STATE_R] << endl; //15

     outFile_ << "Amino acid S: " << stateFrqArr_[STATE_S] << endl; //16
     outFile_ << "Amino acid T: " << stateFrqArr_[STATE_T] << endl; //17
     outFile_ << "Amino acid V: " << stateFrqArr_[STATE_V] << endl; //18
     outFile_ << "Amino acid W: " << stateFrqArr_[STATE_W] << endl; //19
     outFile_ << "Amino acid Y: " << stateFrqArr_[STATE_Y] << endl << endl; //20

  } else {
    outFile_ << "Type of sequence input data: Nucleotides" << endl;
    if (model_ == HKY85) {
       outFile_ <<"Model of substitution:  HKY85 (Hasegawa et al. 1985)" << endl;
         
        double tsTvRatio_ = Model::getTsTvRatio ();
        outFile_ << "Transition/transversion ratio (0.5 for JC69): ";

        outFile_.precision (5);
        outFile_ << tsTvRatio_/2.0 << endl;
        if (Model::getTsTvRatioType () == ESTIMATE)
            if (tsTvRatio_ < MIN_TS_TV_RATIO + 0.1 || tsTvRatio_ > MAX_TS_TV_RATIO - 0.1)
               outFile_ <<"Warning!!! Estimated transition/transversion ratio = " << tsTvRatio_/2.0 << " is close to the boundary of admissible values" << endl;
    }  //end of HKY85


    if (model_ == TN93) {
       outFile_ <<"Model of substitution:  TN93 (Tamura-Nei 1993)" << endl;
         
        double tsTvRatio_ = Model::getTsTvRatio ();
        outFile_ << "Transition/transversion ratio (0.5 for JC69): ";
        outFile_.precision (5);
        outFile_ << tsTvRatio_/2.0 << endl;
        if (Model::getTsTvRatioType () == ESTIMATE)
          if (tsTvRatio_ < MIN_TS_TV_RATIO + 0.1 || tsTvRatio_ > MAX_TS_TV_RATIO - 0.1)
            outFile_ <<"Warning!!! Estimated transition/transversion ratio = " << tsTvRatio_/2.0 << " is close to the boundary of admissible values" << endl;

        double pyPuRatio_ = Model::getPyPuRatio ();
        outFile_ << "Pyrimidine/purine ratio: " << pyPuRatio_ << endl;

        if (Model::getPyPuRatioType () == ESTIMATE)
          if (pyPuRatio_ < MIN_PY_PU_RATIO + 0.1 || pyPuRatio_ > MAX_PY_PU_RATIO - 0.1)
            outFile_ <<"Warning!!! Estimated pyrimidine/purine ratio = " << tsTvRatio_/2.0 << " is close to the boundary of admissible values" << endl;
    } //end of YN93  

    if (model_ == GTR) {// model_ == GTR
        outFile_ <<"Model of substitution: General time reversible"  <<  endl;         
        double tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_;
        Model::getGenPam (tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_);  

        outFile_.precision (5);
        outFile_ <<"    3. Transversion rate from A to C: " << tvAC_ << endl;
        outFile_ <<"    1. Transition   rate from A to G: " << tsAG_ << endl;
        outFile_ <<"    4. Transversion rate from A to T: " << tvAT_ << endl;
        outFile_ <<"    5. Transversion rate from C to G: " << tvCG_ << endl;
        outFile_ <<"    2. Transition   rate from C to T: " << tsAG_ << endl;
        outFile_ <<"    6. Transversion rate from G to T: " << tvGT_ << endl;
    }
    
     double baseFrqA = Model::getStateFrq (BASE_A);
     double baseFrqC = Model::getStateFrq (BASE_C);
     double baseFrqG = Model::getStateFrq (BASE_G);
     double baseFrqT = Model::getStateFrq (BASE_T);


     if (Model::getBaseFrqType () == ESTIMATE) 
       outFile_ << "Nucleotide Frequencies (estimate from the data set):" << endl;
     else
       outFile_ << "Nucleotide Frequencies (user defined):" << endl;
      outFile_.precision (4);
      outFile_ << "    1. Base A: " << baseFrqA << endl;
      outFile_ << "    2. Base C: " << baseFrqC << endl;
      outFile_ << "    3. Base G: " << baseFrqG << endl;
      outFile_ << "    4. Base T: " << baseFrqT << endl;
  }//end of writing model infomration


  outFile_ << endl << "Rate Heterogeneity: ";

  RATE_TYPE rateType_;
  rateType_ = Rate::getType ();
  if (rateType_ == UNIFORM)
    outFile_ << "Uniform substitution rate" << endl;
  else 
    if (rateType_ == SITE_SPECIFIC) {
       outFile_ << "Site-specific substitution rates (see " << rateFileName_ << ")" << endl;
       writeSpecificRate (rateFileName_);
    }  else {
      int nRate_ = Rate::getNRate ();
     double gammaShape_ = Rate::getGammaShape ();

      outFile_ << "Gamma distributed rates" << endl;
      outFile_ << "Gamma distribution parameter alpha: ";

      outFile_.precision (5);
      outFile_ << gammaShape_ << endl;
      outFile_ << "Number of Gamma rate categogies:  " << nRate_ << endl;    
    }

  
  outFile_ <<"================================================================================="<< endl;
  outFile_ <<"                              RESULTS INFORMATION                                                                         " << endl;
  outFile_ <<"================================================================================="<< endl << endl;


  outFile_.precision (10);
  outFile_ << "Log likelihood: " << tree.getLogLi () << endl;

  outFile_ << "The tree with highest likelihood: " << endl;
  tree.writeNewickForm (outFile_);

  outFile_ << endl << "The illustrated picture from the reconstructed tree with highest likelihood:" << endl;
  outFile_ << "Note: it is an unrooted tree!!!" << endl;

   Mat<char> image_;
  tree.createImage (BR_LEN, outGrpNdNo_, image_);
  OutStream::write (image_, outFile_);             
  OutStream::write (tree.getBr (), outFile_);

  outFile_ << endl << "The majority consensus tree from all intermediate trees:" << endl;
  OutStream::write (conTree, outFile_) << endl << endl;

  outFile_ << "Time information:" << endl;  
  outFile_ << "Date and time: " << startedDate;  
  int hour_ = static_cast<int> (progTime /3600.0);
  int min_ =  static_cast<int> ( (progTime - hour_ * 3600)/60.0);
  int sec_ =  static_cast<int> ( progTime - hour_ * 3600 - min_ * 60);

  outFile_ << "Runtime: " << hour_ << "h:" << min_ << "m:" << sec_ << "s" << endl;
  outFile_.close ();
   
  tree.writeTop(topoFileName_, 0);  
  std::cout << "The results were written to following files: " << endl;
  std::cout << "   1. " << outFileName_ << endl;
  std::cout << "   2. " <<  topoFileName_ << endl;
  std::cout << "   3. " << treeLsFileName_ << endl;
  if (Rate::getType () == SITE_SPECIFIC)
     std::cout << "   4. " << rateFileName_ << endl;

}



private:
int nSeq_;
};
#endif
    
