#include "iqp.h"
#include <cstring>
#include <math.h>
#include <time.h>

#include "constant.h"
#include "model.h"
#include "ali.h"
#include "utl.h"
#include "outstream.h"
#include "usertree.h"
#include "br.h"
#include "treels.h"
#include "bionj.h"
#include "clusterarr.h"
#include "rate.h"
#include "prediction.h"
#include "rannum.h"


using namespace std;
extern int isContinuous;

//--------------------------------------------------------------------
//we will init all things here before starting testing this method with data
void IQP::init () {
  nSeq_ = Ali::getNSeq ();
}


//=================================================================
double IQP::escape (Vec<int> &orderNdNoLs, UrTree &tree, UrTree &bestTree, ClusterArr &bestClusterArr) {
//  std::cout << "escaping..." << endl;
  int esArr_[MAX_NUM_SEQ];
  int seqNo_;
  for (seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
    int time_ = clock ();
    if (time_ < 0)
      time_ = -time_;    
    int timeFac_ = time_ % 100;
    int ranFac_ = rand () % 100;
    int ranNum_ = (timeFac_ * ranFac_) % 100;   
    esArr_[seqNo_] = ( ranNum_ < proSel_ * 100);
  }

  int nSeqDeled_ = 0;
  for (seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    if (esArr_[seqNo_] == 1 && nSeq_) {         
      tree.delExNd (seqNo_);
      nSeqDeled_ ++;

      if (nSeqDeled_ + 2 == nSeq_) {
        for (int count_ = seqNo_ + 1; count_ < nSeq_; count_ ++)
          esArr_[count_] = 0;
        break;
      }

    }

  for (int count_ = 0; count_ < nSeq_; count_ ++) {
    seqNo_ = orderNdNoLs[count_];
    if (esArr_[seqNo_] == 1)
      tree.insertNd (seqNo_, nRep_);
 }

 
  tree.reCreateRootedTree (orderNdNoLs[0]);
  double logLi_ = tree.reCmpLogLi (COMPLETE, bestTree, bestClusterArr, 1);
//  double logLi_ = -INFINITIVE;
//  double logLi_ = tree.reCmpLogLi (COMPLETE,  1);
 return logLi_;
}

//****************************************************************************************************************************************************************
//****************************************************************************************************************************************************************
//****************************************************************************************************************************************************************
void IQP::optPam () {
  int nChanged_;

   UrTree bestTree_;
   bestTree_.getOptUrTree (); 

  int nIter_ = 0;
  while (nIter_ < MAX_IT_ALL_PAM_OPT) {
    nIter_ ++;

    nChanged_ = 0;

    double oldGammaShape_ = Rate::getGammaShape ();

    if (Rate::getType () == GAMMA && Rate::getGammaShapeType () == ESTIMATE) {
       Rate::optRatePam ();
      double newGammaShape_ = Rate::getGammaShape ();
      if (fabs (newGammaShape_ - oldGammaShape_) > EPS_GAMMA_SHAPE_ERROR)
        nChanged_ ++;
    }

    int isChangedPam_;
    if (Rate::getType () == GAMMA && Rate::getGammaShapeType () == ESTIMATE 
        && nChanged_ == 1)
      isChangedPam_ = Model::optPam (1);
    else
      isChangedPam_ = Model::optPam ();

    nChanged_ += isChangedPam_;
    
    if  (nChanged_ <= 1)
      break;
  } 
}


//****************************************************************************************************************************************************************
//****************************************************************************************************************************************************************
//****************************************************************************************************************************************************************
double IQP::tuneFineParameter (UrTree &bestTree) {
  bestTree.createOptUrTree();
  if (Model::isEstedPam () == 1 ||  
     (Rate::getType () == GAMMA && Rate::getGammaShapeType () == ESTIMATE) || 
     Rate::getType () == SITE_SPECIFIC) {  
     
     std::cout <<"Re-Optimizing parameters..." << endl;   optPam ();
     optPam ();
     Ali::cmpGenDis (Ali::fileName_);
  }

  double bestLogLi_ = OptUrTree::cmpLogLi ();
  if (Rate::getType () == SITE_SPECIFIC)  {
     int nIter_ = 0;
      do {
         nIter_ ++;
         ClusterArr bestClusterArr_;
         bestTree.createCluster(bestClusterArr_);

         Vec<SeqPair> indSeqPairArr_;
         bestTree.createIndSeqPair (indSeqPairArr_);
     
//         bestTree.draw(NON_BR_LEN, 0);
//         OutStream::write (indSeqPairArr_, std::cout);

         if (nIter_ > 1)
            std::cout <<"Re-";
 
         int isChanged_ = Rate::optSpecificSite (indSeqPairArr_);
         if (isChanged_ == 0)
            break;

         std::cout <<"Optimizing the tree topology as well as branch lengths using site-specific substitution rates... " << endl;
         bestLogLi_ = bestTree.doConOpt (1);

         int topDis_ = bestTree.cmpDis(bestClusterArr_) ;
         if (topDis_ == 0)
            break;
      } while (nIter_ < MAX_IT_TREE_SITE_SPECIFIC_RATE_OPT);
     

  }

  std::cout <<"Optimizing the final tree topology as well as branch lengths..." << endl;
  bestLogLi_ = bestTree.doConOpt(1);
  std::cout.precision (10);
  std::cout <<"Final best log likelihood: " << bestLogLi_ << endl;   
  return bestLogLi_;
}

//****************************************************************************************************************************************************************
//load all already reconstructed intermediate trees from the file into the memory
void IQP::loadTreeLs (const char *treeLsFileName, int &nIter, ClusterArr &bestClusterArr, ClusterArr &allCluArr, 
                                          UrTree &bestTree, double &bestLogLi,
                                          int &nImpTree, const char *predictionFileName,
                                          double &progTime) {

   ifstream treeLsFile_;
   treeLsFile_.open (treeLsFileName);
   ofstream predictionFile_;
   predictionFile_.open (predictionFileName);
   
   bestLogLi = -INFINITIVE;
   nIter = 0;
   nImpTree = 0;

   while (!treeLsFile_.eof ()) {
       char aLine_[MAX_LINE_LEN];

       double tmpProgTime_;
       treeLsFile_.getline (aLine_, MAX_LINE_LEN - 1);
       char *endPos_;
       tmpProgTime_ = strtod (aLine_, &endPos_);

       double curLogLi_;
       treeLsFile_.getline (aLine_, MAX_LINE_LEN - 1);
       curLogLi_ = strtod (aLine_, &endPos_);


       Vec<char> newickTree_;
       treeLsFile_.getline (aLine_, MAX_LINE_LEN - 1);
       newickTree_.set (aLine_);
       
       if (newickTree_.getSize () > 0) {  
          UserTree curTree_;
          curTree_.setNewickTree (newickTree_);
          int status_ = curTree_.createUrTree ();
          if (status_ == NEWICK_TREE_FORM)  {
             nIter ++;
             progTime = tmpProgTime_;

             ClusterArr newClusterArr_;
             curTree_.createCluster(newClusterArr_);
             for (int count_ = 0; count_ < newClusterArr_.totalNCluster_; count_ ++) 
                allCluArr.addConTree(newClusterArr_[count_]);
 
             if (curLogLi_ > bestLogLi + ZERO)  {
                bestLogLi = curLogLi_;
                bestTree = curTree_;
                if (bestTree.cmpDis( bestClusterArr) != 0) {
                   nImpTree ++;
                   predictionFile_ << nIter << endl;
                   bestTree.createCluster (bestClusterArr);
                }
             } //end of if curLogLi_
           } //end of if status
       }  //end of if newickTree

    }  // end of while
 
   treeLsFile_.close ();
   predictionFile_.close ();
}


//****************************************************************************************************************************************************************
void IQP::writePamCheckPoint ( char  *checkPointFileName, 
                        int        nIter,       
                        double     proSel,
                        int        nRep,
                        STOPPING_TYPE     stoppingRule,
                        int         outGrpSeqNo,
                        int         isFinished) {

   ofstream checkPointFile_;
   checkPointFile_.open (checkPointFileName);

   checkPointFile_ << "Number interations: " << nIter << endl;
   checkPointFile_ << "Probability of deleting a sequence: " << proSel << endl;
   checkPointFile_ << "Number representatives: " << nRep - 1<< endl;
   checkPointFile_ << "Stopping rule (0: YES, 1: YES_MIN_ITER, 2: YES_MAX_ITER, 3: NO): " << stoppingRule << endl;

   checkPointFile_ << "Type of data (0:NUCLEOTIDE, 1:AMINO_ACID): " << Model::getDataType () << endl;


   checkPointFile_ << "Substitution model (0:HKY85, 1: TN93, 2:GTR, 3:WAG, 4:JTT, 5:VT, 6:MtREV24, 7:Blosum62, 8:Dayhoff): " << Model::getType () << endl;
   DVec20 baseFrqArr_;
   Model::getStateFrq (baseFrqArr_);
   checkPointFile_ << "Frequency of Base A: " << baseFrqArr_[BASE_A] << endl;
   checkPointFile_ << "Frequency of Base C: " << baseFrqArr_[BASE_C] << endl;
   checkPointFile_ << "Frequency of Base G: " << baseFrqArr_[BASE_G] << endl;
   checkPointFile_ << "Frequency of Base T: " << baseFrqArr_[BASE_T] << endl;
   checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED, 2: EQUAL): " << Model::getBaseFrqType () << endl;
   

   checkPointFile_ << "Transition/transversion ratito: " << Model::getTsTvRatio () / 2.0 << endl;
   checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << Model::getTsTvRatioType () << endl;

   checkPointFile_ << "Pyridimine/purine ratito: " << Model::getPyPuRatio () << endl;
   checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << Model::getPyPuRatioType () << endl;

   double tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_;
   Model::getGenPam (tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_);

   checkPointFile_ << "Transition rate from A to G: " << tsAG_ << endl;
   checkPointFile_ << "Transition rate from C to T: " << tsCT_ << endl;

   checkPointFile_ << "Transversion rate from A to C: " << tvAC_ << endl;
   checkPointFile_ << "Transversion rate from A to T: " << tvAT_ << endl;
   checkPointFile_ << "Transversion rate from C to G: " << tvCG_ << endl;
   checkPointFile_ << "Transversion rate from G to T: " << tvGT_ << endl;
   checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << Model::getPyPuRatioType () << endl;

   RATE_TYPE rateType_ = Rate::getType ();
   checkPointFile_ << "Type of rate heterogeneity (0:UNIFORM, 1:SITE_SPECIFIC, 2:GAMMA): " << rateType_ << endl;
   checkPointFile_ << "Number rates: " << Rate::getNRate () << endl;
   
   checkPointFile_ << "Gamma distribution parameter alpha: " << Rate::getGammaShape () << endl ;
   checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << Rate::getGammaShapeType () << endl;

   checkPointFile_ << "Out group sequence: " << outGrpSeqNo << endl;
   
   checkPointFile_ << "Finished: " << isFinished << endl;

   checkPointFile_.close ();
}

//****************************************************************************************************************************************************************
//****************************************************************************************************************************************************************
double IQP::escapeLocalOpt (int &nIter, double proSel, int nRep, STOPPING_TYPE stoppingRule, UrTree &bestTree,  int outGrpSeqNo, Vec<char> &conTree, 
                            time_t &beginTime, double &progTime) {
  char predictionFileName_[FILENAME_LEN];
  char treeLsFileName_[FILENAME_LEN];

  char bionjFileName_[FILENAME_LEN];
  char distFileName_[FILENAME_LEN];
  char checkPointFileName_[FILENAME_LEN];

  strcpy (predictionFileName_, Ali::fileName_);
  strcpy (treeLsFileName_, Ali::fileName_);
  strcpy (bionjFileName_, Ali::fileName_);
  strcpy (distFileName_, Ali::fileName_);
  strcpy (checkPointFileName_, Ali::fileName_);

  char predictionTail_[FILENAME_LEN] = ".iqpnni.prediction";
  char treeLsTail_[FILENAME_LEN] = ".iqpnni.treels";
  char bionjTail_[FILENAME_LEN] = ".iqpnni.bionj";
  char distTail_[FILENAME_LEN] = ".iqpnni.dist";

  char checkPointTail_[FILENAME_LEN] = ".iqpnni.checkpoint";

  strcat (predictionFileName_, predictionTail_);
  strcat (treeLsFileName_, treeLsTail_);
  strcat (bionjFileName_, bionjTail_);
  strcat (distFileName_, distTail_);
  strcat (checkPointFileName_, checkPointTail_);

  nSeq_ = Ali::getNSeq ();
  nRep_ = nRep;
  proSel_ = proSel;
  
  int nSite_ = Ali::getNSite (); 
  std::cout <<"number sequences = " << nSeq_ << "; sequence length = " << nSite_ << endl;

  ClusterArr allCluArr_;
  allCluArr_.setupIndex ();

  int nPredictedIter_ = MAX_ITERATION;
  int nImpTree_ = 1;
  int iterNo_ = 1;
  double bestLogLi_ = -INFINITIVE;

  Prediction prediction_;

  double stoppedProgTime_ = 0.0;
  int maxNIter_ = nIter;
  
  if (isContinuous == 0) { 
     BioNj bioNj_;
     bioNj_.create (distFileName_, bionjFileName_);
     UserTree bionjTree_;
     bionjTree_.readFile (bionjFileName_);
  

     bionjTree_.createUrTree ();
//   bionjTree_.draw (BR_LEN, 0);
     double bionjLogLi_ =  bionjTree_.doConOpt (0, 1);
     //  std::cout << bionjLogLi_ << endl; 
  
     if ( Model::isEstedPam () == 1 || 
        (Rate::getType () == GAMMA && Rate::getGammaShapeType () == ESTIMATE) ) {  
         optPam ();
         Ali::cmpGenDis (Ali::fileName_);
     }
 
      bionjLogLi_ = bionjTree_.doConOpt (1);  

       bestTree = bionjTree_;
       bestLogLi_ = bionjLogLi_; 
       bestTree.createCluster (bestClusterArr_);

       for (int clusterNo_ = 0; clusterNo_ < bestClusterArr_.totalNCluster_; clusterNo_ ++)
          allCluArr_.addConTree(bestClusterArr_[clusterNo_]);
     
      time_t currentTime_;  
      time(&currentTime_);


      progTime = difftime (currentTime_, beginTime);

      bestTree.writeNewickForm (treeLsFileName_, 0,  progTime);

       std::cout << "We have constructed the initial tree !!!" << endl;
       std::cout.precision (10);
       std::cout << "The currently best log likelihood = " << bestLogLi_  << endl; 

       std::ofstream predictionFile_;
       predictionFile_.open (predictionFileName_);
       predictionFile_ << 1 << endl;
       predictionFile_.close ();
       int isFinished_ = 0;
       writePamCheckPoint(checkPointFileName_,                         
                          nIter, proSel, nRep, stoppingRule, outGrpSeqNo, 
                          isFinished_); 

   } else {//if isContinuous == 0
       std::cout <<"Loading the data from the last stopped point ...";
       loadTreeLs (treeLsFileName_, iterNo_, bestClusterArr_, allCluArr_, bestTree, bestLogLi_, nImpTree_, predictionFileName_, stoppedProgTime_);
 
       if (nImpTree_ >= START_PREDICTION && stoppingRule != NO) {
           double tmpNPreIter_;
           prediction_.predict(predictionFileName_, 0.95, tmpNPreIter_);
           nPredictedIter_ = static_cast<int> (tmpNPreIter_);
           if (nPredictedIter_ <= iterNo_)  
             nPredictedIter_ = MAX_ITERATION;      
          else {
             if (stoppingRule == YES)
               maxNIter_ = nPredictedIter_;
             if (stoppingRule == YES_MIN_ITER)
               maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
             if (stoppingRule == YES_MAX_ITER) 
               maxNIter_ = Utl::getMin (nIter, nPredictedIter_);
          }
       }

       std::cout << endl << "Continuing from the last stopped point ..." << endl;
  }

  double lastProgTime_ = 0.0;  
//************************************************************************************************************************************************
//************************************************************************************************************************************************
//************************************************************************************************************************************************
//************************************************************************************************************************************************
//  start IQPNNI
  while (iterNo_ < maxNIter_) {
     iterNo_ ++;

     Vec<int> orderNdNoLs_;
     RanNum::createOrderNd (nSeq_, orderNdNoLs_);
     UrTree curTree_;
     curTree_ = bestTree;
     double curLogLi_ = escape (orderNdNoLs_, curTree_, bestTree, bestClusterArr_);


     ClusterArr newClusterArr_;
     curTree_.createCluster(newClusterArr_);
     for (int count_ = 0; count_ < newClusterArr_.totalNCluster_; count_ ++) 
        allCluArr_.addConTree(newClusterArr_[count_]);
     

     time_t currentTime_;  
     time(&currentTime_);
     progTime = difftime (currentTime_, beginTime) + stoppedProgTime_;

      int nHour_ = static_cast<int> (progTime / 3600);
      int nMin_ = static_cast<int> ( (progTime - nHour_ * 3600) / 60);
      int nSec_ = static_cast<int> (progTime - nMin_ * 60 - nHour_ * 3600);

     curTree_.writeNewickForm (treeLsFileName_, 1, progTime);    

    if (curLogLi_ > bestLogLi_) {
      if (curTree_.cmpDis( bestClusterArr_) != 0) {
         ofstream predictionFile_;
         predictionFile_.open (predictionFileName_, ios::app);
         predictionFile_ << iterNo_ << endl;
         predictionFile_.close ();

         nImpTree_ ++;
         if (nImpTree_ >= START_PREDICTION && stoppingRule != NO) {
           double tmpNPreIter_;
           prediction_.predict(predictionFileName_, 0.95, tmpNPreIter_);
           if (static_cast<int> (tmpNPreIter_) > iterNo_) {
              nPredictedIter_ = static_cast<int> (tmpNPreIter_);
              if (stoppingRule == YES)
                maxNIter_ = nPredictedIter_;
              if (stoppingRule == YES_MIN_ITER)
                maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
              if (stoppingRule == YES_MAX_ITER) 
                maxNIter_ = Utl::getMin (nIter, nPredictedIter_);
           }  // end of if tmpNPreIter_ > iterNo
         }   
   

         bestLogLi_ = curLogLi_;
         bestTree = curTree_;
         bestTree.createCluster (bestClusterArr_);

         std::cout.precision (10);

         std::cout <<"GOOD NEWS: BETTER TREE FOUND: THE CURRENTLY BEST LOG LIKELIHOOD = " << bestLogLi_ << endl;
         if (nImpTree_ >= START_PREDICTION && nPredictedIter_ < MAX_ITERATION) {
           double neededTime_ = ( progTime / iterNo_) * (nPredictedIter_ - iterNo_);
           int nNeededHour_ = static_cast<int> (neededTime_ / 3600);
           int nNeededMin_ = static_cast<int> ((neededTime_ - nNeededHour_ * 3600)/ 60);
           int nNeededSec_ = static_cast<int> (neededTime_ - nNeededMin_ * 60 - nNeededHour_ * 3600);
           std::cout <<"PREDICTION: WE WILL RUN ANOTHER " << 
             (nPredictedIter_ - iterNo_) << " ITERATIONS (ABOUT " << nNeededHour_ <<"h:" << nNeededMin_ <<"m:"<<nNeededSec_<<"s)" <<endl;          
         }
      }
    }
    
    

    int remNHour_ = 0;
    int remNMin_ = 0;
    int remNSec_ = 1;    
    if (maxNIter_ > iterNo_) {
       int remTime_ = static_cast<int> ( (progTime / iterNo_) * (maxNIter_ - iterNo_));
       remNHour_ = static_cast<int> (remTime_ / 3600);
       remNMin_ = static_cast<int> ( (remTime_ - remNHour_ * 3600) / 60);
       remNSec_ = static_cast<int> (remTime_ - remNHour_ * 3600 - remNMin_ * 60);
    }

    if (progTime -  lastProgTime_ > 10.0) {
      std::cout << iterNo_ << " Iterations / time elapsed = " << nHour_ <<"h:" << nMin_<<"m:" << nSec_<<"s";
      std::cout << "  (will finish in " << remNHour_<<"h:" << remNMin_ <<"m:" << remNSec_ <<"s)" << endl; 
      lastProgTime_ = progTime; 
    }
//    char answer_;
    if (iterNo_ >= maxNIter_) {      
       std::cout << "WE HAVE FOUND THE BEST TREE WITH 95% CONFIDENCE" << endl;
       break;
    }

    if (iterNo_ >= nIter && nPredictedIter_ == MAX_ITERATION) {      
       std::cout << "WE HAVE CREATED " << iterNo_ << " TREES" << endl;
       break;
    } 
  } //end of while
                                                                                              
   nIter = iterNo_;  
   bestLogLi_ = tuneFineParameter (bestTree);

   std::cout <<"Constructing the majoriy rule consensus tree..." << endl;
   allCluArr_.createConTree(nIter, conTree);  
   int isFinished_ = 1;
   writePamCheckPoint(checkPointFileName_,                         
                                  nIter, proSel, nRep, stoppingRule, outGrpSeqNo, isFinished_); 
   time_t currentTime_;  
   time(&currentTime_);
   progTime = difftime (currentTime_, beginTime) + stoppedProgTime_;
   return bestLogLi_;
}
