/***************************************************************************
                          libr.cpp  -  description
                             -------------------
    begin                : Thu Apr 24 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <math.h>
#include <time.h>

#include "constant.h"
#include "libr.h"
#include "ptnls.h"
#include "rate.h"
#include "model.h"
#include "opturtree.h"
#include "dmat20.h"
#include "ptnratecube.h"

extern int nto[NUM_CHAR];

//the constructor
LiBr::LiBr () {
  smaLiNd_ = new LiNd;
  greLiNd_ = new LiNd;
}


//======================================================
//copy the state br into this likelihood branch
void LiBr::copyBase (Br<double> &br) {
  this->id_ = br.getId ();
  this->smaNdNo_ = br.getSmaNd ();
  this->greNdNo_ = br.getGreNd ();
  this->isEx_ = br.isEx ();
  this->len_ = br.getLen ();
  this->isStG_ = br.isStG ();

  greLiNd_->set (greNdNo_, IN, smaNdNo_, id_);  
  if (isEx_ == 1)
    smaLiNd_->set (smaNdNo_, EX, greNdNo_, id_);
  else
    smaLiNd_->set (smaNdNo_, IN, greNdNo_, id_);
}


//======================================================
int LiBr::isOutGrp () {
  if (id_ == OptUrTree::outGrpBrNo_ )
    return 1;
  else 
    return 0;
}







//======================================================
//get the remaining liNd of this branch
LiNd &LiBr::getRemLiNd (int existNdNo) {
  if (existNdNo != smaNdNo_)
    return *smaLiNd_;
  else
    return *greLiNd_;
}

//-------------------------------------------------------------------
//get the sma liNd of this branch
LiNd &LiBr::getSmaLiNd () {
  return *smaLiNd_;
}

//-------------------------------------------------------------------
//get the gre liNd of this branch
LiNd &LiBr::getGreLiNd () {
  return *greLiNd_;
}


void LiBr::writeInf () {
  std::cout << id_ << " / " << smaNdNo_ << " / " << greNdNo_ << " / " << isStG_ << endl;
  getHeadLiNd ().writeInf ();
  getTailLiNd ().writeInf ();
}

void LiBr::setTailLiNd (LiNd *newTailLiNd) {
  if (isStG_ == 1) 
    greLiNd_ = newTailLiNd;
  else
    smaLiNd_ = newTailLiNd;
  int tailNdNo_ = newTailLiNd->getId ();

  isEx_ = isExNd (tailNdNo_);

  if (smaLiNd_->getId () > greLiNd_->getId () ) 
    Utl::swap (smaLiNd_, greLiNd_);
  smaNdNo_ = smaLiNd_->getId ();
  greNdNo_ = greLiNd_->getId ();
  if (greNdNo_ == tailNdNo_)
    isStG_ = 1;
  else
    isStG_ = 0;
  smaLiNd_->setParNdBr (greNdNo_, id_);
  greLiNd_->setParNdBr (smaNdNo_, id_);
}
//-------------------------------------------------------------------
//get the tail liNd of this branch
LiNd &LiBr::getTailLiNd () {

  if (isStG_ == 1)
    return *greLiNd_;
  else
    return *smaLiNd_;
}

//-------------------------------------------------------------------
//get the head liNd of this branch
LiNd &LiBr::getHeadLiNd () {
  if (isStG_ == 1)
    return *smaLiNd_;
  else
    return *greLiNd_;
}



//-------------------------------------------------------------------
LiNd &LiBr::getLiNd (int ndNo) {
  if (ndNo == smaNdNo_)
    return *smaLiNd_;
  else
    return *greLiNd_;
}

//-------------------------------------------------------------------
//compute the likelihood for liNd of this branch
void LiBr::cmpLiNd (int ndNo) {

  if (ndNo == greNdNo_)
    greLiNd_->cmpLi ();

  if (ndNo == smaNdNo_)
    smaLiNd_->cmpLi ();
}

//-------------------------------------------------------------------
//compute the likelihood for tail liNds of this branch
void LiBr::cmpTailLiNd () {
  if (getTailNd () == greNdNo_) //the case, the tail node is greNdNo_
    greLiNd_->cmpLi ();
  else // the case the tail node is small node
    smaLiNd_->cmpLi ();
}

//-------------------------------------------------------------------

//compute the likelihood for head liNds of this branch
void LiBr::cmpHeadLiNd () {
  if (getHeadNd () == greNdNo_) //the case, the head node is greNdNo_
    greLiNd_->cmpLi ();
  else // the case the head node is small node
    smaLiNd_->cmpLi ();

}

//-------------------------------------------------------------------
//reCompute the likelihood for tail liNds of this branch
void LiBr::reCmpTailLiNd () {
  if (getTailNd () == greNdNo_) //the case, the tail node is greNdNo_
    greLiNd_->reCmpLi ();
  else 
    smaLiNd_->reCmpLi ();
}

//-------------------------------------------------------------------
//reCompute the likelihood for head liNds of this branch
void LiBr::reCmpHeadLiNd () {
  if (getHeadNd () == greNdNo_) //the case, the head node is greNdNo_
    greLiNd_->reCmpLi ();
  else
    smaLiNd_->reCmpLi ();
}

//create the external descendant for all liNds of this branch
void LiBr::createExDesLiNd () {  
  greLiNd_->createExDesNd ();
  smaLiNd_->createExDesNd ();
}

inline double cmpLogLi (long double li) {
  double logLi_ = 0.0;
  while (li < MIN_EXP_VAL) {
    li /= MIN_EXP_VAL;
    logLi_ -= MIN_EXP;
  }
  
  logLi_ += log (li);
  return logLi_;
}


//--------------------------------------------------------------------
//compute the likelihood of the tree, the case this branch is an internal branch
double LiBr::cmpInLogLiUniformRate () {
  OptUrTree::nCmpInLogLi_ ++;

  LiNd *headLi_;
  LiNd *tailLi_;

  headLi_ = &getHeadLiNd ();
  tailLi_ = &getTailLiNd ();
     
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  DMat20 proChangeMat_;
  Model::cmpProbChange (len_, proChangeMat_);
  int nState_ = Model::getNState ();

  double inLogLi_ = 0.0;
  int nPtn_ = PtnLs::getNPtn ();
  for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
     long double ptnLi_ = 0.0;
     for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

        long double stateLi_ = 0.0;
        for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) 
           stateLi_ += headLi_->liNdCube_[ptnNo_][0][headStateNo_] * 
                             proChangeMat_[headStateNo_][tailStateNo_] *
                             tailLi_->liNdCube_[ptnNo_][0][tailStateNo_];
        
        ptnLi_ += stateLi_ * stateFrqArr_[headStateNo_];
    }

    if (ptnLi_ <= 0.0)
      Utl::announceError (BAD_CMP_LI);

     double logPtnLi_ = cmpLogLi (ptnLi_);
     inLogLi_ += logPtnLi_ * PtnLs::weightArr_[ptnNo_];
  }

  return inLogLi_;
}



//--------------------------------------------------------------------
//compute the likelihood of the tree, the case this branch is an internal branch
double LiBr::cmpInLogLiSpecificRate () {
  OptUrTree::nCmpInLogLi_ ++;

  LiNd *headLi_;
  LiNd *tailLi_;

  headLi_ = &getHeadLiNd ();
  tailLi_ = &getTailLiNd ();
     
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  int nPtn_ = PtnLs::getNPtn ();    

  Vec<double> *ptnRateArr_;
  ptnRateArr_ = Rate::getPtnRate ();

  int nState_ = Model::getNState ();

  double inLogLi_ = 0.0;
  for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

      double ptnRate_ = (*ptnRateArr_)[ptnNo_];
      DMat20 probChangePtnMat_;
      Model::cmpProbChange (len_ * ptnRate_, probChangePtnMat_);

     long double ptnLi_ = 0.0;
     for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

        long double stateLi_ = 0.0;
        for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) 
           stateLi_ += headLi_->liNdCube_[ptnNo_][0][headStateNo_] * 
                             probChangePtnMat_[headStateNo_][tailStateNo_] *
                             tailLi_->liNdCube_[ptnNo_][0][tailStateNo_];
        
        ptnLi_ += stateLi_ * stateFrqArr_[headStateNo_];
    }

    if (ptnLi_ <= 0.0)
      Utl::announceError (BAD_CMP_LI);

     double logPtnLi_ = cmpLogLi (ptnLi_);
     inLogLi_ += logPtnLi_ * PtnLs::weightArr_[ptnNo_];
  }

  return inLogLi_;
}

//--------------------------------------------------------------------
//compute the likelihood of the tree, the case this branch is an internal branch
double LiBr::cmpInLogLiGammaRate () {  
  OptUrTree::nCmpInLogLi_ ++;

  LiNd *headLi_;
  LiNd *tailLi_;

  headLi_ = &getHeadLiNd ();
  tailLi_ = &getTailLiNd ();
     
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  int nState_ = Model::getNState ();
  int nRate_ = Rate::getNRate ();
  double probRate_ = Rate::getProb ();
  
  DMat20 probChangeCube_[MAX_NUM_RATE];
  int rateNo_;
  for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
    double rate_ = Rate::getRate (rateNo_);
    Model::cmpProbChange (len_ * rate_, probChangeCube_[rateNo_]);
  }

  double inLogLi_ = 0.0;
  for (int ptnNo_ = 0; ptnNo_ < PtnLs::getNPtn (); ptnNo_ ++) {
    long double ptnLi_ = 0.0;

    for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
    long double ratePtnLi_ = 0.0;

    for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

      long double stateLi_ = 0.0;
        for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) 
          stateLi_ += headLi_->liNdCube_[ptnNo_][rateNo_][headStateNo_] * 
                  probChangeCube_[rateNo_][headStateNo_][tailStateNo_] *
                  tailLi_->liNdCube_[ptnNo_][rateNo_][tailStateNo_];

        ratePtnLi_ += stateLi_ * stateFrqArr_[headStateNo_];
      }

       ptnLi_ += ratePtnLi_ * probRate_;
    }
      
    if (ptnLi_ <= 0.0)
    Utl::announceError (BAD_CMP_LI);
  
  double logPtnLi_ = cmpLogLi (ptnLi_);
  inLogLi_ += logPtnLi_ * PtnLs::weightArr_[ptnNo_];
  }

  return inLogLi_;
}


//-------------------------------------------------------------------
//compute the log likelihood of this tree via this branch
double LiBr::cmpExLogLiUniformRate () {
  Cube *inLiCube_;   
  inLiCube_ = &greLiNd_->liNdCube_;
  
  Cube *exLiCube_;   
  exLiCube_ = &smaLiNd_->liNdCube_;

  Seq *seq_;
  seq_ = &Ali::getSeq (smaNdNo_);
  
  int nState_ = Model::getNState ();
  
  DMat20 probChangeMat_;
  Model::cmpProbChange (len_, probChangeMat_);


  double logLi_ = 0.0;
  int nPtn_ = PtnLs::getNPtn ();    
  
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
       int siteNo_ = PtnLs::siteArr_ [ptnNo_];
       int exStateNo_ = nto [ static_cast<int> ((*seq_).items_[siteNo_]) ];
       long double exPtnLi_ = 0.0;

    if (exStateNo_ < nState_) {    
       for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) 
          exPtnLi_ += (*inLiCube_)[ptnNo_][0][inStateNo_] * 
                  probChangeMat_[exStateNo_][inStateNo_];
                        
       exPtnLi_ *= stateFrqArr_[exStateNo_];
    } else {
     for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

         double exStateLi_ = 0.0;
       if (exStateNo_ == BS_UNKNOWN)
         exStateLi_ = 1.0;
       else
         for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) 
           exStateLi_ += probChangeMat_[inStateNo_][exStateNo_] * 
                       (*exLiCube_)[ptnNo_][0][exStateNo_];
         
           exPtnLi_ += (*inLiCube_)[ptnNo_][0][inStateNo_] * 
                     stateFrqArr_[inStateNo_] * exStateLi_;
       }
    }  
  
  if (exPtnLi_ <= 0.0) {
    Utl::announceError (BAD_CMP_LI);
  }

    double logExPtnLi_ = cmpLogLi (exPtnLi_);
    logLi_ += logExPtnLi_ * PtnLs::weightArr_[ptnNo_];
  //std::cout << ptnNo_ << " / " << logLi_ << endl;
  }

  return logLi_;
}

//-------------------------------------------------------------------
//compute the log likelihood of this tree via this branch
double LiBr::cmpExLogLiSpecificRate () {
  Cube *inLiCube_;   
  inLiCube_ = &greLiNd_->liNdCube_;
  
  Cube *exLiCube_;   
  exLiCube_ = &smaLiNd_->liNdCube_;

  Seq *seq_;
  seq_ = &Ali::getSeq (smaNdNo_);
  
  int nState_ = Model::getNState ();
  
  int nPtn_ = PtnLs::getNPtn ();    

  Vec<double> *ptnRateArr_;
  ptnRateArr_ = Rate::getPtnRate ();

  double logLi_ = 0.0;

  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);


  for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

    int siteNo_ = PtnLs::siteArr_ [ptnNo_];
    int exStateNo_ = nto [ static_cast<int> ((*seq_).items_[siteNo_]) ];
    long double exPtnLi_ = 0.0;

    if (exStateNo_ < nState_) {    

      double ptnRate_ = (*ptnRateArr_)[ptnNo_];
      DVec20 probChangePtnVec_;
      Model::cmpProbChange (exStateNo_, len_ * ptnRate_, probChangePtnVec_);

      for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
          exPtnLi_ += (*inLiCube_)[ptnNo_][0][inStateNo_] * 
                             probChangePtnVec_[inStateNo_];
      }
                        
       exPtnLi_ *= stateFrqArr_[exStateNo_];
    } else {
      double ptnRate_ = (*ptnRateArr_)[ptnNo_];
      DMat20 probChangePtnMat_;
      Model::cmpProbChange (len_ * ptnRate_, probChangePtnMat_);

      for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

         double exStateLi_ = 0.0;
         if (exStateNo_ == BS_UNKNOWN)
           exStateLi_ = 1.0;
         else
           for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) 
             exStateLi_ += probChangePtnMat_[inStateNo_][exStateNo_] * 
                                (*exLiCube_)[ptnNo_][0][exStateNo_];
         
           exPtnLi_ += (*inLiCube_)[ptnNo_][0][inStateNo_] * 
                              stateFrqArr_[inStateNo_] * exStateLi_;
      }
    }

    if (exPtnLi_ <= 0.0) {
       Utl::announceError (BAD_CMP_LI);
    }


    double logExPtnLi_ = cmpLogLi (exPtnLi_);
    logLi_ += logExPtnLi_ * PtnLs::weightArr_[ptnNo_];
  //std::cout << ptnNo_ << " / " << logLi_ << endl;
  }

  return logLi_;
}

//-------------------------------------------------------------------
//compute the log likelihood of this tree via this branch
double LiBr::cmpExLogLiGammaRate () {
  Cube *inLiCube_;   
  inLiCube_ = &greLiNd_->liNdCube_;
  
  Cube *exLiCube_;   
  exLiCube_ = &smaLiNd_->liNdCube_;

  Seq *seq_;
  seq_ = &Ali::getSeq (smaNdNo_);
  
  int nState_ = Model::getNState ();
  int nRate_ = Rate::getNRate ();
  double probRate_ = Rate::getProb ();
  
  DMat20 probChangeCube_[MAX_NUM_RATE];
  int rateNo_;
  for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
    double rate_ = Rate::getRate (rateNo_);
    Model::cmpProbChange (len_ * rate_, probChangeCube_[rateNo_]);
  }

  double logLi_ = 0.0;
  int nPtn_ = PtnLs::getNPtn ();    
  
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
    int siteNo_ = PtnLs::siteArr_ [ptnNo_];
  int exStateNo_ = nto [ static_cast<int> ((*seq_).items_[siteNo_]) ];
  long double exPtnLi_ = 0.0;

  for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {

       long double rateExPtnLi_ = 0.0;
       if (exStateNo_ < nState_) {    
         for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) 
           rateExPtnLi_ += (*inLiCube_)[ptnNo_][rateNo_][inStateNo_] * 
                       probChangeCube_[rateNo_][exStateNo_][inStateNo_];
                        
         rateExPtnLi_ *= stateFrqArr_[exStateNo_];
       } else {
       for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
         double exStateLi_ = 0.0;

       if (exStateNo_ == BS_UNKNOWN)
         exStateLi_ = 1.0;
       else
         for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) 
           exStateLi_ += probChangeCube_[rateNo_][inStateNo_][exStateNo_] * 
                       (*exLiCube_)[ptnNo_][rateNo_][exStateNo_];
         
           rateExPtnLi_ += (*inLiCube_)[ptnNo_][rateNo_][inStateNo_] * 
                          stateFrqArr_[inStateNo_] * exStateLi_;
         }
       }  

     exPtnLi_ += rateExPtnLi_ * probRate_;
  }
    
  if (exPtnLi_ <= 0.0) 
      Utl::announceError (BAD_CMP_LI);
  

    double logExPtnLi_ = cmpLogLi (exPtnLi_);
    logLi_ += logExPtnLi_ * PtnLs::weightArr_[ptnNo_];
  //std::cout << ptnNo_ << " / " << logLi_ << endl;
  }

  return logLi_;

}


double LiBr::cmpExLogLi () {

  RATE_TYPE rateType_ = Rate::getType ();
   if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && Rate::isOptedSpecificRate () == 0))
      return cmpExLogLiUniformRate ();
   else
     if (rateType_ == SITE_SPECIFIC)
       return cmpExLogLiSpecificRate ();
    else   
      return cmpExLogLiGammaRate ();
}

double LiBr::cmpInLogLi () {
  RATE_TYPE rateType_ = Rate::getType ();
   if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && Rate::isOptedSpecificRate () == 0) )
      return cmpInLogLiUniformRate ();
   else
     if (rateType_ == SITE_SPECIFIC) 
       return cmpInLogLiSpecificRate ();
    else   
      return cmpInLogLiGammaRate ();
}

//-------------------------------------------------------------------
//set the two liNds to be not cmped
void LiBr::openLiNd () {
  greLiNd_->turnOffIsCmped ();
  smaLiNd_->turnOffIsCmped ();
}

//-------------------------------------------------------------------
//set the two liNds to be not cmped
void LiBr::openTailLiNd () {
  if (getTailNd () == greNdNo_ ) 
    greLiNd_->turnOffIsCmped ();
  else
    smaLiNd_->turnOffIsCmped ();
}

//-------------------------------------------------------------------
//set the two liNds to be not cmped
void LiBr::openHeadLiNd () {
  if (getHeadNd () == greNdNo_ ) 
    greLiNd_->turnOffIsCmped ();
  else
    smaLiNd_->turnOffIsCmped ();
}


//======================================================
//set memory for this liBr

void LiBr::setLimit () {
}

//======================================================
//release all memory of this class
void LiBr::release () {
  smaLiNd_->release ();
  greLiNd_->release ();
}

//======================================================
//the destructor
LiBr::~LiBr () {
  release ();
}
