// innd.cpp: implementation of the inNd class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>

#include "outstream.h"
#include "lind.h"
#include "opturtree.h"
#include "model.h"
#include "ptnls.h"
#include "rate.h"
#include "ptnratecube.h"

extern int nto[256];

//--------------------------------------------------------------------
//the constructor function
LiNd::LiNd () {
  isCmped_ = 0;
  isEx_ = -1;  

  int nPtn_ = PtnLs::getNPtn ();

  int nRate_ = Rate::getNRate ();
  int nState_ = Model::getNState ();  
  Utl::getMem (liNdCube_, nPtn_, nRate_, nState_);  
}


//--------------------------------------------------------------------
//set the id, ndType, parrent node, and branch no for this node
void LiNd::set (int id, EX_IN ndType, int parNdNo, int parBrNo) {
  id_ = id;
  if (ndType == EX)
    isEx_ = 1;
  else
    isEx_ = 0;

  parNdNo_ = parNdNo;
  parBrNo_ = parBrNo;

  isExDesNdCreated_ = 0;
  isCmped_ = 0;
}


//--------------------------------------------------------------------
void LiNd::turnOffIsCmped () {
  isCmped_ = 0;
}

//--------------------------------------------------------------------
int LiNd::getId () {
  return id_;
}

Vec<int> &LiNd::getExDesNd () {
  return exDesNdNoLs_;
}

//--------------------------------------------------------------------
//the the parrent node, and branch no for this node
void LiNd::getPar (int &parNdNo, int &parBrNo) {
  parNdNo = parNdNo_;
  parBrNo = parBrNo_;
}

void LiNd::setParNdBr (int newParNdNo, int newParBrNo) {
  parNdNo_ = newParNdNo;
  parBrNo_ = newParBrNo;
}

void LiNd::writeInf () {
  std::cout << this << " / " << id_ << " / " << parNdNo_ << " / " << parBrNo_ << " / " << endl;
}

//-------------------------------------------------------------------
//compute the likehood of (ptnNo, rateNo, stateNo) via the parBrNo_
long double LiNd::cmpLi (int ptnNo, int rateNo, int stateNo) {
  if (isCmped_ == 0)
    cmpLi ();
  return liNdCube_[ptnNo][rateNo][stateNo];
} //end of function


//-------------------------------------------------------------------
//reCompute the likelihood for all ptnNo, rateNo, stateNo at this node 
void LiNd::reCmpLi () {
  isCmped_ = 0;
  cmpLi ();
}

//-------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiUniformRate () {  
  if (isCmped_ == 1)
    return;

//we do not need to reCmp again
  DVec20 stateFrqArr_;


  Model::getStateFrq (stateFrqArr_);
  //then, compute the likelihood of this internal node
  int nState_ = Model::getNState ();
//    std::cout << id_ << endl;

  
  if (isEx_ == 1) { //the case of external case
    Seq *seq_;
    seq_ = &Ali::getSeq (id_);
    int nPtn_ = PtnLs::getNPtn ();
    for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
    int siteNo_ = PtnLs::siteArr_[ptnNo_];
    int ndStateNo_ = nto[ static_cast<int> ((*seq_).items_[siteNo_]) ];
      
    if (ndStateNo_ < nState_) {      
       for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
          liNdCube_[ptnNo_][0][stateNo_] = 0.0;
     
        liNdCube_[ptnNo_][0][ndStateNo_] = 1.0;

     } else {
      for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
          liNdCube_[ptnNo_][0][stateNo_] = 1.0;

      if (ndStateNo_ != BS_UNKNOWN && Model::getDataType () == NUCLEOTIDE)  
        switch (ndStateNo_) {
          case BASE_AC:
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;
          case BASE_AG:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
         liNdCube_[ptnNo_][0][BASE_T] = 0.0;
           break;
          case BASE_AT:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
         liNdCube_[ptnNo_][0][BASE_G] = 0.0;
           break;
          case BASE_CG:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
         liNdCube_[ptnNo_][0][BASE_T] = 0.0;
           break;
          case BASE_CT:






             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
         liNdCube_[ptnNo_][0][BASE_G] = 0.0;
           break;
          case BASE_GT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
            break;
         case BASE_CGT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
              break;
          case BASE_AGT:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
             break;
          case BASE_ACT:
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             break;
          case BASE_ACG:
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;  
            } // end of switch      
        } //end of else   
    } //end of for

  } else { //the case of internal node

    //first compute the likelihood for all childrent
    int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;   

    OptUrTree::inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);
    double chiLen1_ = OptUrTree::getBrLen (chiBrNo1_);
    double chiLen2_ = OptUrTree::getBrLen (chiBrNo2_);

    int nState_ = Model::getNState ();
    
    DMat20 proChangeMat1_;
    DMat20 proChangeMat2_;

    Model::cmpProbChange (chiLen1_, proChangeMat1_);
    Model::cmpProbChange (chiLen2_, proChangeMat2_);

  
    LiNd *chiLi1_;

   chiLi1_ = &OptUrTree::liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);
  
    chiLi1_->cmpLiUniformRate ();
    
  LiNd *chiLi2_;
  chiLi2_ = &OptUrTree::liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
  chiLi2_->cmpLiUniformRate ();
  
    int nPtn_ = PtnLs::getNPtn ();
    for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
        for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {
       
       int chiStateNo_;
       long double chiLiBr1_ = 0.0;
       for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++) {
         chiLiBr1_ += chiLi1_->liNdCube_[ptnNo_][0][chiStateNo_] * 
                    proChangeMat1_[stateNo_][chiStateNo_];
       }

       long double chiLiBr2_ = 0.0;

       for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++)
         chiLiBr2_ += chiLi2_->liNdCube_[ptnNo_][0][chiStateNo_] * 
                             proChangeMat2_[stateNo_][chiStateNo_];

           long double liPtn_ = chiLiBr1_ * chiLiBr2_;
                

           liNdCube_[ptnNo_][0][stateNo_] = liPtn_;

        }//end of for stateNo_

    } //end of for ptnNo_
    
  } // end of else

  isCmped_ = 1;
}


//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiSpecificRate () {  
  if (isCmped_ == 1)
    return;

//we do not need to reCmp again
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);
  //then, compute the likelihood of this internal node
  int nState_ = Model::getNState ();
//    std::cout << id_ << endl;

  
  if (isEx_ == 1) { //the case of external case
    Seq *seq_;
    seq_ = &Ali::getSeq (id_);
    int nPtn_ = PtnLs::getNPtn ();
    for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
    int siteNo_ = PtnLs::siteArr_[ptnNo_];
    int ndStateNo_ = nto[ static_cast<int> ((*seq_).items_[siteNo_]) ];
      
    if (ndStateNo_ < nState_) {      
       for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
          liNdCube_[ptnNo_][0][stateNo_] = 0.0;
     
        liNdCube_[ptnNo_][0][ndStateNo_] = 1.0;

     } else {
      for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
          liNdCube_[ptnNo_][0][stateNo_] = 1.0;

      if (ndStateNo_ != BS_UNKNOWN && Model::getDataType () == NUCLEOTIDE)  
        switch (ndStateNo_) {
          case BASE_AC:
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;
          case BASE_AG:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
         liNdCube_[ptnNo_][0][BASE_T] = 0.0;
           break;
          case BASE_AT:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
         liNdCube_[ptnNo_][0][BASE_G] = 0.0;
           break;
          case BASE_CG:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
         liNdCube_[ptnNo_][0][BASE_T] = 0.0;
           break;
          case BASE_CT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
         liNdCube_[ptnNo_][0][BASE_G] = 0.0;
           break;
          case BASE_GT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
            break;
         case BASE_CGT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
              break;

          case BASE_AGT:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
             break;
          case BASE_ACT:
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             break;
          case BASE_ACG:
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;  
            } // end of switch      
        } //end of else   
    } //end of for

  } else { //the case of internal node

    //first compute the likelihood for all childrent
    int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;   

    OptUrTree::inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);

    LiNd *chiLi1_;

   chiLi1_ = &OptUrTree::liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);
  
    chiLi1_->cmpLiSpecificRate ();
    
   LiNd *chiLi2_;
   chiLi2_ = &OptUrTree::liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
   chiLi2_->cmpLiSpecificRate ();
  
    int nState_ = Model::getNState ();
    int nPtn_ = PtnLs::getNPtn ();
    
    double chiLen1_ = OptUrTree::getBrLen (chiBrNo1_);
    double chiLen2_ = OptUrTree::getBrLen (chiBrNo2_);

    Vec<double> *ptnRateArr_;
    ptnRateArr_ = Rate::getPtnRate ();

    for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
       double ptnRate_ = (*ptnRateArr_)[ptnNo_];
       DMat20 probChangePtnMat1_;
       DMat20 probChangePtnMat2_;
       Model::cmpProbChange (chiLen1_ * ptnRate_, probChangePtnMat1_);
       Model::cmpProbChange (chiLen2_ * ptnRate_, probChangePtnMat2_);

       for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {
       
          int chiStateNo_;
          long double chiLiBr1_ = 0.0;
          for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++)
            chiLiBr1_ += chiLi1_->liNdCube_[ptnNo_][0][chiStateNo_] * 
                               probChangePtnMat1_[stateNo_][chiStateNo_];

          long double chiLiBr2_ = 0.0;
          for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++)

            chiLiBr2_ += chiLi2_->liNdCube_[ptnNo_][0][chiStateNo_] * 
                               probChangePtnMat2_[stateNo_][chiStateNo_];

          long double liPtn_ = chiLiBr1_ * chiLiBr2_;
                

          liNdCube_[ptnNo_][0][stateNo_] = liPtn_;

       }//end of for stateNo_

    } //end of for ptnNo_
    
  } // end of else

  isCmped_ = 1;
}

//-------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiGammaRate () {  
//we do not need to reCmp again
  if (isCmped_ == 1)
    return;


  OptUrTree::nCmpLi_ ++;

  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);
  //then, compute the likelihood of this internal node
  int nRate_ = Rate::getNRate ();
  int nState_ = Model::getNState ();
  //std::cout << id_ << endl;  
  if (isEx_ == 1) { //the case of external case
//    std::cout << id_ << endl;
    Seq *seq_;

  seq_ = &Ali::getSeq (id_);
    int nPtn_ = PtnLs::getNPtn ();
    for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
    int siteNo_ = PtnLs::siteArr_[ptnNo_];
    int ndStateNo_ = nto[ static_cast<int> ((*seq_).items_[siteNo_]) ];
      
    if (ndStateNo_ < nState_) 
      for (int rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
        
      for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
            liNdCube_[ptnNo_][rateNo_][stateNo_] = 0.0;
     
          liNdCube_[ptnNo_][rateNo_][ndStateNo_] = 1.0;

        }   
    else 
      for (int rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {         

      for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
            liNdCube_[ptnNo_][rateNo_][stateNo_] = 1.0;

      if (ndStateNo_ != BS_UNKNOWN && Model::getDataType () == NUCLEOTIDE)  
          switch (ndStateNo_) {
          case BASE_AC:
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;
          case BASE_AG:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;
          case BASE_AT:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             break;
          case BASE_CG:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;
          case BASE_CT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             break;
          case BASE_GT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
             break;

          case BASE_CGT:
             liNdCube_[ptnNo_][0][BASE_A] = 0.0;
             break;
          case BASE_AGT:
             liNdCube_[ptnNo_][0][BASE_C] = 0.0;
             break;
          case BASE_ACT:
             liNdCube_[ptnNo_][0][BASE_G] = 0.0;
             break;
          case BASE_ACG:
             liNdCube_[ptnNo_][0][BASE_T] = 0.0;
             break;  
         } // end of switch      
      } //end of for rateNo

    }

  } else { //the case of internal node

    //first compute the likelihood for all childrent
    int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;   

    OptUrTree::inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);
  double chiLen1_ = OptUrTree::getBrLen (chiBrNo1_);
  double chiLen2_ = OptUrTree::getBrLen (chiBrNo2_);

    int nState_ = Model::getNState ();
    int nRate_ = Rate::getNRate ();
//    double probRate_ = Rate::getProb ();
    
     DMat20 *proChangeCube1_;
     proChangeCube1_ = new DMat20[nRate_];

     DMat20 *proChangeCube2_;
     proChangeCube2_ = new DMat20[nRate_];

//    DMat20 proChangeCube1_[MAX_NUM_RATE];
//    DMat20 proChangeCube2_[MAX_NUM_RATE];

    int rateNo_;
    for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
      double rate_ = Rate::getRate (rateNo_);
      Model::cmpProbChange (chiLen1_ * rate_, proChangeCube1_[rateNo_]);
      Model::cmpProbChange (chiLen2_ * rate_, proChangeCube2_[rateNo_]);
    }

  
    LiNd *chiLi1_;
    chiLi1_ = &OptUrTree::liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);
    chiLi1_->cmpLi ();
    
    LiNd *chiLi2_;
    chiLi2_ = &OptUrTree::liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
    chiLi2_->cmpLi ();
  
    int nPtn_ = PtnLs::getNPtn ();
    for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

      for (int rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) 
        for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {
       
       int chiStateNo_;
       long double chiLiBr1_ = 0.0;
       for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++)
         chiLiBr1_ += chiLi1_->liNdCube_[ptnNo_][rateNo_][chiStateNo_] * 
                            proChangeCube1_[rateNo_][stateNo_][chiStateNo_];

       long double chiLiBr2_ = 0.0;
       for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++)
         chiLiBr2_ += chiLi2_->liNdCube_[ptnNo_][rateNo_][chiStateNo_] * proChangeCube2_[rateNo_][stateNo_][chiStateNo_];

           long double liPtn_ = chiLiBr1_ * chiLiBr2_;
                

           liNdCube_[ptnNo_][rateNo_][stateNo_] = liPtn_;

        }//end of for stateNo_

    } //end of for ptnNo_

    delete [] proChangeCube1_;
    delete [] proChangeCube2_;
    
  } // end of else

  isCmped_ = 1;
}


void LiNd::cmpLi () {
  RATE_TYPE rateType_ = Rate::getType ();
   if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && Rate::isOptedSpecificRate () == 0))
      cmpLiUniformRate ();
   else
     if (rateType_ == SITE_SPECIFIC)
       cmpLiSpecificRate ();
    else   
       cmpLiGammaRate ();
}


//--------------------------------------------------------------------
//create the external descendant nodes no list of this node
void LiNd::createExDesNd (Vec<int> &exDesNdNoLs) {
  if (isExDesNdCreated_ == 0) {  
    if (isEx_ == 1) {
      exDesNdNoLs_.set (1, 0);
      exDesNdNoLs_ += id_;
    } else {
      int chiBrNo1_, chiBrNo2_;
      OptUrTree::inNdArr_[id_].get2RemBr (parBrNo_, chiBrNo1_, chiBrNo2_);

      exDesNdNoLs_.set (MAX_NUM_SEQ, 0);
  
      Vec<int> chiExDesNdNoLs1_;
      OptUrTree::liBrArr_[chiBrNo1_].getRemLiNd (id_).createExDesNd (chiExDesNdNoLs1_);
      exDesNdNoLs_ += chiExDesNdNoLs1_;

      Vec<int> chiExDesNdNoLs2_;
      OptUrTree::liBrArr_[chiBrNo2_].getRemLiNd (id_).createExDesNd (chiExDesNdNoLs2_);
      exDesNdNoLs_ += chiExDesNdNoLs2_;
    }
  }


  exDesNdNoLs = exDesNdNoLs_;
  isExDesNdCreated_ = 1;
}

//--------------------------------------------------------------------
//create the external descendant nodes no list of this node
void LiNd::createExDesNd () {
  createExDesNd (exDesNdNoLs_);
}

//--------------------------------------------------------------------
//set memory for this class
void LiNd::setLimit () {
}

//--------------------------------------------------------------------
//release the memory of this class
void LiNd::release () {
  exDesNdNoLs_.release ();
  int nPtn_ = PtnLs::getNPtn ();
  int nRate_ = Rate::getNRate ();
//  int nState_ = Model::getNState ();  
  Utl::delMem (liNdCube_, nPtn_, nRate_);  
}

//-------------------------------------------------------------------
//the destructor function
LiNd::~LiNd () {
  release ();
//  std::cout << "this is the destructor of liInNd class " << endl;
}
