/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Fri Apr 11 10:08:35 CEST 2003
    copyright            : (C) 2003 by
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>

#include "constant.h"
#include "utl.h"
#include "vec.h"
#include "mat.h"
#include "cube.h"
#include "seq.h"
#include "ali.h"
#include "ptn.h"
#include "ptnls.h"
#include "brent.h"
#include "exnd.h"
#include "exndarr.h"
#include "innd.h"
#include "inndarr.h"
#include "br.h"
#include "brarr.h"
#include "basem.h"
#include "model.h"
#include "optpairseq.h"
#include "rate.h"
#include "opturtree.h"
#include "urtree.h"
#include "outstream.h"
#include "testmethod.h"
#include "bionj.h"
#include "interface.h"
#include "rannum.h"
#include "tn93m.h"


int isContinuous = 0;
int nto[NUM_CHAR];

void createPosGenChar () {
  for (int count_ = 0; count_ < NUM_CHAR; count_ ++)
    nto [ count_ ] = BS_UNKNOWN;

//here come for Nucleotides
  nto[static_cast<int> ('A')] = BASE_A;
  nto[static_cast<int> ('C')] = BASE_C;
  nto[static_cast<int> ('T')] = BASE_T;
  nto[static_cast<int> ('G')] = BASE_G;


  
  nto[static_cast<int> ('M')] = BASE_AC;
  nto[static_cast<int> ('R')] = BASE_AG;
  nto[static_cast<int> ('W')] = BASE_AT;

  nto[static_cast<int> ('S')] = BASE_CG;
  nto[static_cast<int> ('Y')] = BASE_CT;
  nto[static_cast<int> ('K')] = BASE_GT;

  nto[static_cast<int> ('B')] = BASE_CGT;
  nto[static_cast<int> ('D')] = BASE_AGT;
  nto[static_cast<int> ('H')] = BASE_ACT;
  nto[static_cast<int> ('V')] = BASE_ACG;

//here come for AA
  nto[static_cast<int> ('A')] = STATE_A;
  nto[static_cast<int> ('C')] = STATE_C;
  nto[static_cast<int> ('T')] = STATE_T;
  nto[static_cast<int> ('G')] = STATE_G;

  nto[static_cast<int> ('R')] = STATE_R;
  nto[static_cast<int> ('N')] = STATE_N;
  nto[static_cast<int> ('D')] = STATE_D;
  nto[static_cast<int> ('Q')] = STATE_Q;

  nto[static_cast<int> ('E')] = STATE_E;
  nto[static_cast<int> ('H')] = STATE_H;
  nto[static_cast<int> ('I')] = STATE_I;
  nto[static_cast<int> ('L')] = STATE_L;

  nto[static_cast<int> ('K')] = STATE_K;
  nto[static_cast<int> ('M')] = STATE_M;
  nto[static_cast<int> ('F')] = STATE_F;
  nto[static_cast<int> ('P')] = STATE_P;

  nto[static_cast<int> ('S')] = STATE_S;
  nto[static_cast<int> ('W')] = STATE_W;
  nto[static_cast<int> ('Y')] = STATE_Y;
  nto[static_cast<int> ('V')] = STATE_V;
}
//===============================================
void createAmino () {
  for (int count_ = 0; count_ < NUM_CHAR; count_ ++)
    nto [ count_ ] = BS_UNKNOWN;
  
  nto[static_cast<int> ('A')] = STATE_A;
  nto[static_cast<int> ('C')] = STATE_C;
  nto[static_cast<int> ('T')] = STATE_T;
  nto[static_cast<int> ('G')] = STATE_G;

  nto[static_cast<int> ('R')] = STATE_R;
  nto[static_cast<int> ('N')] = STATE_N;
  nto[static_cast<int> ('D')] = STATE_D;
  nto[static_cast<int> ('Q')] = STATE_Q;

  nto[static_cast<int> ('E')] = STATE_E;
  nto[static_cast<int> ('H')] = STATE_H;
  nto[static_cast<int> ('I')] = STATE_I;
  nto[static_cast<int> ('L')] = STATE_L;

  nto[static_cast<int> ('K')] = STATE_K;
  nto[static_cast<int> ('M')] = STATE_M;
  nto[static_cast<int> ('F')] = STATE_F;
  nto[static_cast<int> ('P')] = STATE_P;

  nto[static_cast<int> ('S')] = STATE_S;
  nto[static_cast<int> ('W')] = STATE_W;
  nto[static_cast<int> ('Y')] = STATE_Y;
  nto[static_cast<int> ('V')] = STATE_V;
}


void createDNA () {
  for (int count_ = 0; count_ < NUM_CHAR; count_ ++)
    nto [ count_ ] = BS_UNKNOWN;
  
  nto[static_cast<int> ('A')] = BASE_A;
  nto[static_cast<int> ('C')] = BASE_C;
  nto[static_cast<int> ('T')] = BASE_T;
  nto[static_cast<int> ('G')] = BASE_G;

  
  nto[static_cast<int> ('M')] = BASE_AC;
  nto[static_cast<int> ('R')] = BASE_AG;
  nto[static_cast<int> ('W')] = BASE_AT;

  nto[static_cast<int> ('S')] = BASE_CG;
  nto[static_cast<int> ('Y')] = BASE_CT;
  nto[static_cast<int> ('K')] = BASE_GT;

  nto[static_cast<int> ('B')] = BASE_CGT;
  nto[static_cast<int> ('D')] = BASE_AGT;
  nto[static_cast<int> ('H')] = BASE_ACT;
  nto[static_cast<int> ('V')] = BASE_ACG;
}

void init () {
  OptUrTree::doConstructor ();
  OptPairSeq::doConstructor ();
  Ali::doConstructor ();
  PtnLs::doConstructor ();
  createPosGenChar ();
}


//overload the operator <<
std::ostream &operator << (std::ostream &out, Vec<char> &vec) {
  return OutStream::write (vec, out);
}

//overload the operator <<
std::ofstream &operator << (std::ofstream &out, Vec<char> &vec) {

  return OutStream::write (vec, out);
}


int main(int argc, char *argv[])
{
  RanNum::initRandom (-1);

  char aliFileName_[FILENAME_LEN];
  if (argc == 2)   
     strcpy (aliFileName_, argv[1]);

  init ();

  MODEL model_           = HKY85;  
  double tsTvRatio_      = 4.0;
  double pyPuRatio_      = 1.0;


  double tsAG_ = 1.0 / 6.0;  
  double tsCT_ = 1.0 / 6.0;  
  double tvAC_ = 1.0 / 6.0;  
  double tvAT_ = 1.0 / 6.0;  
  double tvCG_ = 1.0 / 6.0;  
  double tvGT_ = 1.0 / 6.0;  

  double baseA_ = 0.25;
  double baseC_ = 0.25;
  double baseG_ = 0.25;
  double baseT_ = 0.25;

  int nOptTree_;
  double probSel_;
  int nRep_;
  int outGrpSeqNo_;

  Interface inter_;

  DATA_TYPE dataType_; //DNA or Amino Acid
  PAM_TYPE  baseFrqType_;
  PAM_TYPE  tsTvRatioType_;
  PAM_TYPE  pyPuRatioType_;
  PAM_TYPE   genPamType_;
  PAM_TYPE  gammaShapeType_;
  
  RATE_TYPE rateType_;
  int nRate_;
  double gammaShape_;

  STOPPING_TYPE stoppingRule_;

  int status_ = inter_.getUserPam (aliFileName_, 
                                   nOptTree_, probSel_, nRep_, stoppingRule_,
                                   dataType_, model_,
                                   baseA_, baseC_, baseG_, baseT_,
                                   baseFrqType_,
 
                                   tsTvRatio_,
                                   tsTvRatioType_,

                                   pyPuRatio_,
                                   pyPuRatioType_,
 
                                   tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_,
                                   genPamType_,     

                                   rateType_, nRate_, 
                                   gammaShape_,
                                   gammaShapeType_,

                                   outGrpSeqNo_);

  if (status_ == 0)
    return 0;
  char userTreeFileName_[]   = "usertree.iqpnni";
  
  if (dataType_ == NUCLEOTIDE)
    createDNA ();
  else
    createAmino ();
  
  TestMethod testMethod_;
   
  testMethod_.create (userTreeFileName_, aliFileName_,                    
                      nOptTree_, probSel_, nRep_, stoppingRule_,
                      dataType_, model_,

                      baseA_, baseC_, baseG_, baseT_,
                      baseFrqType_,

                      tsTvRatio_,
                      tsTvRatioType_,

                      pyPuRatio_,
                      pyPuRatioType_,
 
                      tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_,
                      genPamType_,     

                      rateType_, nRate_, 
                      gammaShape_,
                      gammaShapeType_,
                      outGrpSeqNo_);


  return EXIT_SUCCESS;
}

