/***************************************************************************
                          model.cpp  -  description
                             -------------------
    begin                : Wed Mar 5 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include "model.h"
#include "mat.h"
#include "constant.h"
#include "ali.h"
#include "dmat20.h"

DVec20 Model::stateFrqArr_;

double Model::tsTvRatio_;
double Model::pyPuRatio_;

double Model::tsAG_;
double Model::tsCT_;

double Model::tvAC_;
double Model::tvAT_;
double Model::tvCG_;
double Model::tvGT_;

MODEL Model::modelType_;
PAM_TYPE Model::genPamType_;
PAM_TYPE Model::tsTvRatioType_;
PAM_TYPE Model::pyPuRatioType_;
PAM_TYPE Model::baseFrqType_;

DATA_TYPE Model::dataType_;

int Model::nState_;

BaseM     *Model::model_;
HKY85M     Model::hKY85M_;
TN93M     Model::tN93M_;
GTRM        Model::gtrM_;
AminoM     Model::aminoM_;


//the constructor function of this class
Model::Model () {
}


//====================================
void Model::setDataType (DATA_TYPE dataType) {
   dataType_ = dataType;
}

//====================================
DATA_TYPE Model::getDataType () {
   return dataType_;
}


//====================================
void Model::setStateFrq (double stateFrqA, double stateFrqC, double stateFrqG, double stateFrqT) {
  stateFrqArr_[BASE_A] = stateFrqA;
  stateFrqArr_[BASE_C] = stateFrqC;
  stateFrqArr_[BASE_G] = stateFrqG;
  stateFrqArr_[BASE_T] = stateFrqT;
}

void Model::setBaseFrqType (PAM_TYPE baseFrqType) {
  baseFrqType_ = baseFrqType;
}

//============================================
int Model::getNState () {
  return nState_;
}

//============================================
MODEL Model::getType () {
  return modelType_;  
}

PAM_TYPE Model::getTsTvRatioType () {
   return tsTvRatioType_;
}

PAM_TYPE Model::getPyPuRatioType () {
   return pyPuRatioType_;
}

PAM_TYPE Model::getGenPamType () {
   return genPamType_;
}

PAM_TYPE Model::getBaseFrqType () {
   return baseFrqType_;
}

void Model::setNState (int nState) {
  nState_ = nState;
}

//set the model of evolutional change
void Model::setModelType (const MODEL modelType) {
  modelType_ = modelType;

  if (modelType_ == HKY85 || modelType_ == TN93 || modelType_ == GTR) {
    if (modelType_ == HKY85) 
      model_ = &hKY85M_;
    else
      if (modelType_ == TN93)
        model_ = &tN93M_;
      else
        model_ = &gtrM_;

    nState_ = NUM_BASE;
  } else {
      model_ = &aminoM_;
      nState_ = NUM_AMINO;
  }

  model_->setModel (modelType_);
}  

//----------------------------------------------------------------------------------------------------------------------------------------
int Model::isEstedPam () {
  return ( (modelType_ == HKY85 && tsTvRatioType_ == ESTIMATE) ||
           (modelType_ == TN93 && tsTvRatioType_ == ESTIMATE) ||
           (modelType_ == TN93 && pyPuRatioType_ == ESTIMATE) ||
           (modelType_ == GTR &&  genPamType_ == ESTIMATE) );
    
}


//----------------------------------------------------------------------------------------------------------------------------------------
//set the tsTvRatio
void Model::setTsTvRatio (const double tsTvRatio) {
  tsTvRatio_ = tsTvRatio;
}

//----------------------------------------------------------------------------------------------------------------------------------------
//set the tsTvRatio
void Model::setTsTvRatioType (const PAM_TYPE tsTvRatioType) {
  tsTvRatioType_ = tsTvRatioType;
} 

//----------------------------------------------------------------------------------------------------------------------------------------
//set the pyPuRatio
void Model::setPyPuRatio (const double pyPuRatio) {
  pyPuRatio_ = pyPuRatio;
}

//----------------------------------------------------------------------------------------------------------------------------------------
//set the pyPuRatioType
void Model::setPyPuRatioType (const PAM_TYPE pyPuRatioType) {
  pyPuRatioType_ = pyPuRatioType;
}

//----------------------------------------------------------------------------------------------------------------------------------------
void Model::setGenPam (const double tsAG, const double tsCT,
                                     const double tvAC, const double tvAT, const double tvCG, const double tvGT) {
   tsAG_  = tsAG;
   tsCT_  = tsCT;
   tvAC_  = tvAC;
   tvAT_  = tvAT;
   tvCG_  = tvCG;
   tvGT_  = tvGT;
}

//----------------------------------------------------------------------------------------------------------------------------------------
void Model::setGenPamType (const PAM_TYPE genPamType) {
  genPamType_ = genPamType;
}


//--------------------------------------------------------------------

//init all necessary things for using this class
void Model::init () {
  if (dataType_ == NUCLEOTIDE) {
    if (baseFrqType_ == ESTIMATE) 
       Ali::estimateStateFrq (stateFrqArr_);

    model_->setStateFrq (stateFrqArr_);

    if (tsTvRatioType_ == ESTIMATE) 
      tsTvRatio_ = 4.0;
   model_->setTsTvRatio(tsTvRatio_);

    if (pyPuRatioType_ == ESTIMATE) 
       pyPuRatio_ = 1.0;
    model_->setPyPuRatio (pyPuRatio_);
 
    if (genPamType_ == ESTIMATE && modelType_ == GTR) 
       Ali::estimateGenPam (tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_); 

    model_->setGenPam (tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_);
    
  }  // end of if dataType_

  
  model_->init ();

  if (dataType_ == AMINO_ACID) 
     model_->getStateFrq (stateFrqArr_);
}

//================================================================
double Model::getTsTvRatio () {
  return model_->getTsTvRatio ();
}

//================================================================
double Model::getPyPuRatio () {
  return model_->getPyPuRatio ();
}

//================================================================
void Model::getGenPam (double &tsAG, double & tsCT,
                                     double & tvAC, double & tvAT, double & tvCG, double & tvGT) {
  
   tsAG  = tsAG_;
   tsCT  = tsCT_;
   tvAC  = tvAC_;
   tvAT  = tvAT_;
   tvCG  = tvCG_;
   tvGT  = tvGT_;
}

//--------------------------------------------------------------------
/*optimize all parameter evolving to sequence evolution
1. tsTvRatio
3. general model parameters
*/

int Model::optPam (const int maxNIter) {
  int isChanged_ = 0;

  if (modelType_ == HKY85 && tsTvRatioType_ == ESTIMATE) {
       double oriTsTvRatio_ = tsTvRatio_;
       model_->optPam ();
       tsTvRatio_ = model_->getTsTvRatio ();  
       if (fabs (tsTvRatio_ - oriTsTvRatio_) > EPS_MODEL_PAM_ERROR)
          isChanged_ = 1;
  } // end if HKY85

  if (modelType_ == TN93) {
    if (tsTvRatioType_ == ESTIMATE || pyPuRatioType_ == ESTIMATE) {
       double oriTsTvRatio_ = tsTvRatio_;
       double oriPyPuRatio_ = pyPuRatio_;
       
       model_->optPam (maxNIter, tsTvRatioType_, pyPuRatioType_);
       tsTvRatio_ = model_->getTsTvRatio ();  
       pyPuRatio_ = model_->getPyPuRatio ();
 
       if (fabs (tsTvRatio_ - oriTsTvRatio_) > EPS_MODEL_PAM_ERROR ||
           fabs (pyPuRatio_ - oriPyPuRatio_) > EPS_MODEL_PAM_ERROR )
          isChanged_ = 1;
     }       
  } //end of TN93

  
  if (modelType_ == GTR && genPamType_ == ESTIMATE) {
       double oriTsAG_ = tsAG_;
       double oriTsCT_ = tsCT_;
       double oriTvAC_ = tvAC_;
       double oriTvAT_ = tvAT_;
       double oriTvCG_ = tvCG_;
       double oriTvGT_ = tvGT_;
       
        
       model_->optPam (maxNIter);

       model_->getGenPam (tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_);

       if ( fabs (tsAG_ - oriTsAG_) > EPS_MODEL_PAM_ERROR || 
            fabs (tsCT_ - oriTsCT_) > EPS_MODEL_PAM_ERROR ||  
            fabs (tvAC_ - oriTvAC_) > EPS_MODEL_PAM_ERROR || 
            fabs (tvAT_ - oriTvAT_) > EPS_MODEL_PAM_ERROR || 
            fabs (tvCG_ - oriTvCG_) > EPS_MODEL_PAM_ERROR || 
            fabs (tvGT_ - oriTvGT_) > EPS_MODEL_PAM_ERROR  )

            isChanged_ = 1;
   } 

    return isChanged_;
}


//--------------------------------------------------------------------
//compute the probability mat
void Model::cmpProbChange (const double brLen, DMat20 &probMat) {
  model_->cmpProbChange (brLen, probMat);
}


//--------------------------------------------------------------------
//compute the probability vec
void Model::cmpProbChange (const int stateNo, const double brLen, DVec20 &probVec) {
  model_->cmpProbChange (stateNo, brLen, probVec);
}


//--------------------------------------------------------------------
/*compute the probability of changing from stateNo1 into stateNo2
after a period of brLen / subRate_
*/
double Model::cmpProbChange (const int stateNo1, const int stateNo2, const double brLen) {
  return model_->cmpProbChange (stateNo1, stateNo2, brLen);
}

//--------------------------------------------------------------------
//return the state frequence
void Model::getStateFrq (DVec20 &stateFrqArr) {
  for (int count_ = 0; count_ < nState_; count_ ++)
    stateFrqArr[count_] = stateFrqArr_[count_];
}

//--------------------------------------------------------------------
//return the type of model

MODEL Model::getModelType () {
  return modelType_;
}



//--------------------------------------------------------------------
//return the state frequence
double Model::getStateFrq (const int stateNo) {
  return stateFrqArr_[stateNo];
}


//--------------------------------------------------------------------
//release the memory of this class

void Model::release () {
}

//--------------------------------------------------------------------
//the destruction function of this class

Model::~Model () {
  std::cout << "This is destructor of model class " << endl;
}
