/***************************************************************************
                          model.h  -  description
                             -------------------
    begin                : Wed Mar 5 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MODEL_H
#define MODEL_H

#include "constant.h"
#include "mat.h"
#include "basem.h"

#include "gtrm.h"

#include "hky85m.h"
#include "tn93m.h"



#include "aminom.h"
#include "dvec20.h"
#include "dmat20.h"

//this class is an interface for all methods
class Model {
public: 
//the constructor function
Model();


static int getNState ();


//set the type of evolutional change model
static void setModelType (const MODEL modelType);
static void setDataType (DATA_TYPE dataType);
static void setTsTvRatioType (PAM_TYPE tsTvRatioType);
static void setPyPuRatioType (PAM_TYPE pyPuRatioType);

static void setPyPuRatio (PAM_TYPE pyPuRatio);
static void setGenPamType (PAM_TYPE genPamType);
static void setBaseFrqType (PAM_TYPE baseFrqType);
static int isEstedPam ();


//get the type of evolutional change model
static MODEL getModelType ();

//get the type of evolutional change model
static DATA_TYPE getDataType ();

static PAM_TYPE getTsTvRatioType ();
static PAM_TYPE getPyPuRatioType ();
static PAM_TYPE getGenPamType ();
static PAM_TYPE getBaseFrqType ();

//set the transition transvertion ratio
static void setTsTvRatio (const double tsTvRatio);
static void setPyPuRatio (const double pyPuRatio);

static void setGenPam (const double tsAG, const double tsCT,
                       const double tvAC, const double tvAT, const double tvCG, const double tvGT);


static double getTsTvRatio ();
static double getPyPuRatio ();


//everything is inited here
static void init ();
static void setNState (int nState);

static void setStateFrq (double stateFrqA, double stateFrqC, double stateFrqG, double stateFrqT);

//return the state frequence
static void getStateFrq (DVec20 &stateFrqArr);
//================================================================
static void getGenPam (double &tsAG, double & tsCT, double & tvAC, double & tvAT, double & tvCG, double & tvGT);


//return the state frequence
static double getStateFrq (const int stateNo);
static MODEL getType ();


/*optimize all parameter evolving to sequence evolution
1. tsTvRatio

2. TsRatio
3. find state frequencies
*/
static int optPam (const int maxNIter = MAX_IT_ALL_PAM_OPT);

//compute the probability mat
static void cmpProbChange (const double time, DMat20 &probMat);
//compute the first order probability mat

//compute the probability vec
static void cmpProbChange (const int stateNo, const double time, DVec20 &probVec);
//compute the first order probability vec


//compute the probability of changing from stateNo1 into stateNo2 after a period time of time
static double cmpProbChange (const int stateNo1, const int stateNo2, const double time);
//compute the first order probability of changing from stateNo1 into stateNo2 after a period time of time


//release all memmory for this class
void release ();

//the constructor
virtual ~Model ();

/************************************************************************************
************************************************************************************
************************************************************************************
************************************************************************************
************************************************************************************/

//the type of evolutional change
static MODEL modelType_;

static PAM_TYPE baseFrqType_;
static PAM_TYPE tsTvRatioType_;
static PAM_TYPE pyPuRatioType_;
static PAM_TYPE genPamType_;


//the transition transvertion ratio
static double tsTvRatio_;
static double pyPuRatio_;

static double tsAG_;
static double tsCT_;

static double tvAC_;
static double tvAT_;
static double tvCG_;
static double tvGT_;


//the state frequency list

static DVec20 stateFrqArr_;

static int nState_;
static DATA_TYPE dataType_;

static BaseM     *model_;
static GTRM      gtrM_;
static HKY85M     hKY85M_;
static TN93M     tN93M_;
static AminoM     aminoM_;

}; //end of class model
#endif
