#include <math.h>
#include <iostream>

#include "optpairseq.h"
#include "mat.h"
#include "model.h"
#include "brent.h"
#include "constant.h"
#include "rate.h"
#include "ptnls.h"

extern int nto[NUM_CHAR];

Seq * OptPairSeq::seq1_;
Seq * OptPairSeq::seq2_;
int OptPairSeq::ptnMat_[NUM_CAN_STATE][NUM_CAN_STATE];

//the constructor function
OptPairSeq::OptPairSeq () {
}

//this function is used for constructing this class
void OptPairSeq::doConstructor () {
}

//-------------------------------------------------------------
//compute the observed distance between two sequences
double OptPairSeq::cmpObsDis (Seq &seq1, Seq &seq2) {
  int nDif = 0;
  for (int site_ = 0; site_ < seq1.getNSite (); site_ ++)
    if (seq1[site_] != seq2[site_] && seq1[site_] != BS_UNKNOWN && seq2[site_] != BS_UNKNOWN) 
      nDif ++;

  return static_cast<double> (nDif) / seq1.getNSite ();
}


//-------------------------------------------------------------
//compute the negative log likelihood when the distance between them is brLen
double OptPairSeq::cmpLogLiSpecificRate (double brLen) {
  int nPtn_ = PtnLs::getNPtn ();

  Vec<double> *ptnRateArr_;
  ptnRateArr_ = Rate::getPtnRate ();
  
  
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  double logLi_ = 0.0;
  for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
     int siteNo_ = PtnLs::siteArr_ [ptnNo_];
     int stateNo1_ = nto [ static_cast<int> ((*seq1_)[siteNo_]) ];
     int stateNo2_ = nto [ static_cast<int> ((*seq2_)[siteNo_]) ];

     double ptnRate_ = (*ptnRateArr_)[ptnNo_];
     double prob_ = Model::cmpProbChange (stateNo1_, stateNo2_, brLen * ptnRate_);
   
     double logPtnLi_ = log ( stateFrqArr_[stateNo1_] * prob_);
     logLi_ += logPtnLi_ * PtnLs::weightArr_[ptnNo_];
  }

  return logLi_;
}

//compute the negative log likelihood when the distance between them is brLen
double OptPairSeq::cmpLogLiGammaRate(double brLen) {
  double logLi_ = 0.0;
  int stateNo1_, stateNo2_;
  int nState_ = Model::getNState ();
  
  double probRate_ = Rate::getProb ();
  DMat20 probChangeCube_[MAX_NUM_RATE];
  int rateNo_;
  int nRate_ = Rate::getNRate ();
  for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
    double rate_ = Rate::getRate (rateNo_);
    Model::cmpProbChange (brLen * rate_, probChangeCube_[rateNo_]);
  }
  
  DVec20 stateFrqArr_;
  Model::getStateFrq (stateFrqArr_);

  for (stateNo1_ = 0; stateNo1_ < nState_; stateNo1_ ++) 
    for (stateNo2_ = 0; stateNo2_ < nState_; stateNo2_ ++) {
      double liPtn_ = 0.0;    
    
      //the rate of this pattern follows gamma distribution
      for (int rateNo_ = 0; rateNo_ < Rate::getNRate (); rateNo_ ++) {
        double probChange_ = probChangeCube_[rateNo_][stateNo1_][stateNo2_];
        liPtn_ += probRate_ * stateFrqArr_[stateNo1_] * probChange_;
      }

      logLi_ += log (liPtn_) * ptnMat_[stateNo1_][stateNo2_]; 
    }

  return logLi_;
}

//compute the negative log likelihood when the distance between them is brLen
double OptPairSeq::cmpNegLogLi(double brLen) {
   double logLi_;
   if (Rate::getType () == SITE_SPECIFIC && Rate::isOptedSpecificRate () == 1)
       logLi_ = cmpLogLiSpecificRate (brLen);
   else
     logLi_ = cmpLogLiGammaRate (brLen);

  return -logLi_;
}

//-------------------------------------------------------------
//create the pair pattern matrix of two sequences
void OptPairSeq::createPairPtn () {
 int stateNo1_, stateNo2_;
 int nState_ = Model::getNState ();
 for (stateNo1_ = 0; stateNo1_ < nState_; stateNo1_ ++)
   for (stateNo2_ = 0; stateNo2_ < nState_; stateNo2_ ++)     
     ptnMat_[stateNo1_][stateNo2_] = 0;

 for (int siteNo_ = 0; siteNo_ < seq1_->getSize (); siteNo_ ++) {
   stateNo1_ = nto [ static_cast<int> ((*seq1_)[siteNo_]) ];
   stateNo2_ = nto [ static_cast<int> ((*seq2_)[siteNo_]) ];
   if (stateNo1_ < nState_ && stateNo2_ < nState_)
     ptnMat_[stateNo1_][stateNo2_] ++;
 }
}

//-------------------------------------------------------------
//return the genetic distance between two sequences
double OptPairSeq::cmp (Seq &seq1, Seq &seq2, double initGenDis) {
  double genDis_;
//  seq1.writeOut ();
//  seq2.writeOut ();  
  double jCDis_;
  double observedDis_ = cmpObsDis (seq1, seq2);
  if (initGenDis < 0.0) {    
    //the case, two sequences are too far from each other
    if (1.0 - (4.0 / 3.0) * observedDis_ < ZERO)
      jCDis_ = observedDis_;
    else
      jCDis_ =  - (3.0 / 4.0) * log (1.0 - (4.0 / 3.0) * observedDis_);

//    genDis_ = jCDis_ * COF_BR;
    genDis_ = jCDis_;
  }
  else
    genDis_ = initGenDis;
  
  //if two sequences are identical, of course, the distance between them is 0.0
  if (seq1.getId () == seq2.getId () )
    return 0.0;
  else {
    seq1_ = &seq1;
    seq2_ = &seq2;
    createPairPtn ();
  }

  //put the distance into the range of a posible length
  if (genDis_ <= MIN_BR_LEN)
    genDis_ = MIN_BR_LEN + 1.0;
  if (genDis_ >=MAX_BR_LEN)
    genDis_ = MAX_BR_LEN - 1.0;

  double fx, f2x;
  
  genDis_ = Brent::optOneDim (MIN_BR_LEN, genDis_, MAX_BR_LEN,
                              OptPairSeq::cmpNegLogLi,
                              EPS_OPT_PAIR_SEQ_BR, &fx, &f2x);

/*  double oriFx_ = OptPairSeq::cmpNegLogLi (jCDis_ * COF_BR);
  double optFx_ =  OptPairSeq::cmpNegLogLi (genDis_);

  cout.precision (20);
  cout << oriFx_ - optFx_ << endl;

  cout.precision (20);
  cout << genDis_ << endl;
*/
  
  return genDis_;
}

//--------------------------------------------------------------------
//release the memory of this class
void OptPairSeq::release () {
}

//-------------------------------------------------------------
//the destructor function
OptPairSeq::~OptPairSeq () {
  release ();
  std::cout << "this is destructor of optimal pair sequences class " << endl;
}
