#ifndef OPTURTREE_H
#define OPTURTREE_H

#include "exndarr.h"
#include "inndarr.h"
#include "brarr.h"
#include "libr.h"
#include "lind.h"
#include "librarr.h"
#include "vec.h"
#include "dmat20.h"

/*this is one of the most important class, it describles all things involving
optimizing an unrooted tree */

class OptUrTree
{
public:
friend class LiNd;
friend class LiBr;

//the constructor function
OptUrTree();

//use this function for constructing this class
static void doConstructor ();

static void changeNd (int parNdno, int sNdno, int chiNdNo, int dNdNo);

static double optLocalBr (int brNo);

static int optLocal (int parNdNo, int sNdNo, int dNdNo);

static void sortScoreBr ();

static int findNPosSwapBr (int maxNSwapBr);
static void trySwapBr (double lamda);
static int swapBr ();
static void swapBrBack ();
static double cmpSwapLogLi ();
static double optSwapLogLi ();
static void createScoreBr ();
static double optLocal ();
static void draw ();

static double oriLenArr_[MAX_NUM_SEQ * 2];
static double canLenArr_[MAX_NUM_SEQ * 2];
static double optedLenArr_[MAX_NUM_SEQ * 2];

//set out group node no for this tree
static void setOutGrpNd (int outGrpNdNo);

//set the exNdArr for this class, construct the structure of this unrooted tree
static void setExNd (ExNdArr &exNdArr);

//set the inNdArr for this class, construct the structure of this unrooted tree
static void setInNd (InNdArr &inNdArr);

//set the brArr for this class, construct the structure of this unrooted tree
static void setBr (BrArr<double> &brArr);

//get the brArr of this tree after optimizing
static void getBr (BrArr<double> &brArr);

//get the exNdArr of this tree after optimizing
static void getExNd (ExNdArr &ExNdArr);

//get the InNdArr of this tree after optimizing
static void getInNd (InNdArr &InNdArr);

static double getBrLen (int brNo);

static double opt(int nIt = MAX_IT_UR_BR);


//opt all childrent bres of this node
static void opt(int ndNo, int &nChangedBr);

//clean all content of this class
static void clean();

//everything is inited here
static void init ();

//return the log likelihood of this tree.
static double getLogLi ();

//write all information of this optimal unrooted tree into the screen
static void write ();

static void updateBestLogLi (double logLi);

//release all memmory for this class

void release ();

//the destructor of this class
virtual ~OptUrTree();

/***********************************************************************************
***********************************************************************************
***********************************************************************************

***********************************************************************************
***********************************************************************************/

//private :

static Vec<int> inBrNoLs_;

static double bestLogLi_;

//the current number of inNds in this tree
static int curNInNd_;

//the current number of ExNds in this tree
static int curNExNd_;

//the current number of bres in this tree
static int curNBr_;

//the maximum number of inNds in this tree
static int maxNInNd_;

//the maximum number of ExNds in this tree
static int maxNExNd_;

//the maximum number of bres in this tree
static int maxNBr_;

//the outGrpNdNo if this unrooted tree
static int outGrpNdNo_;


//the outGrpBrNo if this unrooted tree
static int outGrpBrNo_;


//the artificial internal root node if this unrooted tree
static int artInRootNo_;

//the array of all internal nds of this UnrootTree
static InNdArr inNdArr_;

//the array of all ex nds of this UnrootTree
static ExNdArr exNdArr_;

//the array of all bres of this UnrootTree
static LiBrArr liBrArr_;

//isLiCmped = 0, it means that the likelihood of this tree has NO been computed
static int isLiCmped_;


//isBrLenEsted = 1, it means that we have estimated the len of all branches
static int isBrLenEsted_;

//contains the branch no which is changed
static int actBrNo_;

//it contains the genetic distance between any two sequences
static Mat<double> *genDisMat_;

//the number of calling function cmpNegLogLiBr
static int nCmpNegLogLiBr_;

//the number of compute likehood of an internal node
static int nCmpLi_;

static int nCmpExLogLi_;
static int nCmpInLogLi_;

//the number of calling cmpLogLi
static int nCmpLogLi_;

static int isInNd (int ndNo);

//check whether or not it is a external node
static int isExNd (int ndNo);

//create exDesLiInNd for all internal nodes of this tree
static void createExDesLiNd ();

//create pattern for all internal node of this tree
static void createPtnLiNd ();

/*create the list of all exnternal nodes which belong to the same one root */
static void createGrpExNd (const int parGrpNo, const int grpNo, Vec<int> &grpExNdNoLs);


/*compute the distance between two groups, group 1 and group 2.
group1 is a subtree whose the root is grpNo1
group2 is a subtree whose the root is grpNo2
parentNo is the parent of grpNo1, and grpNo2
*/
static double cmpGrpDis (const int parGrpNo, const int grpNo1, const int grpNo2);
static double cmpGrpDis (Vec<int> &grpExNdNoLs1, Vec<int> &grpExNdNoLs2);
static double cmpGrpDis (int exNdNo, Vec<int> &grpExNdNoLs);


/*using Rzhetsky and Nei method to estimate the branch of this tree as
a starting point for maximum likelihood*/
static double estExBrLen (Br<double> &exBr);

/*using Rzhetsky and Nei method to estimate the branch of this tree as
a starting point for maximum likelihood*/
static double estInBrLen (Br<double> &inBr);

/*using Rzhetsky and Nei method to estimate the branch of this tree as
a starting point for maximum likelihood*/
static void estBrLen ();


//assign values for all LiNd_ of all LiNd
static void cmpLiNd ();

//assign values for all LiNd_ of all LiNd
static void cmpHeadLiNd ();

//compute the likelihood at a given pattern site

static long double cmpPtnLi (int ptnNo_);


//cmp the log likelihood of this tree
static double cmpLogLi ();

//serve for cmpLiBr and cmpLogLiBr
static void changeBr (double len);

//cmp the negative log likelihood of this tree when only this br changes
//using for optimizing by brent method
static double cmpNegLogLiBr (double len);

//optimize the br whose name is brNo, return the length of br
static double optBr (int brNo);


}; //end of class optimizing unrooted tree
#endif // OPTIMALUNROOTTREE


