#ifndef RATE_H
#define RATE_H

#include "vec.h"
#include "seqpair.h"

class Rate {
//====================================================================================

public :

//====================================================================================
/*...Being specified more
*/
//set the number of catagories 
static void setNRate (const int nRate);

//set the shape of gamma distribution
static void setGammaShape (const double gammaShape);
static void setGammaShapeType (const PAM_TYPE gammaShapeType);

//return the fraction of number of invariable sites over the number of sites
static double getInvarRatio ();

static PAM_TYPE getGammaShapeType ();

//return the shape of gamma distribution
static double getGammaShape ();

// get the nCategories mean rates of nCatefories different categories
static Vec<double> &getRate ();

//return the rateArr_[rateNo];
static double getRate (int rateNo);

//get the prob of rateNo
static double getProb ();

//return the number of categories rate
static int getNRate ();

static double cmpNegLogLiGammaShape (double gammaShape);

//optimize all parameters of Rate class
static void optRatePam ();

//init everything here
static void init ();

//create the mean rate list using Gamma distribution, using existed parameters
static void create ();

static void setType (RATE_TYPE rateType);
static RATE_TYPE getType ();

//====================================================================================
//optimize specific site substitution rates using Sonja Meyer and Arndt von Haeseler, MBE-2003
static int optSpecificSite (Vec<SeqPair> &indSeqPairArr);

//==================================================================
//optimize specific site substitution rates using Sonja Meyer and Arndt von Haeseler, MBE-2003
static Vec<double> *getPtnRate ();

//==================================================================
//optimize specific site substitution rates using Sonja Meyer and Arndt von Haeseler, MBE-2003
static double getPtnRate (int ptnNo);

static int isOptedSpecificRate ();

//optimize the gamma shape parameter using Brent method
//static void optGammaShape ();

//release all memmory for this class
void release ();

//the destructor function 
virtual ~Rate ();
/********************************************************************************
********************************************************************************
********************************************************************************

********************************************************************************
********************************************************************************
********************************************************************************/

private :
//the number of categories
static int nRate_;

static RATE_TYPE rateType_;
static PAM_TYPE gammaShapeType_;

static Vec<double> rateArr_;
static Vec<double> ptnRateArr_;
static Vec<SeqPair> indSeqPairArr_;

static int actPtnNo_;

static int isOptedSpecificRate_;

//====================================================================================
//cmp likelihood when active pattern changes using Sonja Meyer and Arndt von Haeseler, MBE-2003
static double cmpNegLogLiPtn (double ptnRate);

static void scalePtnRate ();

//====================================================================================
//optimize specific site substitution rate for PTNNO using Sonja Meyer and Arndt von Haeseler, MBE-2003
static double optPtn (int ptnNo);

//====================================================================================
//optimize specific site substitution rates using Sonja Meyer and Arndt von Haeseler, MBE-2003
static void estimatePtnRate ();

//the shape parameter of gamma distribution
static double xxx_gammaShape_;

/* discrete Gamma according to Yang 1994 (JME 39:306-314) */
static void doYangDiscreteGamma (const double shape, const int nRate, Vec<double> &rateArr);



//Normally, beta is assigned equal to alpha
static double cmpPerPointGamma (const double prob, const double shape);



/*NUMERICAL SUBROUTINES
THE FOLLOWING CODE COMES FROM tools.c in Yang's PAML package */

/* returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
   Stirling's formula is used for the central polynomial part of the procedure.
   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684

*/
static double cmpLnGamma (double alpha);

/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
   limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   (1) series expansion     if (alpha>x || x<=1)
   (2) continued fraction   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
static double cmpIncompleteGamma (double x, double alpha, double ln_gamma_alpha);

/* functions concerning the CDF and percentage points of the gamma and
   Chi2 distribution
   returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.
*/
static double cmpPointNormal (double prob);


/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
static double cmpPointChi2 (double prob, double v);

/* THE END OF THE CODES COMMING FROM tools.c in Yang's PAML package */

};//end of class Rate
#endif
