#include "testmethod.h"

#include <cstring>
#include <ctime>
#include <fstream>
#include <math.h>
#include <time.h>
#include <ctime>

#include "testmethod.h"
#include "constant.h"
#include "rate.h"
#include "model.h"
#include "ali.h"
#include "utl.h"
#include "outstream.h"
#include "usertree.h"
#include "ptnls.h"
#include "treels.h"
#include "interface.h"
#include "iqp.h"

using namespace std;

//--------------------------------------------------------------------
//we will init all things here before starting testing this method with data
void TestMethod::init () {
}

//--------------------------------------------------------------------
//set all paramaters of this class for Model class, Rate class
void TestMethod::setPam (DATA_TYPE dataType,
                         MODEL model,
                         double baseFrqA, double baseFrqC, double baseFrqG, double baseFrqT,
                         PAM_TYPE baseFrqType,     

                         double tsTvRatio,
                         PAM_TYPE tsTvRatioType,
           
                         double pyPuRatio,
                         PAM_TYPE pyPuRatioType,

                         double tsAG, double tsCT, double tvAC, double tvAT, double tvCG, double tvGT,
                         PAM_TYPE genPamType,
     
                         RATE_TYPE rateType,  
                         int nRate,

                         double gammaShape,
                         PAM_TYPE gammaShapeType) {

  Model::setDataType (dataType);   
  Model::setModelType (model);


  Model::setTsTvRatio (tsTvRatio * 2.0);
  Model::setTsTvRatioType (tsTvRatioType);

  Model::setPyPuRatio (pyPuRatio);
  Model::setPyPuRatioType (pyPuRatioType);

  Model::setGenPam (tsAG, tsCT, tvAC, tvAT, tvCG, tvGT);
  Model::setGenPamType (genPamType);

  Model::setStateFrq (baseFrqA, baseFrqC, baseFrqG, baseFrqT);
  Model::setBaseFrqType (baseFrqType);

  Model::init ();

  Rate::init ();
  Rate::setType (rateType);

  if (rateType == GAMMA)
     Rate::setNRate (nRate);  
  Rate::setGammaShape (gammaShape);
  Rate::setGammaShapeType (gammaShapeType);
  Rate::create ();
}


//--------------------------------------------------------------------
/*do the quartet puzzling method for all data set from the input file
return the number of correct predicted tree
*/
int TestMethod::create ( const char *userTreeFile, const char *aliFile, 
                        int        nIter,       
                        double     probSel,
                        int        nRep,
                        STOPPING_TYPE     stoppingRule,
                        DATA_TYPE  dataType,
                        MODEL      model,
                        double      baseA, double baseC, double baseG, double baseT,
                        PAM_TYPE    baseFrqType,

                        double      tsTvRatio,
                        PAM_TYPE    tsTvRatioType,

                        double      pyPuRatio,
                        PAM_TYPE    pyPuRatioType,

                        double      tsAG, 
                        double      tsCT, 
                        double      tvAC, 
                        double      tvAT, 
                        double      tvCG, 
                        double      tvGT, 
                        PAM_TYPE    gTRPamType,

                        RATE_TYPE   rateType,
                        int         nRate,
                        
                        double      gammaShape,
                        PAM_TYPE    gammaShapeType,

                        int         outGrpSeqNo) {


//  long double start_ = clock ();
  time_t beginTime_;
  time (&beginTime_);
   
   char *startedDate_;
   startedDate_ = ctime(&beginTime_);      
   std::cout << "Date and time: " << startedDate_;

   int nSeq_;
   nSeq_ = Ali::getNSeq ();
           
  //create the pattern list
   PtnLs::init ();
   PtnLs::create ();
   std::cout << "Number patterns = " << PtnLs::getNPtn () << endl;
//   PtnLs::write (std::cout);
  
   setPam (dataType, model, 
           baseA, baseC, baseG, baseT,
           baseFrqType,     

           tsTvRatio,
           tsTvRatioType,
           
           pyPuRatio,
           pyPuRatioType,

           tsAG, tsCT, tvAC, tvAT, tvCG, tvGT,
           gTRPamType,
     
           rateType,  
           nRate,
           gammaShape,
           gammaShapeType);
      
   Ali::init ();
   std::ifstream userTreeFile_;
   userTreeFile_.open (userTreeFile);

      
   if (userTreeFile_ == 0) {
//      staFile_ << "can not open userTreeFile : " << userTreeFile << endl;
      userTreeFile_.close ();
   } else {
      UserTree userTree_;
      userTree_.readFile (userTreeFile);
      userTree_.createUrTree ();
      //userTree_.draw (BR_LEN, 0, std::cout);
   
      double userLogLi_ =  userTree_.doConOpt (0);    

      std::cout.precision (10);
      std::cout << "User log likelihood: " << userLogLi_ << endl;
   }
  UrTree tree_;
  IQP iqp_;
  int acceptedNRep_ = Utl::getMin (nRep + 1, MAX_NUM_REP - 1);
  Vec<char> conTree_;
  double progTime_;
  iqp_.escapeLocalOpt (nIter, probSel, acceptedNRep_, stoppingRule, tree_, outGrpSeqNo, conTree_, beginTime_, progTime_);



  Interface interface_;
  tree_.createRootedTree(outGrpSeqNo);   

  interface_.writeOut (aliFile,  tree_, nIter, probSel, nRep, stoppingRule, conTree_, startedDate_, progTime_);
  

  PtnLs::clean ();
  Ali::clean ();
   
  time_t finishedTime_;
  time (&finishedTime_);
  char *finishedDate_;
  finishedDate_ = ctime(&finishedTime_);      
  std::cout << "Date and time: " << finishedDate_;

  return 1;
}

