/***************************************************************************
                          general time reversible modell.h  -  description
                             -------------------
    begin                : Thu Feb 20 2003
    copyright            : (C) 2003 by Vinh Le Sy and Arndt von Haeseler
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TN93M_H
#define TN93M_H

#include "basem.h"
 /*
general time reversible model
*/
class TN93M: public BaseM
{
public :


//cmp the base frequencies for this alignment
//virtual void cmpBaseFrq ();

//all things are inited here
virtual void init ();

/*
cmp the probability of changing from nucleotide baseNo1 into nucleotide baseNo2
after a period of brLen  
*/
virtual double cmpProbChange (const int baseNo1, const int baseNo2, const double brLen);


//cmp the vector probability of change from baseNo
virtual void cmpProbChange (const int baseNo, const double brLen, DVec20 &probVec);

/*
cmp the array of probability of changing from
one nucleotide into one nucleotide after a period of brLen / subRate_
*/
virtual void cmpProbChange (const double brLen, DMat20 &probMat);


virtual void optPam (const int maxNIter, const PAM_TYPE tsTvRatioType, const PAM_TYPE pyPuRatioTyp);


static void reInit ();
/*compute the log likelihood when given a value
this function is used to optimize the parameter of this model
*/
static double cmpNegLogLi (double value);


static void elmhes(DMat20 a, Vec<int> &ordr, int n);
static void eltran(DMat20 a, DMat20 zz, Vec<int> &ordr, int n);
static void mcdiv(double ar, double ai, double br, double bi,
                          double *cr, double *ci);

static void hqr2(int n, int low, int hgh, DMat20 h,
                        DMat20 zz, DVec20 &wr, DVec20 &wi);
static void onepamratematrix(DMat20 a);
static void eigensystem(DVec20 &eval, DMat20 &evec);
static void luinverse(DMat20 inmat, DMat20 imtrx, int size);
static void checkevector(DMat20 evec, DMat20 ivec, int nn);
static void tranprobmat();
static void rtfdata (DMat20 &a, DVec20 &forg);
static void createRateMat ();

void writeMat (DMat20 &rate);
void writeInt (Vec<int> &vec);
void writeDouble (DVec20 &vec);



//release all memmory for this class
void release ();

}; //end of General model class
#endif
